/* (C) 2003 XDA Developers  itsme@xs4all.nl
 *
 * $Header: /var/db/cvs/xdautils/common/debug.cpp,v 1.15 2004/12/25 12:30:14 itsme Exp $
 */
#include <windows.h>

#include "debug.h"        // declarations for this file.
#include <string>

#include "stringutils.h"
#include "vectorutils.h"

#ifndef _WIN32_WCE
#ifndef _NO_RAPI
#include <rapi.h>
#endif
#endif

#define DBG_OUTPUTDEBUGSTRING 1
#define DBG_MESSAGEBOX        2
#define DBG_DEBUGWINDOW       4
#define DBG_LOGFILE           8
#define DBG_STDOUT           16

#ifdef _WIN32_WCE
//  define these for wince use
#define DBG_WCHAROUTPUT (DBG_OUTPUTDEBUGSTRING|DBG_MESSAGEBOX|DBG_DEBUGWINDOW)
#define DBG_CHAROUTPUT (DBG_LOGFILE|DBG_STDOUT)
#else
#ifdef _UNICODE
//  define these for unicode desktop use
#define DBG_WCHAROUTPUT (DBG_OUTPUTDEBUGSTRING|DBG_MESSAGEBOX|DBG_DEBUGWINDOW)
#define DBG_CHAROUTPUT (DBG_LOGFILE|DBG_STDOUT)
#else
//  define these for non unicode desktop use
#define DBG_WCHAROUTPUT (0)
#define DBG_CHAROUTPUT (DBG_LOGFILE|DBG_STDOUT|DBG_OUTPUTDEBUGSTRING|DBG_MESSAGEBOX|DBG_DEBUGWINDOW)
#endif
#endif

int g_debugOutputFlags;
char g_logfilename[MAX_PATH+1];
HANDLE g_log_mutex;
CDebugWindow *g_debugWindow;

#define MAX_DEBUG_LENGTH 16384

void DebugSetLogfile(const char *filename)
{
    g_log_mutex= CreateMutex(NULL, false, NULL);
    strncpy(g_logfilename, filename, MAX_PATH);
    g_logfilename[MAX_PATH]=0;
    g_debugOutputFlags |= DBG_LOGFILE;
}
void DebugRegisterWindow(CDebugWindow *wnd)
{
	if (wnd==NULL) {
		g_debugOutputFlags &= ~DBG_DEBUGWINDOW;
		g_debugWindow= wnd;
	}
	else {
		g_debugWindow= wnd;
		g_debugOutputFlags |= DBG_DEBUGWINDOW;
	}
}
void DebugStdOut()
{
    g_debugOutputFlags |= DBG_STDOUT;
}
void DebugMessagebox()
{
    g_debugOutputFlags |= DBG_MESSAGEBOX;
}
void DebugOutputDebugString()
{
    g_debugOutputFlags |= DBG_OUTPUTDEBUGSTRING ;
}
void wdebugoutput(const WCHAR *wbuf)
{
#if DBG_WCHAROUTPUT&DBG_OUTPUTDEBUGSTRING
    if (g_debugOutputFlags&DBG_OUTPUTDEBUGSTRING) {
        int len= _tcslen(wbuf);
        for (int i= 0 ; i<len ; i+=512)
        {
            WCHAR smallbuf[513];
            wcsncpy(smallbuf, wbuf+i, 512);
            smallbuf[512]= 0;
            OutputDebugString(smallbuf);
        }
    }
#endif
#if DBG_WCHAROUTPUT&DBG_MESSAGEBOX
    if (g_debugOutputFlags&DBG_MESSAGEBOX) {
        MessageBox(0,wbuf,L"debug",0);
    }
#endif
#if DBG_WCHAROUTPUT&DBG_DEBUGWINDOW
    if (g_debugOutputFlags&DBG_DEBUGWINDOW && g_debugWindow) {
        g_debugWindow->appendString(wbuf);
    }
#endif
#if DBG_WCHAROUTPUT&DBG_STDOUT
    if (g_debugOutputFlags&DBG_STDOUT) {
        fputws(wbuf, stdout);
    }
#endif
#if DBG_WCHAROUTPUT&DBG_LOGFILE
        WaitForSingleObject(g_log_mutex, INFINITE);
        FILE *f= fopen(g_logfilename, "a+");
        fputws(wbuf, f);
        fclose(f);
        ReleaseMutex(g_log_mutex);
#endif
}
void debugoutput(const char *buf)
{
#if DBG_CHAROUTPUT&DBG_OUTPUTDEBUGSTRING
    if (g_debugOutputFlags&DBG_OUTPUTDEBUGSTRING) {
        int len= (int)strlen(buf);
        for (int i= 0 ; i<len ; i+=512)
        {
            char smallbuf[513];
            strncpy(smallbuf, buf+i, 512);
            smallbuf[512]= 0;
            OutputDebugString(smallbuf);
        }
    }
#endif
#if DBG_CHAROUTPUT&DBG_MESSAGEBOX
    if (g_debugOutputFlags&DBG_MESSAGEBOX) {
        MessageBox(0,buf,"debug",0);
    }
#endif
#if DBG_CHAROUTPUT&DBG_DEBUGWINDOW
    if (g_debugOutputFlags&DBG_DEBUGWINDOW && g_debugWindow) {
        g_debugWindow->appendString(buf);
    }
#endif
#if DBG_CHAROUTPUT&DBG_STDOUT
    if (g_debugOutputFlags&DBG_STDOUT) {
        fputs(buf, stdout);
    }
#endif
#if DBG_CHAROUTPUT&DBG_LOGFILE
    if (g_debugOutputFlags&DBG_LOGFILE) {
        WaitForSingleObject(g_log_mutex, INFINITE);
        FILE *f= fopen(g_logfilename, "a+");
        fputs(buf, f);
        fclose(f);
        ReleaseMutex(g_log_mutex);
    }
#endif
}

std::wstring vwdebugmsg(const WCHAR *msg, va_list ap)
{
	std::wstring wbuf; wbuf.resize(MAX_DEBUG_LENGTH);

	_vsnwprintf(stringptr(wbuf), wbuf.size(), msg, ap);
	wbuf.resize(wcslen(wbuf.c_str()));

	return wbuf;
}
std::string vdebugmsg(const char *msg, va_list ap)
{
	std::string buf; buf.resize(MAX_DEBUG_LENGTH);

	_vsnprintf(stringptr(buf), buf.size(), msg, ap);
	buf.resize(strlen(buf.c_str()));

	return buf;
}



void vwdebug(const WCHAR *msg, va_list ap)
{
	std::wstring wbuf= vwdebugmsg(msg, ap);
    if (g_debugOutputFlags&DBG_CHAROUTPUT) {
        debugoutput(ToString(wbuf).c_str());
    }
    wdebugoutput(wbuf.c_str());
}

void vdebug(const char *msg, va_list ap)
{
	std::string buf= vdebugmsg(msg, ap);
    if (g_debugOutputFlags&DBG_WCHAROUTPUT) {
        wdebugoutput(ToWString(buf).c_str());
    }
    debugoutput(buf.c_str());
}
void wdebug(const WCHAR *msg, ...)
{
    va_list ap;

    va_start(ap, msg);
    vwdebug(msg, ap);
    va_end(ap);
}
void debug(const char *msg, ...)
{
    va_list ap;

    va_start(ap, msg);
    vdebug(msg, ap);
    va_end(ap);
}

void verror(DWORD dwErrorCode, const char *msg, va_list ap)
{
    // todo: make this charsize independent
    TCHAR* msgbuf;
    int rc= FormatMessage(
            FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
            NULL, dwErrorCode, 0, (TCHAR*) &msgbuf, 0, NULL);
    if (rc)
    {
#ifdef _UNICODE
        debug("ERROR: %hs - %ls\n", vdebugmsg(msg, ap).c_str(), msgbuf);
#else
        debug("ERROR: %hs - %hs\n", vdebugmsg(msg, ap).c_str(), msgbuf);
#endif
        LocalFree(msgbuf);
    }
    else
        debug("ERROR: %hs - UNKNOWNERROR: 0x%08lx\n", vdebugmsg(msg, ap).c_str(), dwErrorCode);
}
void error(DWORD dwErrorCode, const char *msg, ...)
{
    va_list ap;
    va_start(ap, msg);
    verror(dwErrorCode, msg, ap);
    va_end(ap);
}

void error(const char *msg, ...)
{
    va_list ap;
    va_start(ap, msg);
    verror(GetLastError(), msg, ap);
    va_end(ap);
}

#ifndef _WIN32_WCE
#ifndef _NO_RAPI
void ceerror(const char *msg, ...)
{
    va_list ap;
    va_start(ap, msg);

    DWORD dwError= CeRapiGetError();
    if (!dwError)
        dwError= CeGetLastError();

    verror(dwError, msg, ap);
    va_end(ap);
}
#endif
#endif

void debugt(char *msg, ...)
{
	SYSTEMTIME now;
	GetLocalTime(&now);
	
	va_list ap;
    va_start(ap, msg);
	debug("TIMESTAMP: %04d-%02d-%02d %02d:%02d:%02d.%03d %hs", 
		now.wYear, now.wMonth, now.wDay,
		now.wHour, now.wMinute, now.wSecond, now.wMilliseconds,
		vdebugmsg(msg, ap).c_str());
    va_end(ap);
}

std::string dumponeunit(const BYTE *p, size_t len, DumpUnitType unittype)
{
    switch(unittype) {
        case DUMPUNIT_BYTE: return stringformat("%02x", *p);
        case DUMPUNIT_WORD:
                if (len==1)
                    return stringformat("__%02x", *p);
                else
                    return stringformat("%04x", *(WORD*)p);
                break;
        case DUMPUNIT_DWORD:
                if (len==1)
                    return stringformat("______%02x", *p);
                else if (len==2)
                    return stringformat("____%04x", *(WORD*)p);
                else if (len==3)
                    return stringformat("__%02x%04x", p[2], *(WORD*)p);
                else
                    return stringformat("%08x", *(DWORD*)p);
                break;
    }
    return "";
}
std::string hexdumpunit(const BYTE *buf, size_t bytelen, DumpUnitType unittype)
{
    std::string str;

    str.reserve(bytelen*3);
    for (size_t i=0 ; i<bytelen ; i+=DumpUnitSize(unittype)) {
        if (!str.empty())
            str += " ";
        str += dumponeunit(buf+i, bytelen-i, unittype);
    }

    return str;
}


std::string dumpraw(const BYTE *buf, size_t len, size_t &usedlen)
{
    std::string str((char*)buf, len);
    str.resize(strlen(str.c_str()));
    size_t i= str.size();
    while (i<len && buf[i]==0) i++;

    usedlen= i;

    return str;
}
// todo: add 'summarize' option, to generate 'dup' constructs.
//       make strings more readable, by breaking on NUL.
std::string dumpstrings(const BYTE *buf, size_t len, size_t &usedlen)
{
    std::string result;
    bool bQuoted= false;
    bool bThisIsEolChar= false;
    std::string escaped= "\n\r\t";

    size_t i;
    for (i=0 ; i<len ; i++)
    {
        bool bNeedsEscape= escaped.find((char)buf[i])!=escaped.npos 
            || buf[i]=='\"' 
            || buf[i]=='\\';

        if (isprint(buf[i]) || bNeedsEscape) {
            if (!bQuoted) {
				if (!result.empty())
					result += ",";
                result += "\"";
                bQuoted= true;
            }
            if (bNeedsEscape) {
                std::string escapecode;
                switch(buf[i]) {
                    case '\n': escapecode= "\\n"; break;
                    case '\r': escapecode= "\\r"; break;
                    case '\t': escapecode= "\\t"; break;
                    case '\"': escapecode= "\\\""; break;
                    case '\\': escapecode= "\\\\"; break;
                    default:
                       escapecode= stringformat("\\x%02x", buf[i]);
                }
                result += escapecode;
            }
            else {
                result += (char) buf[i];
            }
        }
        else {
            if (bQuoted) {
                result += "\"";
                bQuoted= false;
            }
			if (!result.empty())
				result += ",";
            result += stringformat("%02x", buf[i]);
        }
        if (i+1<len) {
            if (i==0)
                bThisIsEolChar = (buf[i]==0x0a || buf[i]==0x0d);

            bool bNextIsEolChar= (buf[i+1]==0x0a || buf[i+1]==0x0d);
            if (bThisIsEolChar && !bNextIsEolChar) {
                i++;
                break;
            }
            bThisIsEolChar= bNextIsEolChar;
        }
    }

	if (bQuoted) {
		result += "\"";
		bQuoted= false;
	}

    usedlen= i;

    return result;
}
void bighexdump(const ByteVector& data, DWORD flags/*=hexdumpflags(16, DUMPUNIT_BYTE, DUMP_HEX_ASCII)*/)
{
    bighexdump(0, data, flags & ~ (1<<16));
}
void bighexdump(DWORD dwOffset, const ByteVector& data, DWORD flags/*=hexdumpflags(16, DUMPUNIT_BYTE, DUMP_HEX_ASCII)*/)
{
    DumpUnitType dumpunittype= dumpunit_from_flags(flags);
    DumpFormat dumpformat= dumpformat_from_flags(flags);
    int unitsperline= unitsperline_from_flags(flags);
    bool bWithOffset= (flags&HEXDUMP_WITH_OFFSET)!=0;
    bool bSummarize= (flags&HEXDUMP_SUMMARIZE)!=0;

    bool bLastBlock= (flags&HEXDUMP_MOREFOLLOWS)==0;

    size_t bytesperline= unitsperline*DumpUnitSize(dumpunittype);

    if (dumpformat==DUMP_STRINGS || dumpformat==DUMP_RAW) {
        bytesperline= data.size();
    }

    std::string prevline;
    bool bSummaryPrinted= false;

    for (size_t i=0 ; i<data.size() ; i+=bytesperline) {
        // not using 'min' since msvc's header files are broken, and make it
        // quite hard to include them in an order as not to redefine 'min' in
        // an inconvenient way.

        size_t len= bytesperline; if (len > data.size()-i) len= data.size()-i;

        std::string line;
        if (dumpformat==DUMP_STRINGS) {
            line= dumpstrings(vectorptr(data)+i, data.size()-i, bytesperline);
        }
        if (dumpformat==DUMP_RAW) {
            line= dumpraw(vectorptr(data)+i, data.size()-i, bytesperline);
        }
        if (dumpformat==DUMP_HEX_ASCII || dumpformat==DUMP_HEX) {
            line= hexdumpunit(vectorptr(data)+i, len, dumpunittype);
            if (len < bytesperline) {
                int charsinfullline= (2*DumpUnitSize(dumpunittype)+1)*unitsperline-1;
                line.append(charsinfullline-line.size(), ' ');
            }
        }
        if (dumpformat==DUMP_HEX_ASCII || dumpformat==DUMP_ASCII)  {
            if (!line.empty())
                line += "  ";
            line += asciidump(vectorptr(data)+i, len);
            if (len < bytesperline) {
                line.append(bytesperline-len, ' ');
            }
        }

        if (bSummarize && line == prevline) {
            if (!bSummaryPrinted)
                debug("*\n");       // -- this is 'the summary'
            bSummaryPrinted= true;
        }
        else {
            bSummaryPrinted= false;

            if (bWithOffset)
                debug("%08lx: %hs\n", dwOffset+i, line.c_str());
            else
                debug("%hs\n", line.c_str());
        }

        prevline= line;
    }
    if (bLastBlock && bSummaryPrinted)
        debug("%08lx: \n", dwOffset+data.size());
}
