#include <windows.h>
#include <psyscall.h>
#include <storemgr.h>
#include "tffsreader.h"
#include "ceioctl.h"
#include "debug.h"
#include "ptrutils.h"
#include "stringutils.h"

#include "diskio.h"     // for SG_REQ

// in itsutils
int GetDiskInfo(HANDLE hDsk, DISK_INFO &info);

#include "cenk.h"

HDATA *cvHandle2HDataPtr(HANDLE h)
{
	struct KDataStruct *k= (struct KDataStruct *)0xffffc800;

    return (HDATA *)(k->handleBase+((DWORD)h&0x1ffffffc));
}
std::string cinfostring(const CINFO *ci)
{
    if (ci)
        return stringformat("cinfo{%c%c%c%c, n=%d, svr=%08lx}", ci->acName[0], ci->acName[1], ci->acName[2], ci->acName[3], ci->cMethods, ci->pServer);
    else
        return "(null)";
}
std::string handleinfostring(HANDLE h)
{
	HDATA *hi= cvHandle2HDataPtr(h);
    return stringformat("hdata{hv=%08lx ci=%hs pvo=%08lx}", hi->hValue, cinfostring(hi->pci).c_str(), hi->pvObj);
}

HANDLE FindFirstStore(STOREINFO *psi)
{
    return (FindFirstFileW( L"\\StoreMgr", (WIN32_FIND_DATA *)psi));
}
bool GetStoreMgrList(std::vector<STOREINFO>& list)
{
    STOREINFO si; memset(&si, 0, sizeof(STOREINFO));
    si.cbSize= sizeof(STOREINFO);
    int currentStore = 0;
    HANDLE hEnumStores = FindFirstStore(&si);
    if (hEnumStores == INVALID_HANDLE_VALUE) {
        error("FindFirstFile");
        return false;
    }
    list.push_back(si);
    while (PSLFindNextStore(hEnumStores, &si)) {
        list.push_back(si);
    }

    if (GetLastError()!=ERROR_NO_MORE_ITEMS) {
        error("PSLFindNextStore");
        return !list.empty();
    }

    if (!PSLFindCloseStore(hEnumStores)) {
        error("PSLFindCloseStore");
        return !list.empty();
    }

    return true;
}
void cvSTOREINFO2DevStoreinfo(const STOREINFO& si, DevStoreInfo& dsi)
{
    dsi.llStoreSize= si.snNumSectors * si.dwBytesPerSector;

    wcsncpy(dsi.szDeviceName, si.szDeviceName, DEVICENAMESIZE);

    debug("STOREINFO dev='%ls'  store='%ls'", si.szDeviceName, si.szStoreName);
    debug("  nsect=%x bpsect=%x free=%x maxpartsize=%x\n",
            (DWORD)si.snNumSectors,
            si.dwBytesPerSector,
            (DWORD)si.snFreeSectors,
            (DWORD)si.snBiggestPartCreatable);
}

ITSUTILS_API HRESULT STDAPICALLTYPE ITGetStoreMgrList(
        DWORD cbInput, void *pbInput,
        DWORD *pcbOutput, GetStoreMgrListResult **ppbOutput,
        IRAPIStream *pStream)
{
    *pcbOutput= 0;
    GetStoreMgrListResult *pOut= *ppbOutput= NULL;

    std::vector<STOREINFO> list;

    if (!GetStoreMgrList(list)) {
        error("GetStoreMgrList");
        return GetLastError();
    }

    *pcbOutput= PTR_DIFF(pOut, &pOut->storeInfo[list.size()]);

    pOut= *ppbOutput= (GetStoreMgrListResult*)LocalAlloc(LPTR, *pcbOutput);
    if (pOut==NULL) {
        error("LocalAlloc(%d)", *pcbOutput);
        return GetLastError();
    }

    pOut->nStores = list.size();

    for (size_t i=0 ; i<list.size() ; i++) {
        cvSTOREINFO2DevStoreinfo(list[i], pOut->storeInfo[i]);
    }

    return 0;
}
bool ReadProcessDword(HANDLE hProc, DWORD dwMemOffset, DWORD &dw)
{
    DWORD nCopied;
    if (!ReadProcessMemory(hProc, (void*)dwMemOffset, &dw, sizeof(DWORD), &nCopied))
        return false;

    if (nCopied!=sizeof(DWORD)) {
        SetLastError(ERROR_INTERNAL_ERROR);
        return false;
    }

    return true;
}
bool GetSTRGHandleList(std::vector<DWORD> &list)
{
	HDATA *hi= cvHandle2HDataPtr((HANDLE)GetCurrentThreadId());
	HDATA *ha;
	ha=hi;

    BOOL bMode = SetKMode(TRUE);
    DWORD dwPerm = SetProcPermissions(0xFFFFFFFF);

	do {
        DWORD dw;
        if (memcmp(ha->pci->acName, "STRG", 4)==0
                && ReadProcessDword(ha->pci->pServer->hProc, (DWORD)ha->pvObj, dw)
                && dw == 0x64615900) {
		    list.push_back((DWORD)ha->hValue);
        }
		ha= (HDATA*)ha->linkage.fwd;
	} while (ha!=hi && ha!=NULL);

    SetProcPermissions(dwPerm);
    SetKMode(bMode);

    return true;
}

ITSUTILS_API HRESULT STDAPICALLTYPE ITGetSTRGHandleList(
        DWORD cbInput, void *pbInput,
        DWORD *pcbOutput, GetSTRGHandleListResult **ppbOutput,
        IRAPIStream *pStream)
{
    *pcbOutput= 0;
    GetSTRGHandleListResult *pOut= *ppbOutput= NULL;

    std::vector<DWORD> list;

    if (!GetSTRGHandleList(list)) {
        debug("ERROR: GetSTRGHandleList failed\n");
        return ERROR_INTERNAL_ERROR;
    }

    *pcbOutput= PTR_DIFF(pOut, &pOut->handle[list.size()]);

    pOut= *ppbOutput= (GetSTRGHandleListResult*)LocalAlloc(LPTR, *pcbOutput);
    if (pOut==NULL) {
        error("LocalAlloc(%d)", *pcbOutput);
        return GetLastError();
    }
    pOut->nHandles= list.size();

    for (size_t i=0 ; i<list.size() ; i++) {
        pOut->handle[i]= list[i];
    }

    return 0;
}


HANDLE WINAPI OpenStore(LPCTSTR szDeviceName)
{
    TCHAR szFileName[MAX_PATH];
    memset( szFileName, 0, sizeof(szFileName));
    wcscpy( szFileName, L"\\StoreMgr\\");
    __try {
        wcsncat( szFileName, szDeviceName, MAX_PATH-15);
    } __except(EXCEPTION_EXECUTE_HANDLER) {
        SetLastError(ERROR_BAD_ARGUMENTS);
    }
    return CreateFile( szFileName, 0, 0, NULL, 0, 0, NULL);
}
bool GetStorePartitionlist(HANDLE hStore, std::vector<PARTINFO>& list)
{
    PARTINFO pi; memset(&pi, 0, sizeof(pi));
    pi.cbSize= sizeof(PARTINFO);

    HANDLE hEnumPartitions = PSLFindFirstPartition(hStore, &pi);
    if (hEnumPartitions == INVALID_HANDLE_VALUE) {
        error("PSLFindFirstPartition");
        return false;
    }
    list.push_back(pi);
    while (PSLFindNextPartition(hEnumPartitions, &pi))
        list.push_back(pi);
   
    if (GetLastError()!=ERROR_NO_MORE_ITEMS) {
        error("PSLFindNextPartition");
        return !list.empty();
    }
    if (!PSLFindClosePartition(hEnumPartitions)) {
        error("PSLFindClosePartition");
        return !list.empty();
    }
    return true;
}
void cvPARTINFO2DevPartitioninfo(const STOREINFO& si, const PARTINFO& pi, DevPartitionInfo& dpi)
{
    dpi.llPartitionSize= pi.snNumSectors * si.dwBytesPerSector;

    wcsncpy(dpi.szPartitionName, pi.szPartitionName, PARTITIONNAMESIZE);

    debug("PARTINFO name='%ls'  filesys='%ls' volname='%ls'",
        pi.szPartitionName, pi.szFileSys, pi.szVolumeName);
    debug("  end=%x  type=%02x\n",
            (DWORD)pi.snNumSectors, pi.bPartType);
}
bool GetStoreInfo(HANDLE hStore, STOREINFO& si)
{
    memset(&si, 0, sizeof(STOREINFO));
    si.cbSize= sizeof(STOREINFO);
    if (!PSLGetStoreInfo(hStore, &si)) {
        error("PSLGetStoreInfo");
        return false;
    }
    return true;
}
bool GetPartitionInfo(HANDLE hPartition, PARTINFO& pi)
{
    memset(&pi, 0, sizeof(PARTINFO));
    pi.cbSize= sizeof(PARTINFO);
    if (!PSLGetPartitionInfo(hPartition, &pi)) {
        error("PSLGetPartitionInfo");
        return false;
    }
    return true;
}

ITSUTILS_API HRESULT STDAPICALLTYPE ITGetPartitionList(
        DWORD cbInput, GetPartitionListParams *pbInput,
        DWORD *pcbOutput, GetPartitionListResult **ppbOutput,
        IRAPIStream *pStream)
{
    *pcbOutput= 0;
    GetPartitionListResult *pOut= *ppbOutput= NULL;

    std::vector<PARTINFO> list;

    HANDLE hStore = pbInput->hStore ? (HANDLE)pbInput->hStore : OpenStore(pbInput->szDeviceName);
    if (hStore == INVALID_HANDLE_VALUE) {
        error("GetStorePartitionlist: OpenStore('%ls')", pbInput->szDeviceName);
        return false;
    }
 
    STOREINFO si;
    if (!GetStoreInfo(hStore, si))
        return GetLastError();

    if (!GetStorePartitionlist(hStore, list))
        return GetLastError();

    if (pbInput->hStore==NULL)
        CloseHandle(hStore);

    *pcbOutput= PTR_DIFF(pOut, &pOut->partInfo[list.size()]);
    pOut= *ppbOutput= (GetPartitionListResult*)LocalAlloc(LPTR, *pcbOutput);
    if (pOut==NULL) {
        error("LocalAlloc(%d)", *pcbOutput);
        return GetLastError();
    }

    pOut->nPartitions = list.size();

    for (size_t i=0 ; i<list.size() ; i++) {
        cvPARTINFO2DevPartitioninfo(si, list[i], pOut->partInfo[i]);
    }

    return 0;
}


#define BDK_SIGNATURE "BIPO"

// if partition==""  ->  open device as file.
bool OpenTFFSDisk(const WCHAR *szDeviceName, const WCHAR *szPartitionName, HANDLE& hStore, HANDLE& hPartition)
{

    if (szPartitionName && wcslen(szPartitionName)) {
        hStore = OpenStore(szDeviceName);
        if (hStore == INVALID_HANDLE_VALUE || hStore == NULL) {
            error("OpenTFFSDisk: OpenStore('%ls')", szDeviceName);
            return false;
        }
        hPartition = PSLOpenPartition(hStore, szPartitionName);
        if (hPartition == INVALID_HANDLE_VALUE || hPartition == NULL) {
            error("PSLOpenPartition('%ls')", szPartitionName);
            CloseHandle(hStore);
            hStore= NULL;
            return false;
        }
    }
    else {
        hStore= NULL;
        hPartition= CreateFile( szDeviceName,
            GENERIC_READ, FILE_SHARE_READ|FILE_SHARE_WRITE,
            NULL, OPEN_EXISTING, 0, 0 );
    }

    return true;
}
bool CloseTFFSDisk(HANDLE hStore, HANDLE hPartition)
{
    CloseHandle(hPartition);
    CloseHandle(hStore);
    return true;
}
bool TFFS_BDK_GetInfo(HANDLE hDisk, DWORD dwBinaryPartitionNr, DWORD &dwLength, DWORD &dwCount)
{
    flBDKOperationInput in;  memset(&in, 0, sizeof(in));

    in.partitionNumber = (BYTE)dwBinaryPartitionNr;
    in.type = BDK_GET_INFO;
    memcpy(in.bdkStruct.oldSign, BDK_SIGNATURE, strlen(BDK_SIGNATURE));
    in.bdkStruct.signOffset = 8;
    in.bdkStruct.startingBlock = 0;

    flOutputStatusRecord out; memset(&out, 0, sizeof(out));

    DWORD nReturned;
    if (!DeviceIoControl(hDisk, FL_IOCTL_BDK_OPERATION, &in, sizeof(in), &out, sizeof(out), &nReturned, NULL)) {
        error("DeviceIoControl(FL_IOCTL_BDK_OPERATION, BDK_GET_INFO)");
        return false;
    }
    if (out.status != flOK) {
        error("DeviceIoControl(FL_IOCTL_BDK_OPERATION, BDK_GET_INFO) - status=%08lx", out.status);
        SetLastError(ERROR_INTERNAL_ERROR);
        return false;
    }
    //  used length:  in.bdkStruct.length
    dwLength= in.bdkStruct.startingBlock;
    dwCount= in.bdkStruct.flags;

    return true;
}
bool TFFS_OTP_GetSize(HANDLE hDisk, DWORD& dwSize, DWORD &dwMaxSize)
{
    flOtpInput in;  memset(&in, 0, sizeof(in));
    flOutputStatusRecord out; memset(&out, 0, sizeof(out));

    in.type= OTP_SIZE;

    DWORD nReturned;
    if (!DeviceIoControl(hDisk, FL_IOCTL_OTP, &in, sizeof(in), &out, sizeof(out), &nReturned, NULL)) {
        error("DeviceIoControl(FL_IOCTL_OTP, OTP_SIZE)");
        return false;
    }
    if (out.status != flOK) {
        error("DeviceIoControl(FL_IOCTL_OTP, OTP_SIZE) - status=%08lx", out.status);
        SetLastError(ERROR_INTERNAL_ERROR);
        return false;
    }

    // in.lockedFlag

    dwSize= in.usedSize;
    dwMaxSize= in.length;
    return true;
}
bool TFFS_GetCustomerId(HANDLE hDisk, DWORD &id)
{
    flCustomerIdOutput out; memset(&out, 0, sizeof(out));

    DWORD nReturned;
    if (!DeviceIoControl(hDisk, FL_IOCTL_CUSTOMER_ID, NULL, 0, &out, sizeof(out), &nReturned, NULL)) {
        error("DeviceIoControl(FL_IOCTL_CUSTOMER_ID)");
        return false;
    }
    if (out.status != flOK) {
        error("DeviceIoControl(FL_IOCTL_CUSTOMER_ID) - status=%08lx", out.status);
        SetLastError(ERROR_INTERNAL_ERROR);
        return false;
    }
    memcpy(&id, out.id, 4);
    return true;
}
std::string GuidAsString(GUID& id)
{
    return stringformat("{%08lx-%04x-%04x-%02x%02x-%02x%02x%02x%02x%02x%02x}", 
            id.Data1, id.Data2, id.Data3, id.Data4[0], id.Data4[1],
            id.Data4[2], id.Data4[3], id.Data4[4], id.Data4[5], id.Data4[6], id.Data4[7]);
}
bool TFFS_GetUniqueId(HANDLE hDisk, GUID& id)
{
    flUniqueIdOutput out; memset(&out, 0, sizeof(out));

    DWORD nReturned;
    if (!DeviceIoControl(hDisk, FL_IOCTL_UNIQUE_ID, NULL, 0, &out, sizeof(out), &nReturned, NULL)) {
        error("DeviceIoControl(FL_IOCTL_UNIQUE_ID)");
        return false;
    }
    if (out.status != flOK) {
        error("DeviceIoControl(FL_IOCTL_UNIQUE_ID) - status=%08lx", out.status);
        SetLastError(ERROR_INTERNAL_ERROR);
        return false;
    }
    memcpy(&id, out.id, 16);
    return true;
}
bool TFFS_GetNrOfPartitions(HANDLE hDisk, DWORD& dwPartitionCount)
{
    flCountPartitionsOutput out; memset(&out, 0, sizeof(out));

    DWORD nReturned;
    if (!DeviceIoControl(hDisk, FL_IOCTL_NUMBER_OF_PARTITIONS, NULL, 0, &out, sizeof(out), &nReturned, NULL)) {
        error("DeviceIoControl(FL_IOCTL_NUMBER_OF_PARTITIONS)");
        return false;
    }
    if (out.status != flOK) {
        error("DeviceIoControl(FL_IOCTL_NUMBER_OF_PARTITIONS) - status=%08lx", out.status);
        SetLastError(ERROR_INTERNAL_ERROR);
        return false;
    }
    dwPartitionCount= out.noOfPartitions;
    return true;
}

bool TFFS_BDK_InitRead(HANDLE hDisk, DWORD dwBinaryPartitionNr, ULONGLONG llOffset, DWORD dwSize)
{
    flBDKOperationInput in;  memset(&in, 0, sizeof(in));

    in.partitionNumber = (BYTE)dwBinaryPartitionNr;
    in.type = BDK_INIT_READ;
    memcpy(in.bdkStruct.oldSign, BDK_SIGNATURE, strlen(BDK_SIGNATURE));
    in.bdkStruct.signOffset = 8;
    in.bdkStruct.startingBlock = (DWORD)(llOffset/TFFS_BLOCK_SIZE);
    in.bdkStruct.length = dwSize;

    flOutputStatusRecord out; memset(&out, 0, sizeof(out));

    DWORD nReturned;
    if (!DeviceIoControl(hDisk, FL_IOCTL_BDK_OPERATION, &in, sizeof(in), &out, sizeof(out), &nReturned, NULL)) {
        error("DeviceIoControl(FL_IOCTL_BDK_OPERATION, BDK_INIT_READ)");
        return false;
    }
    if (out.status != flOK) {
        error("DeviceIoControl(FL_IOCTL_BDK_OPERATION, BDK_INIT_READ) - status=%08lx", out.status);
        SetLastError(ERROR_INTERNAL_ERROR);
        return false;
    }
    return true;
}

bool TFFS_BDK_ReadNextSector(HANDLE hDisk, DWORD dwBinaryPartitionNr, ULONGLONG llOffset, DWORD dwLength, BYTE *buffer, DWORD &dwRead, DWORD dwBlockSize)
{
    flBDKOperationInput in;  memset(&in, 0, sizeof(in));
    BYTE *localbuf= (BYTE*)alloca(dwLength);

    memset(localbuf, 0, dwLength);

    in.partitionNumber = (BYTE)dwBinaryPartitionNr;
    in.type = BDK_READ;
    memcpy(in.bdkStruct.oldSign, BDK_SIGNATURE, strlen(BDK_SIGNATURE));
    in.bdkStruct.signOffset = 8;
    in.bdkStruct.startingBlock = (DWORD)(llOffset/dwBlockSize);
    in.bdkStruct.length = dwLength;
    in.bdkStruct.bdkBuffer = localbuf;    // !!! this must be allocated on the stack.

    flOutputStatusRecord out; memset(&out, 0, sizeof(out));

    debug("disk_bdk_readnext: reading %I64x %08lx  ss=%08lx\n", llOffset, dwLength, dwBlockSize);

    DWORD nReturned;
    if (!DeviceIoControl(hDisk, FL_IOCTL_BDK_OPERATION, &in, sizeof(in), &out, sizeof(out), &nReturned, NULL)) {
        error("DeviceIoControl(FL_IOCTL_BDK_OPERATION, BDK_READ)");
        return false;
    }
    if (out.status != flOK) {
        error("DeviceIoControl(FL_IOCTL_BDK_OPERATION, BDK_READ) - status=%08lx", out.status);
        SetLastError(ERROR_INTERNAL_ERROR);
        return false;
    }

    memcpy(buffer, localbuf, dwLength);

    dwRead= dwLength;

    debug("disk_bdk_readnext: read %08lx\n", dwRead);

    return true;
}

bool TFFS_OTP_Read(HANDLE hDisk, ULONGLONG llOffset, DWORD dwLength, BYTE *buffer, DWORD &dwRead)
{
    flOtpInput in;  memset(&in, 0, sizeof(in));
    flOutputStatusRecord out; memset(&out, 0, sizeof(out));

    BYTE *localbuf= (BYTE*)alloca(dwLength);
    memset(localbuf, 0, dwLength);

    in.type   = OTP_READ;
    in.usedSize= 0;
    in.length = dwLength;
    in.buffer= localbuf;

    debug("tffs_otp_read: reading %I64x %08lx\n", llOffset, dwLength);

    DWORD nReturned;
    if (!DeviceIoControl(hDisk, FL_IOCTL_OTP, &in, sizeof(in), &out, sizeof(out), &nReturned, NULL)) {
        error("DeviceIoControl(FL_IOCTL_OTP, OTP_READ)");
        return false;
    }
    if (out.status != flOK) {
        error("DeviceIoControl(FL_IOCTL_OTP, OTP_READ) - status=%08lx", out.status);
        SetLastError(ERROR_INTERNAL_ERROR);
        return false;
    }

    memcpy(buffer, localbuf, in.length);
    dwRead= in.length;

    debug("tffs_otp_read: read %08lx\n", dwRead);
    return true;
}

//..........
bool TFFS_ReadSectors(HANDLE hDisk, ULONGLONG llOffset, DWORD dwLength, BYTE *buffer, DWORD &dwBytesRead, DWORD dwSectorSize)
{
    flReadWriteInput in;  memset(&in, 0, sizeof(in));
    BYTE *localbuf= (BYTE*)alloca(dwLength);

    memset(localbuf, 0, dwLength);

    if (dwLength > dwLength)
        dwLength= dwLength;

    in.firstSector= (DWORD)(llOffset/dwSectorSize);
    in.numberOfSectors= dwLength/dwSectorSize;
    in.buf= localbuf;

    flReadWriteOutput out; memset(&out, 0, sizeof(out));

    DWORD nReturned;
    if (!DeviceIoControl(hDisk, FL_IOCTL_READ_SECTORS, &in, sizeof(in), &out, sizeof(out), &nReturned, NULL)) {
        error("DeviceIoControl(FL_IOCTL_READ_SECTORS)");
        return false;
    }
    if (out.status != flOK) {
        error("DeviceIoControl(FL_IOCTL_READ_SECTORS) - status=%08lx", out.status);
        SetLastError(ERROR_INTERNAL_ERROR);
        return false;
    }
    dwBytesRead= out.numberOfSectors*dwSectorSize;
    if (dwBytesRead > dwLength)
        dwBytesRead= dwLength;

    memcpy(buffer, localbuf, dwBytesRead);

    return true;
}


//...............................
bool Disk_ReadSectors(HANDLE hDisk, ULONGLONG llOffset, DWORD dwLength, BYTE *buffer, DWORD &dwBytesRead, DWORD dwSectorSize)
{
    SG_REQ req;

    req.sr_start= (DWORD)(llOffset/dwSectorSize);
    req.sr_num_sec= dwLength/dwSectorSize;
    req.sr_num_sg= 1;
    req.sr_status= 0;
    req.sr_callback= NULL;
    req.sr_sglist[0].sb_len= dwLength;
    req.sr_sglist[0].sb_buf= buffer;

    debug("disk_readsectors: reading %I64x %08lx  ss=%08lx\n", llOffset, dwLength, dwSectorSize);

    DWORD nReturned=0;
    if (!DeviceIoControl(hDisk, DISK_IOCTL_READ, &req, sizeof(req), NULL, NULL, &nReturned, NULL)) {
        error("DISK_IOCTL_READ");
        return false;
    }

    dwBytesRead= req.sr_sglist[0].sb_len;

    debug("disk_readsectors: read %08lx\n", dwBytesRead);
    return true;
}
bool Disk_GetDiskSize(HANDLE hDisk, DWORD dwBinaryPartitionNr, ULONGLONG &llDiskSize)
{
    switch (dwBinaryPartitionNr) {
    case BP_TFFSREADSECTOR:
    case BP_WINCEREADSECTOR: 
    {
        DISK_INFO di;
        if (!GetDiskInfo(hDisk, di)) {
            error("GetDiskInfo");
            llDiskSize= 0;
            return false;
        }
        llDiskSize= UInt32x32To64(di.di_total_sectors, di.di_bytes_per_sect);
        return true;
    }

    case BP_OTPREADSECTOR: 
    {
        DWORD dwOtpSize, dwMaxOtpSize;
        if (!TFFS_OTP_GetSize(hDisk, dwOtpSize, dwMaxOtpSize)) {
            error("TFFS_OTP_GetSize");
            llDiskSize= 0;
            return false;
        }
        llDiskSize= dwOtpSize;  // returning current size, .. or should i return max size here?
        return true;
    }

    default: 
    {
        DWORD dwLength, dwNrBParts;
        if (!TFFS_BDK_GetInfo(hDisk, dwBinaryPartitionNr, dwLength, dwNrBParts)) {
            error("TFFS_BDK_GetInfo(%d)", dwBinaryPartitionNr);
            llDiskSize= 0;
            return false;
        }
        llDiskSize= dwLength;
        return true;
    }
    }
}
bool Disk_GetDiskParams(HANDLE hDisk, DWORD dwBinaryPartitionNr, DWORD &dwBlockSize, DWORD &dwSectorSize)
{
    switch (dwBinaryPartitionNr) {
    case BP_TFFSREADSECTOR:
        dwBlockSize= TFFS_BLOCK_SIZE;
        dwSectorSize= TFFS_SECTOR_SIZE;
        return true;

    case BP_WINCEREADSECTOR:
        // todo: read this from GetStoreInfo.
        dwBlockSize= TFFS_BLOCK_SIZE;
        dwSectorSize= TFFS_SECTOR_SIZE;
        return true;

    case BP_OTPREADSECTOR:
        dwBlockSize= TFFS_BLOCK_SIZE;
        dwSectorSize= TFFS_SECTOR_SIZE;
        return true;

    default:
        dwBlockSize= TFFS_BLOCK_SIZE;
        dwSectorSize= TFFS_BLOCK_SIZE;
        return true;
    }
}
bool ReadDisk_Init(HANDLE hDisk, DWORD dwBinaryPartitionNr, ULONGLONG llOffset, DWORD dwLength)
{
    switch (dwBinaryPartitionNr) {
    case BP_TFFSREADSECTOR:
        return true;

    case BP_WINCEREADSECTOR:
        return true;

    case BP_OTPREADSECTOR:
        return true;

    default:
        return TFFS_BDK_InitRead(hDisk, dwBinaryPartitionNr, llOffset, dwLength);
    }
}
bool ReadDisk(HANDLE hDisk, DWORD dwBinaryPartitionNr, ULONGLONG llOffset, DWORD dwLength, BYTE *buffer, DWORD &dwRead, DWORD dwSectorSize)
{
    debug("readdisk: %d : %I64x %08lx  ss=%08lx\n", dwBinaryPartitionNr, llOffset, dwLength, dwSectorSize);
    switch(dwBinaryPartitionNr) {
    case BP_TFFSREADSECTOR:
        return TFFS_ReadSectors(hDisk, llOffset, dwLength, buffer, dwRead, dwSectorSize);
    case BP_WINCEREADSECTOR:
        return Disk_ReadSectors(hDisk, llOffset, dwLength, buffer, dwRead, dwSectorSize);
    case BP_OTPREADSECTOR:
        return TFFS_OTP_Read(hDisk, llOffset, dwLength, buffer, dwRead);
    default:
        // passing blocksize in sectorsize here!
        return TFFS_BDK_ReadNextSector(hDisk, dwBinaryPartitionNr, llOffset, dwLength, buffer, dwRead, dwSectorSize);
    }
}
ITSUTILS_API HRESULT STDAPICALLTYPE ITReadDisk(
        DWORD cbInput, ReadDiskParams *pbInput,
        DWORD *pcbOutput, ReadDiskResult **ppbOutput,
        IRAPIStream *pStream)
{
    *pcbOutput= 0;
    ReadDiskResult *pOut= *ppbOutput= NULL;

    HANDLE hStore, hPartition;
    if (pbInput->hDisk) {
        hPartition= (HANDLE)pbInput->hDisk;
        hStore= NULL;
    }
    else if (!OpenTFFSDisk(pbInput->szDeviceName, pbInput->szPartitionName, hStore, hPartition))
        return GetLastError();

    BOOL bMode = SetKMode(TRUE);
    DWORD dwPerm = SetProcPermissions(0xFFFFFFFF);

    DWORD dwBlockSize, dwSectorSize;
    if (!Disk_GetDiskParams(hPartition, pbInput->dwBinaryPartitionNr, dwBlockSize, dwSectorSize)) {
        if (pbInput->hDisk==NULL) CloseTFFSDisk(hStore, hPartition);
        return GetLastError();
    }

    if (!ReadDisk_Init(hPartition, pbInput->dwBinaryPartitionNr, pbInput->llOffset, pbInput->dwSize)) {
        if (pbInput->hDisk==NULL) CloseTFFSDisk(hStore, hPartition);
        return GetLastError();
    }

    *pcbOutput= PTR_DIFF(pOut, &pOut->buffer[pbInput->dwSize]);
    pOut= *ppbOutput= (ReadDiskResult*)LocalAlloc(LPTR, *pcbOutput);
    if (pOut==NULL) {
        error("LocalAlloc(%d)", *pcbOutput);
        if (pbInput->hDisk==NULL) CloseTFFSDisk(hStore, hPartition);
        return GetLastError();
    }

    DWORD dwRes= 0;

    debug("reading %I64x : %08lx bytes bs=%08lx ss=%08lx\n", pbInput->llOffset, pbInput->dwSize,  dwBlockSize, dwSectorSize);
    DWORD dwRemainingLength= pbInput->dwSize;
    DWORD dwBufOffset= 0;
    ULONGLONG llDiskOffset= pbInput->llOffset;
    while (dwRemainingLength)
    {
        // make sure first wanted makes diskoffset allign to blocksize
        DWORD dwWanted= dwBlockSize - ((DWORD)llDiskOffset&(dwBlockSize-1));
        if (dwWanted > dwRemainingLength) 
            dwWanted= dwRemainingLength;

        debug("w=%08lx b=%08lx o=%I64x r=%08lx\n", dwWanted, dwBlockSize, llDiskOffset, dwRemainingLength);

        DWORD dwRead;
        if (!ReadDisk(hPartition, pbInput->dwBinaryPartitionNr, llDiskOffset, dwWanted, pOut->buffer+dwBufOffset, dwRead, dwSectorSize))
        {
            dwRes= GetLastError();
            if (dwRes==0)
                dwRes= ERROR_INTERNAL_ERROR;
            break;
        }

        if (dwRead==0)
            break;

        dwBufOffset += dwRead;
        llDiskOffset += dwRead; // ?? maybe don't add dwRead for BDK read?
        dwRemainingLength -= dwRead;
    }

    SetProcPermissions(dwPerm);
    SetKMode(bMode);

    if (pbInput->hDisk==NULL) CloseTFFSDisk(hStore, hPartition);

    pOut->dwNumberOfBytesRead= dwBufOffset;

    return dwRes;
}


//.......................................................
bool TFFS_GetInfo(HANDLE hDisk)
{
    flDiskInfoOutput out;

    DWORD nReturned;
    if (!DeviceIoControl(hDisk, FL_IOCTL_GET_INFO, NULL, 0, &out, sizeof(out), &nReturned, NULL)) {
        error("DeviceIoControl(FL_IOCTL_GET_INFO)");
        return false;
    }
    if (out.status != flOK) {
        error("DeviceIoControl(FL_IOCTL_GET_INFO) - status=%08lx", out.status);
        SetLastError(ERROR_INTERNAL_ERROR);
        return false;
    }

    debug("getinfo: logical=%08lx boot=%08lx base=%08lx ft=%04x phys=%08lx unitsize=%08lx\n",
        out.info.logicalSectors, out.info.bootAreaSize, out.info.baseAddress,
        out.info.flashType, out.info.physicalSize, out.info.physicalUnitSize);
    debug("getinfo: doctype=%02x lifetime=%02x  driver='%s'  osak='%s' cyl=%d hd=%d sc=%d\n",
        out.info.DOCType, out.info.lifeTime, out.info.driverVer, out.info.OSAKVer,
        out.info.cylinders, out.info.heads, out.info.sectors);

    return true;
}


std::string SI_DeviceClassString(DWORD dw)
{
    switch(dw)
    {
        case STORAGE_DEVICE_CLASS_BLOCK     : return "BLOCK";
        case STORAGE_DEVICE_CLASS_MULTIMEDIA: return "MULTIMEDIA";
        default: return stringformat("UNKNOWNDEVCLASS_%08lx", dw);
    }
}
std::string SI_DeviceTypeString(DWORD dw)
{
    if (dw==0) return "(none)";
    StringList list;
    if (dw&STORAGE_DEVICE_TYPE_PCIIDE         ) list.push_back("PCIIDE");
    if (dw&STORAGE_DEVICE_TYPE_FLASH          ) list.push_back("FLASH");
    if (dw&STORAGE_DEVICE_TYPE_ATA            ) list.push_back("ATA");
    if (dw&STORAGE_DEVICE_TYPE_ATAPI          ) list.push_back("ATAPI");
    if (dw&STORAGE_DEVICE_TYPE_PCCARD         ) list.push_back("PCCARD");
    if (dw&STORAGE_DEVICE_TYPE_CFCARD         ) list.push_back("CFCARD");
    if (dw&STORAGE_DEVICE_TYPE_SRAM           ) list.push_back("SRAM");
    if (dw&STORAGE_DEVICE_TYPE_DVD            ) list.push_back("DVD");
    if (dw&STORAGE_DEVICE_TYPE_CDROM          ) list.push_back("CDROM");
    if (dw&STORAGE_DEVICE_TYPE_USB            ) list.push_back("USB");
    if (dw&STORAGE_DEVICE_TYPE_1394           ) list.push_back("1394");
    if (dw&STORAGE_DEVICE_TYPE_DOC            ) list.push_back("DOC");
    if (dw&STORAGE_DEVICE_TYPE_UNKNOWN        ) list.push_back("UNKNOWN");
    if (dw&STORAGE_DEVICE_TYPE_REMOVABLE_DRIVE) list.push_back("REMOVABLE_DRIVE");
    if (dw&STORAGE_DEVICE_TYPE_REMOVABLE_MEDIA) list.push_back("REMOVABLE_MEDIA");

    const DWORD knownflags= STORAGE_DEVICE_TYPE_PCIIDE|STORAGE_DEVICE_TYPE_FLASH|STORAGE_DEVICE_TYPE_ATA|STORAGE_DEVICE_TYPE_ATAPI|STORAGE_DEVICE_TYPE_PCCARD|STORAGE_DEVICE_TYPE_CFCARD|STORAGE_DEVICE_TYPE_SRAM|STORAGE_DEVICE_TYPE_DVD|STORAGE_DEVICE_TYPE_CDROM|STORAGE_DEVICE_TYPE_USB|STORAGE_DEVICE_TYPE_1394|STORAGE_DEVICE_TYPE_DOC|STORAGE_DEVICE_TYPE_UNKNOWN|STORAGE_DEVICE_TYPE_REMOVABLE_DRIVE|STORAGE_DEVICE_TYPE_REMOVABLE_MEDIA;
    if (dw&~knownflags) list.push_back(stringformat("UNKNOWNDEVTYPE_%08lx", dw&~knownflags));

    return JoinStringList(list, ",");
}

std::string SI_DeviceFlagsString(DWORD dw)
{
    if (dw==0) return "(none)";
    StringList list;
    if (dw&STORAGE_DEVICE_FLAG_READWRITE ) list.push_back("READWRITE");
    if (dw&STORAGE_DEVICE_FLAG_READONLY  ) list.push_back("READONLY");
    if (dw&STORAGE_DEVICE_FLAG_TRANSACTED) list.push_back("TRANSACTED");
    if (dw&STORAGE_DEVICE_FLAG_MEDIASENSE) list.push_back("MEDIASENSE");

    const DWORD knownflags= STORAGE_DEVICE_FLAG_READWRITE|STORAGE_DEVICE_FLAG_READONLY|STORAGE_DEVICE_FLAG_TRANSACTED|STORAGE_DEVICE_FLAG_MEDIASENSE;
    if (dw&~knownflags) list.push_back(stringformat("UNKNOWNDEVFLAGS_%08lx", dw&~knownflags));

    return JoinStringList(list, ",");
}
std::string GetStorageDeviceInfoString(const STORAGEDEVICEINFO& sdi)
{
    if (sdi.cbSize!=sizeof(STORAGEDEVICEINFO))
        debug("WARNING: sdi.cbSize=%d (!= %d)\n", sdi.cbSize, sizeof(STORAGEDEVICEINFO));
    return stringformat("profile='%ls' class=%s type=%s flags=%s", 
            sdi.szProfile,
            SI_DeviceClassString(sdi.dwDeviceClass).c_str(),
            SI_DeviceTypeString(sdi.dwDeviceType).c_str(),
            SI_DeviceFlagsString(sdi.dwDeviceFlags).c_str());
}
std::string SectorNumString(const SECTORNUM& sn)
{
    return stringformat("%I64x", sn);
}
std::string FileTimeString(const FILETIME& ft)
{
    if (ft.dwLowDateTime==0 && ft.dwHighDateTime==0)
        return "(null)";

    SYSTEMTIME systime;
    FILETIME lfiletime;
    FileTimeToLocalFileTime(&ft, &lfiletime);
    FileTimeToSystemTime(&lfiletime, &systime);
    return stringformat("%04d-%02d-%02d %02d:%02d:%02d.%03d",
            systime.wYear, systime.wMonth, systime.wDay,
            systime.wHour, systime.wMinute, systime.wSecond,systime.wMilliseconds);
}
std::string SI_AttributesString(DWORD dw)
{
    if (dw==0) return "(none)";
    StringList list;
    if (dw&STORE_ATTRIBUTE_READONLY   ) list.push_back("READONLY");
    if (dw&STORE_ATTRIBUTE_REMOVABLE  ) list.push_back("REMOVABLE");
    if (dw&STORE_ATTRIBUTE_UNFORMATTED) list.push_back("UNFORMATTED");
    if (dw&STORE_ATTRIBUTE_AUTOFORMAT ) list.push_back("AUTOFORMAT");
    if (dw&STORE_ATTRIBUTE_AUTOPART   ) list.push_back("AUTOPART");
    if (dw&STORE_ATTRIBUTE_AUTOMOUNT  ) list.push_back("AUTOMOUNT");

    const DWORD knownflags= STORE_ATTRIBUTE_READONLY|STORE_ATTRIBUTE_REMOVABLE|STORE_ATTRIBUTE_UNFORMATTED|STORE_ATTRIBUTE_AUTOFORMAT|STORE_ATTRIBUTE_AUTOPART|STORE_ATTRIBUTE_AUTOMOUNT;
    if (dw&~knownflags) list.push_back(stringformat("UNKNOWNSATTRS_%08lx", dw&~knownflags));

    return JoinStringList(list, ",");
}
std::string PI_AttributesString(DWORD dw)
{
    if (dw==0) return "(none)";
    StringList list;
    if (dw&PARTITION_ATTRIBUTE_EXPENDABLE) list.push_back("EXPENDABLE");
    if (dw&PARTITION_ATTRIBUTE_READONLY  ) list.push_back("READONLY");
    if (dw&PARTITION_ATTRIBUTE_AUTOFORMAT) list.push_back("AUTOFORMAT");
    if (dw&PARTITION_ATTRIBUTE_ACTIVE    ) list.push_back("ACTIVE");
    if (dw&PARTITION_ATTRIBUTE_BOOT      ) list.push_back("BOOT");
    if (dw&PARTITION_ATTRIBUTE_MOUNTED   ) list.push_back("MOUNTED");

    const DWORD knownflags= PARTITION_ATTRIBUTE_EXPENDABLE|PARTITION_ATTRIBUTE_READONLY|PARTITION_ATTRIBUTE_AUTOFORMAT|PARTITION_ATTRIBUTE_ACTIVE|PARTITION_ATTRIBUTE_BOOT|PARTITION_ATTRIBUTE_MOUNTED;
    if (dw&~knownflags) list.push_back(stringformat("UNKNOWNPATTRS_%08lx", dw&~knownflags));

    return JoinStringList(list, ",");
}
std::string StoreInfoString(const STOREINFO& si)
{
    std::string str;
    if (si.cbSize!=sizeof(STOREINFO))
        debug("WARNING: si.cbSize=%d (!= %d)\n", si.cbSize, sizeof(STOREINFO));
    str += stringformat("dev='%ls'  store='%ls' ", si.szDeviceName, si.szStoreName);
    str += stringformat("  nsect=%s bpsect=%x free=%s maxpartsize=%s",
            SectorNumString(si.snNumSectors).c_str(),
            si.dwBytesPerSector,
            SectorNumString(si.snFreeSectors).c_str(),
            SectorNumString(si.snBiggestPartCreatable).c_str());
    str += stringformat("  created=%s  modified=%s",
            FileTimeString(si.ftCreated).c_str(),
            FileTimeString(si.ftLastModified).c_str());
    str += stringformat("  attribs=%s  partitions=%d  mounts=%d\n",
            SI_AttributesString(si.dwAttributes).c_str(),
            si.dwPartitionCount,
            si.dwMountCount);

    str += stringformat("  si=%08lx  -> class=%08lx type=%08lx\n", &si, si.dwDeviceClass, si.dwDeviceType);
    str += stringformat("  class=%s type=%s flags=%s\n",
            SI_DeviceClassString(si.dwDeviceClass).c_str(),
            SI_DeviceTypeString(si.dwDeviceType).c_str(),
            SI_DeviceFlagsString(si.dwDeviceFlags).c_str());
    str += stringformat("  sdi: %s\n", GetStorageDeviceInfoString(si.sdi).c_str());

    return str;
}
std::string PartitionInfoString(const PARTINFO& pi)
{
    std::string str;
    if (pi.cbSize!=sizeof(PARTINFO))
        debug("WARNING: pi.cbSize=%d (!= %d)\n", pi.cbSize, sizeof(PARTINFO));
    str += stringformat("name='%ls'  filesys='%ls' volname='%ls'",
        pi.szPartitionName, pi.szFileSys, pi.szVolumeName);
    str += stringformat("  end=%s  created=%s modified=%s",
            SectorNumString(pi.snNumSectors).c_str(),
            FileTimeString(pi.ftCreated).c_str(),
            FileTimeString(pi.ftLastModified).c_str());
    str += stringformat("  attr=%s type=%02x\n",
            PI_AttributesString(pi.dwAttributes).c_str(),
            pi.bPartType);

    return str;
}
bool GetStorageDeviceInfo(HANDLE hDisk, STORAGEDEVICEINFO &sdi)
{
    DWORD dwReturned;
    if (DeviceIoControl(hDisk, IOCTL_DISK_DEVICE_INFO, &sdi, sizeof(STORAGEDEVICEINFO), NULL, 0, &dwReturned, NULL))
        return true;
    error("IOCTL_DISK_DEVICE_INFO(in)");
    if (DeviceIoControl(hDisk, IOCTL_DISK_DEVICE_INFO, NULL, 0, &sdi, sizeof(STORAGEDEVICEINFO), &dwReturned, NULL))
        return true;
    error("IOCTL_DISK_DEVICE_INFO(out)");
    return false;
}
std::string DiskInfoFlagsString(DWORD dw)
{
    if (dw==0) return "(none)";
    StringList list;
    if (dw&DISK_INFO_FLAG_MBR           ) list.push_back("MBR");
    if (dw&DISK_INFO_FLAG_CHS_UNCERTAIN ) list.push_back("CHS_UNCERTAIN");
    if (dw&DISK_INFO_FLAG_UNFORMATTED   ) list.push_back("UNFORMATTED");
    if (dw&DISK_INFO_FLAG_PAGEABLE      ) list.push_back("PAGEABLE");

    const DWORD knownflags= DISK_INFO_FLAG_MBR|DISK_INFO_FLAG_CHS_UNCERTAIN|DISK_INFO_FLAG_UNFORMATTED|DISK_INFO_FLAG_PAGEABLE;
    if (dw&~knownflags) list.push_back(stringformat("UNKNOWNFLAGS_%08lx", dw&~knownflags));

    return JoinStringList(list, ",");
}
std::string DiskInfoString(const DISK_INFO &di)
{
    return stringformat("tot=0x%08lx bps=0x%x cyls=0x%x hds=0x%x sectpcyl=0x%x flags=%hs",
            di.di_total_sectors, di.di_bytes_per_sect,
            di.di_cylinders, di.di_heads, di.di_sectors,
            DiskInfoFlagsString(di.di_flags).c_str());
}
bool GetDiskStorageId(HANDLE hDisk, ByteVector &mfid, ByteVector &serialnr)
{
    ByteVector data;
    STORAGE_IDENTIFICATION *sid= NULL;

    data.resize(sizeof(STORAGE_IDENTIFICATION)+256);
    sid= (STORAGE_IDENTIFICATION*)vectorptr(data);
    DWORD dwReturned;
    while (!DeviceIoControl(hDisk, IOCTL_DISK_GET_STORAGEID, NULL, 0, sid, data.size(), &dwReturned, NULL)) {
        if (GetLastError()!=ERROR_INSUFFICIENT_BUFFER) {
            error("IOCTL_DISK_GET_STORAGEID(out)");
            return false;
        }
        data.resize(data.size()*2);
        sid= (STORAGE_IDENTIFICATION*)vectorptr(data);
    }
    if ((sid->dwFlags&MANUFACTUREID_INVALID)==0)
    {
        mfid.resize(sid->dwManufactureIDOffset < sid->dwSerialNumOffset 
                ? sid->dwSerialNumOffset-sid->dwManufactureIDOffset 
                : sid->dwSize-sid->dwManufactureIDOffset );
        memcpy(vectorptr(mfid), &data[sid->dwManufactureIDOffset], mfid.size());
    }
    if ((sid->dwFlags&SERIALNUM_INVALID)==0)
    {
        serialnr.resize(sid->dwSerialNumOffset < sid->dwManufactureIDOffset 
                ? sid->dwManufactureIDOffset-sid->dwSerialNumOffset
                : sid->dwSize-sid->dwSerialNumOffset);
        memcpy(vectorptr(serialnr), &data[sid->dwSerialNumOffset], serialnr.size());
    }
    const dwKnownFlags= MANUFACTUREID_INVALID|SERIALNUM_INVALID;
    if (sid->dwFlags&~dwKnownFlags)
        debug("WARNING: STORAGE_IDENTIFICATION: unknown flags: %08lx\n", sid->dwFlags&~dwKnownFlags);

    return true;
}

bool GetDiskName(HANDLE hDisk, std::string &name)
{
    WCHAR namebuf[MAX_PATH];
    DWORD dwReturned;
    if (!DeviceIoControl(hDisk, IOCTL_DISK_GETNAME, NULL, 0, namebuf, sizeof(namebuf), &dwReturned, NULL)) {
        error("IOCTL_DISK_GETNAME(out)");
        return false;
    }

    name= ToString(namebuf);
    return true;
}

ITSUTILS_API HRESULT STDAPICALLTYPE ITTFFSGetInfo(
        DWORD cbInput, TFFSGetInfoParams *pbInput,
        DWORD *pcbOutput, TFFSGetInfoResult **ppbOutput,
        IRAPIStream *pStream)
{
    *pcbOutput= 0;
    TFFSGetInfoResult *pOut= *ppbOutput= NULL;

    HANDLE hStore, hPartition;
    if (pbInput->hDisk) {
        hPartition= (HANDLE)pbInput->hDisk;
        hStore= NULL;
    }
    else if (!OpenTFFSDisk(pbInput->szDeviceName, pbInput->szPartitionName, hStore, hPartition))
        return GetLastError();

    BOOL bMode = SetKMode(TRUE);
    DWORD dwPerm = SetProcPermissions(0xFFFFFFFF);

    DWORD dwBlockSize, dwSectorSize;
    if (!Disk_GetDiskParams(hPartition, pbInput->dwBinaryPartitionNr, dwBlockSize, dwSectorSize)) {
        if (pbInput->hDisk==NULL) CloseTFFSDisk(hStore, hPartition);
        return GetLastError();
    }

    debug("diskparams: bs=%08lx ss=%08lx\n", dwBlockSize, dwSectorSize);

    ULONGLONG llDiskSize;
    if (!Disk_GetDiskSize(hPartition, pbInput->dwBinaryPartitionNr, llDiskSize)) {
        if (pbInput->hDisk==NULL) CloseTFFSDisk(hStore, hPartition);
        return GetLastError();
    }

    debug("disksize: %I64x\n", llDiskSize);

    SetProcPermissions(dwPerm);
    SetKMode(bMode);

    if (pbInput->hDisk==NULL) CloseTFFSDisk(hStore, hPartition);

    *pcbOutput= sizeof(TFFSGetInfoResult);
    pOut= *ppbOutput= (TFFSGetInfoResult*)LocalAlloc(LPTR, *pcbOutput);
    if (pOut==NULL) {
        error("LocalAlloc(%d)", *pcbOutput);
        return GetLastError();
    }

    pOut->dwSectorSize= dwSectorSize;
    pOut->llDiskSize= llDiskSize;

    return 0;

}
ITSUTILS_API HRESULT STDAPICALLTYPE ITLogDiskInfo(
        DWORD cbInput, LogDiskInfoParams *pbInput,
        DWORD *pcbOutput, BYTE **ppbOutput,
        IRAPIStream *pStream)
{
    *pcbOutput= 0;
    *ppbOutput= NULL;

    HANDLE hStore, hPartition;
    if (pbInput->hDisk) {
        hPartition= (HANDLE)pbInput->hDisk;
        hStore= NULL;
        debug("listing diskinfo for existing handle %08lx\n", hPartition);
    }
    else if (!OpenTFFSDisk(pbInput->szDeviceName, pbInput->szPartitionName, hStore, hPartition))
        return GetLastError();
    else 
        debug("listing diskinfo for %ls/%ls\n", pbInput->szDeviceName, pbInput->szPartitionName);

    BOOL bMode = SetKMode(TRUE);
    DWORD dwPerm = SetProcPermissions(0xFFFFFFFF);

    TFFS_GetInfo(hPartition);

    DWORD dwNrParts;
    if (TFFS_GetNrOfPartitions(hPartition, dwNrParts))
        debug("tffs: %d partitions\n", dwNrParts);

    DWORD dwNrBParts=0;
    DWORD dwLength=0;
    if (TFFS_BDK_GetInfo(hPartition, 0, dwLength, dwNrBParts)) {
        debug("tffs: %d binary partitions\n", dwNrBParts);
        debug("bdk%d: %08lx\n", 0, dwLength);
        for (DWORD p=1 ; p<dwNrBParts ; p++) {
            if (TFFS_BDK_GetInfo(hPartition, p, dwLength, dwNrBParts))
                debug("bdk%d: %08lx\n", p, dwLength);
        }
    }

    DWORD dwOtpSize, dwMaxOtpSize;
    if (TFFS_OTP_GetSize(hPartition, dwOtpSize, dwMaxOtpSize))
        debug("otp size: %08lx  max: %08lx\n", dwOtpSize, dwMaxOtpSize);

    DWORD custid;
    if (TFFS_GetCustomerId(hPartition, custid))
        debug("tffs custid: %08lx\n", custid);

    GUID tffsid;
    if (TFFS_GetUniqueId(hPartition, tffsid))
        debug("tffs uniqueid: %hs\n", GuidAsString(tffsid).c_str());

    STOREINFO si;
    if (GetStoreInfo(hStore, si))
        debug("si: %hs\n",  StoreInfoString(si).c_str());

    PARTINFO pi;
    if (GetPartitionInfo(hPartition, pi))
        debug("pi: %hs\n",  PartitionInfoString(pi).c_str());

    STORAGEDEVICEINFO sdi;
    if (GetStorageDeviceInfo(hPartition, sdi))
        debug("%hs\n", GetStorageDeviceInfoString(sdi).c_str());

    DISK_INFO di;
    int ndi;
    if (ndi=GetDiskInfo(hPartition, di))
        debug("di: (%d) %hs\n", ndi, DiskInfoString(di).c_str());

    ByteVector mfid, serialnr;
    if (GetDiskStorageId(hPartition, mfid, serialnr)) {
        debug("mfid: %hs \n", hexdump(mfid).c_str());
        debug("serial: %hs \n", hexdump(serialnr).c_str());
    }

    std::string name;
    if (GetDiskName(hPartition, name))   // IOCTL_DISK_GETNAME
        debug("diskname: %hs\n", name.c_str());

    SetProcPermissions(dwPerm);
    SetKMode(bMode);

    if (pbInput->hDisk==NULL) CloseTFFSDisk(hStore, hPartition);

    return 0;
}


