#ifndef __TFFSREADER_H__
#define __TFFSREADER_H__

//========================================================================
// tffs bin part reading

#include "itsutils.h"

#define TFFS_BLOCK_SIZE 65536
#define TFFS_SECTOR_SIZE 512

#define DEVICENAMESIZE 8
#define PARTITIONNAMESIZE 32

// special values for binary partition number.
#define BP_TFFSREADSECTOR  0xffffffff
#define BP_WINCEREADSECTOR 0xfffffffe
#define BP_OTPREADSECTOR   0xfffffffd

typedef struct tagDevStoreInfo {
    WCHAR szDeviceName[DEVICENAMESIZE];
    ULONGLONG llStoreSize;
} DevStoreInfo;
typedef struct tagDevPartitionInfo {
    WCHAR szPartitionName[PARTITIONNAMESIZE];
    ULONGLONG llPartitionSize;
} DevPartitionInfo;

typedef struct tagGetStoreMgrListResult {
    int nStores;
    DevStoreInfo storeInfo[1];
} GetStoreMgrListResult;
ITSUTILS_API HRESULT STDAPICALLTYPE ITGetStoreMgrList(
        DWORD cbInput, void *pbInput,
        DWORD *pcbOutput, GetStoreMgrListResult**ppbOutput,
        IRAPIStream *pStream);

typedef struct tagGetPartitionListParams {
    WCHAR szDeviceName[DEVICENAMESIZE];
    DWORD hStore;
} GetPartitionListParams;
typedef struct tagGetPartitionListResult {
    int nPartitions;
    DevPartitionInfo partInfo[1];
} GetPartitionListResult;
ITSUTILS_API HRESULT STDAPICALLTYPE ITGetPartitionList(
        DWORD cbInput, GetPartitionListParams *pbInput,
        DWORD *pcbOutput, GetPartitionListResult **ppbOutput,
        IRAPIStream *pStream);

typedef struct tagGetSTRGHandleListResult {
    int nHandles;
    DWORD handle[1];
} GetSTRGHandleListResult;
ITSUTILS_API HRESULT STDAPICALLTYPE ITGetSTRGHandleList(
        DWORD cbInput, void *pbInput,
        DWORD *pcbOutput, GetSTRGHandleListResult**ppbOutput,
        IRAPIStream *pStream);

// these structs are the same for several functions
typedef struct tagReadDiskParams {
    WCHAR szDeviceName[DEVICENAMESIZE];
    WCHAR szPartitionName[PARTITIONNAMESIZE];
    DWORD hDisk;
    DWORD dwBinaryPartitionNr;

    ULONGLONG llOffset;
    DWORD dwSize;
} ReadDiskParams;
typedef struct tagReadDiskResult {
	DWORD dwNumberOfBytesRead;
    BYTE buffer[1];
} ReadDiskResult;
ITSUTILS_API HRESULT STDAPICALLTYPE ITReadDisk(
        DWORD cbInput, ReadDiskParams *pbInput,
        DWORD *pcbOutput, ReadDiskResult **ppbOutput,
        IRAPIStream *pStream);

typedef struct tagWriteDiskParams {
    WCHAR szDeviceName[DEVICENAMESIZE];
    WCHAR szPartitionName[PARTITIONNAMESIZE];
    DWORD hDisk;
    DWORD dwBinaryPartitionNr;

    ULONGLONG llOffset;
    DWORD dwSize;
    BYTE buffer[1];
} WriteDiskParams;
typedef struct tagWriteDiskResult {
	DWORD dwNumberOfBytesWritten;
} WriteDiskResult;
ITSUTILS_API HRESULT STDAPICALLTYPE ITWriteDisk(
        DWORD cbInput, WriteDiskParams *pbInput,
        DWORD *pcbOutput, WriteDiskResult **ppbOutput,
        IRAPIStream *pStream);

typedef struct tagTFFSGetInfoParams {
    WCHAR szDeviceName[DEVICENAMESIZE];
    WCHAR szPartitionName[PARTITIONNAMESIZE];
    DWORD dwBinaryPartitionNr;
    DWORD hDisk;
} TFFSGetInfoParams;
typedef struct tagTFFSGetInfoResult {
    DWORD dwSectorSize;
    ULONGLONG llDiskSize;
} TFFSGetInfoResult;
ITSUTILS_API HRESULT STDAPICALLTYPE ITTFFSGetInfo(
        DWORD cbInput, TFFSGetInfoParams *pbInput,
        DWORD *pcbOutput, TFFSGetInfoResult **ppbOutput,
        IRAPIStream *pStream);

typedef struct tagLogDiskInfoParams {
    WCHAR szDeviceName[DEVICENAMESIZE];
    WCHAR szPartitionName[PARTITIONNAMESIZE];
    DWORD dwBinaryPartitionNr;
    DWORD hDisk;
} LogDiskInfoParams;
ITSUTILS_API HRESULT STDAPICALLTYPE ITLogDiskInfo(
        DWORD cbInput, LogDiskInfoParams *pbInput,
        DWORD *pcbOutput, BYTE **ppbOutput,
        IRAPIStream *pStream);

#endif
