/* (C) 2003 XDA Developers  itsme@xs4all.nl
 *
 * $Header: /var/db/cvs/xdautils/leds/regbk.cpp,v 1.1 2005/01/19 10:58:52 itsme Exp $
 */
#include <windows.h>
#include "debug.h"
#include "args.h"
#include <stdio.h>
#include <string>
#include <map>
#include <vector>
#include "stringutils.h"
#include "vectorutils.h"

#define STD_NAME_BUF_SIZE   256
#define STD_VALUE_BUF_SIZE  1024

#define __RegCloseKey        RegCloseKey
#define __RegCreateKeyEx     RegCreateKeyEx
#define __RegDeleteKey       RegDeleteKey
#define __RegDeleteValue     RegDeleteValue
#define __RegEnumKeyEx       RegEnumKeyEx
#define __RegEnumValue       RegEnumValue
#define __RegOpenKeyEx       RegOpenKeyEx
#define __RegQueryInfoKey    RegQueryInfoKey
#define __RegQueryValueEx    RegQueryValueEx
#define __RegSetValueEx      RegSetValueEx

// todo:
//    - add option to make output dump format compatible with .reg file
//        ( and possibly less readable 'hex' output )
//    - support 'REGEDIT4' fileformat header.
//    - support unicode files
//    - change 'r' option in non-recursive option
//    - add option to choose how to mix keys and values.
//         ( full cross product, or values per key. )

std::string GetNameFromSetSpec(const std::string& spec, size_t start);
std::string GetValueSpecFromSetSpec(const std::string& spec, size_t start);

enum RegAction {
    DO_LOAD,
    DO_LIST,
    DO_DELETE,
    DO_CREATE,
};

class RegistryKey;
/*
class RegistryData {
};
class DwordData : public RegistryData {
public:
    DwordData(ByteVector& data)
    {
    }
};
class BinaryData : public RegistryData {
public:
    BinaryData(ByteVector& data)
    {
    }
};
class StringData : public RegistryData {
public:
    StringData(ByteVector& data)
    {
    }
};
class StringListData : public RegistryData {
public:
    StringListData(ByteVector& data)
    {
    }
};
*/
std::string unescape(const std::string& str);
class RegistryValue {
public:
    RegistryValue()
        : m_dwType(REG_NONE)
    {
        //debug("new regvalue none\n");
    }
    RegistryValue(DWORD dwType, const ByteVector& data)
        : m_dwType(dwType), m_data(data)
    {
        //debug("new regvalue type %d\n", m_dwType);
    }
    RegistryValue(const RegistryValue& rv)
        : m_dwType(rv.GetType()), m_data(rv.GetData())
    {
        //debug("copy of regvalue type %d\n", m_dwType);
    }

    DWORD GetType() const
    {
        return m_dwType;
    }

    // binary value
    RegistryValue(const ByteVector& data)
        : m_dwType(REG_BINARY), m_data(data)
    {
    }
    ByteVector GetData() const
    {
        return m_data;
    }

    // dword value
    RegistryValue(DWORD dwValue)
        : m_dwType(REG_DWORD), m_data(BV_FromDword(dwValue))
    {
    }
    DWORD GetDword() const
    {
        if (m_dwType!=REG_DWORD) { debug("GetDword: not a dword"); exit(1); }
        if (m_data.size()!=4) { debug("GetDword: incorrect size"); exit(1); }

        return BV_GetDword(m_data);
    }

    // string value
    RegistryValue(const std::string& strValue)
        : m_dwType(REG_SZ), m_data(BV_FromWString(ToWString(strValue)))
    {
        BV_AppendWord(m_data, 0); // add terminating (wchar_t)NUL
    }
    std::string GetString() const
    {
        if (m_dwType!=REG_SZ && m_dwType!=REG_EXPAND_SZ) { debug("GetString: not a string"); exit(1); }
        if (m_data.size()&1) { debug("GetString: incorrect size"); exit(1); }

        std::string str;
        for (ByteVector::const_iterator i=m_data.begin() ; i!=m_data.end() && (*i)!=0 ; i+=2)
            str += (char)(*i);

        return str;
    }

    // string list value
    RegistryValue(const StringList& listValue)
        : m_dwType(REG_MULTI_SZ)
    {
        m_data.clear();
        ByteVector::iterator out= m_data.begin();
        for (StringList::const_iterator i= listValue.begin() ; i!=listValue.end() ; ++i)
        {
            BV_AppendWString(m_data, ToWString(*i));
            BV_AppendWord(m_data, 0); // add terminating (wchar_t)NUL
        }
        BV_AppendWord(m_data, 0); // add terminating (wchar_t)NUL
    }
    StringList GetStringList() const
    {
        if (m_dwType!=REG_MULTI_SZ) { debug("GetStringList: not a string"); exit(1); }
        if (m_data.size()&1) { debug("GetStringList: incorrect size"); exit(1); }

        StringList list;
        std::string str;
        for (ByteVector::const_iterator i= m_data.begin() ; i!=m_data.end() ; i+=2)
        {
            if (*i==0)
            {
                list.push_back(str);
                str.clear();
            }
            else
            {
                str += (char)(*i);
            }
        }
        if (!str.empty()) {
            { debug("Invalid multi_sz: no terminating NUL"); exit(1); }
        }
        if (!list.back().empty()) {
            { debug("Invalid multi_sz: no closing element"); exit(1); }
        }
        list.pop_back();
        return list;
    }
    std::string quotestring(const std::string& str) const
    {
        std::string qstr= "\"";
        for (size_t i=0 ; i<str.size() ; i++)
        {
            if (str[i]<' ' || str[i]>'~')
                qstr+=stringformat("\\%02x", str[i]);
            else
                qstr+=str[i];
        }
        return qstr+"\"";
    }
    std::string MultiszAsString() const
    {
        StringList list= GetStringList();

        for (size_t i=0 ; i<list.size() ; ++i)
            list[i]= quotestring(list[i]);
        return std::string("multi_sz:")+JoinStringList(list, ",");
    }
    std::string AsString() const
    {
        switch(m_dwType) {
case REG_NONE:   return "none";
case REG_SZ: return quotestring(GetString());
case REG_EXPAND_SZ: return std::string("expand:")+quotestring(GetString());
case REG_BINARY: return std::string("hex:")+ascdump(m_data);
case REG_DWORD: return stringformat("dword:%08lx", *(DWORD*)vectorptr(m_data));
//case REG_DWORD_BE: return stringformat("dwordbe:%08lx", *(DWORD*)vectorptr(m_data));
case REG_LINK: return "link";
case REG_MULTI_SZ: return MultiszAsString();
case REG_RESOURCE_LIST: return "rsclist";
case REG_FULL_RESOURCE_DESCRIPTOR: return "rscdesc";
case REG_RESOURCE_REQUIREMENTS_LIST: return "rscreqslist";
case REG_QWORD: return "qword";
default: return "UNKNOWN";

        }
    }

private:
    DWORD m_dwType;
    ByteVector m_data;
    //Value* m_value;
};
typedef std::map<std::string,RegistryValue> RegistryValueMap;
typedef std::vector<RegistryKey> RegistryKeyList;


class RegistryKey {
public:
    static RegistryKey FromKeySpec(const std::string& keyspec)
    {
        HKEY hRoot;
        std::string path;

        size_t slashpos= keyspec.find_first_of("/\\");

        if (stringicompare(keyspec.substr(0,slashpos), "hkcr")==0)
            hRoot= HKEY_CLASSES_ROOT;
        else if (stringicompare(keyspec.substr(0,slashpos), "hkey_classes_root")==0)
            hRoot= HKEY_CLASSES_ROOT;
        else if (stringicompare(keyspec.substr(0,slashpos), "hkcu")==0)
            hRoot= HKEY_CURRENT_USER;
        else if (stringicompare(keyspec.substr(0,slashpos), "hkey_current_user")==0)
            hRoot= HKEY_CURRENT_USER;
        else if (stringicompare(keyspec.substr(0,slashpos), "hklm")==0)
            hRoot= HKEY_LOCAL_MACHINE;
        else if (stringicompare(keyspec.substr(0,slashpos), "hkey_local_machine")==0)
            hRoot= HKEY_LOCAL_MACHINE;
        else
            hRoot= HKEY_LOCAL_MACHINE;

        if (slashpos==keyspec.npos)
            path= "";
        else
            path= keyspec.substr(slashpos+1);

        //debug("FromKeySpec: %08lx - %s\n", hRoot, path.c_str());

        return RegistryKey(hRoot, path);
    }

    RegistryKey() : m_hRoot(0), m_path(), m_hKey(0)
    {
    }

    RegistryKey(HKEY hRoot) : m_hRoot(hRoot), m_path(), m_hKey(0)
    {
    }

    RegistryKey(const RegistryKey& hKey) : m_hRoot(hKey.GetRoot()), m_path(hKey.GetPath()), m_hKey(0)
    {
    }
    RegistryKey& operator=(const RegistryKey& rk)
    {
        m_hRoot=rk.GetRoot();
        m_path= rk.GetPath();
        m_hKey= 0;
        return *this;
    }
    RegistryKey(const RegistryKey& hRoot, const std::string& path) : m_hRoot(hRoot.GetRoot()), m_hKey(0)
    {
        m_path= hRoot.GetPath();
        if (!m_path.empty())
            m_path += "\\";
        m_path += path;
    }
    ~RegistryKey()
    {
        Close();
    }

    void Close()
    {
        __RegCloseKey(m_hKey);
        m_hKey= 0;
    }
    bool Open()
    {
        if (m_hKey)
            return true;

        LONG res= __RegOpenKeyEx(m_hRoot, ToWString(m_path).c_str(), 0, KEY_ALL_ACCESS, &m_hKey);
        if (res!=ERROR_SUCCESS)
        {
            error(res, "Failed to open key %hs:'%hs'", GetRootName().c_str(), m_path.c_str());
            m_hKey= 0;
            return false;
        }
        m_dwMaxSubKeyNameLength= STD_NAME_BUF_SIZE;
        m_dwMaxValueNameLength= STD_NAME_BUF_SIZE;
        m_dwMaxValueLength= STD_VALUE_BUF_SIZE;
/*
 *
 * this does not seem to work on ce/rapi
 *
        __RegQueryInfoKey(m_hKey, NULL, NULL, NULL, NULL, &m_dwMaxSubKeyNameLength,
                NULL, NULL, &m_dwMaxValueNameLength, &m_dwMaxValueLength, NULL, NULL);

        if ( m_dwMaxSubKeyNameLength==0 && m_dwMaxValueNameLength==0 && m_dwMaxValueLength==0)
        {
            debug("ERROR: noinfo\n");
            m_dwMaxSubKeyNameLength= STD_NAME_BUF_SIZE;
            m_dwMaxValueNameLength= STD_NAME_BUF_SIZE;
            m_dwMaxValueLength= STD_VALUE_BUF_SIZE;
        }
*/
        return true;
    }
    HKEY GetHKey()
    {
        Open();

        return m_hKey;
    }

    void info()
    {
        std::wstring regclass; regclass.resize(256);
        DWORD cbClass= regclass.size();

        DWORD nSubKeys=0;
        DWORD maxSubKeyNameLength=0;
        DWORD maxClassLength=0;
        DWORD nValues=0;
        DWORD maxValueNameLength=0;
        DWORD maxValueLength=0;

        Open();

        // BUG: only the key and value counts are returned
        LONG res= __RegQueryInfoKey(m_hKey, stringptr(regclass), &cbClass, NULL, &nSubKeys,
                &maxSubKeyNameLength, &maxClassLength, &nValues,
                &maxValueNameLength, &maxValueLength, NULL, NULL);
        if (res != ERROR_SUCCESS)
        {
            error(res, "__RegQueryInfoKey(%hs:'%hs')", GetRootName().c_str(), m_path.c_str());
            return ;
        }
        regclass.resize(cbClass);

        debug("regclass='%hs'  nkeys=%d nvals=%d  maxkey=%d maxcls=%d maxname=%d maxval=%d\n",
                regclass.c_str(), nSubKeys, maxSubKeyNameLength,
                maxClassLength, nValues, maxValueNameLength,&maxValueLength);
    }

    // returns the roothive of this key
    HKEY GetRoot() const
    {
        return m_hRoot;
    }

    // returns the full path to this key
    std::string GetPath() const
    {
        return m_path;
    }

    // returns the name of this key
    std::string GetName() const
    {
        size_t pos= m_path.find_last_of("\\/");
        if (pos!=m_path.npos)
            return m_path.substr(pos+1);
        else
            return m_path;
    }

    // returns the name of the hive of this key
    std::string GetRootName() const
    {
        if (m_hRoot==HKEY_LOCAL_MACHINE) return "HKLM";
        if (m_hRoot==HKEY_CURRENT_USER) return "HKCU";
        if (m_hRoot==HKEY_CLASSES_ROOT) return "HKCR";
        return "unknown";
    }

    // creates a new subkey
    bool CreateSubKey(const std::string& keyname)
    {
        if (!Open())
            return false;

        HKEY hk;
        DWORD disp;
        LONG res= __RegCreateKeyEx(m_hKey, ToWString(keyname).c_str(), 0, NULL, 0, 0, NULL, &hk, &disp);
        if (res != ERROR_SUCCESS)
        {
            error(res, "Failed to create subkey %hs:'%hs'\\'%hs'", GetRootName().c_str(), m_path.c_str(), keyname.c_str());
            return false;
        }
        __RegCloseKey(hk);
        return true;
    }


    // deletes a childless subkey from this key
    bool DeleteSubKey(const std::string& keyname)
    {
        if (!Open())
            return false;
        LONG res= __RegDeleteKey(m_hKey, ToWString(keyname).c_str());
        if (res != ERROR_SUCCESS)
        {
            error(res, "Failed to delete1 subkey %hs:'%hs'\\'%hs'", GetRootName().c_str(), m_path.c_str(), keyname.c_str());
            return false;
        }
        return true;
    }

    // delete all children
    bool DeleteChildren()
    {
        bool bRes= true;
        StringList children= GetSubKeyNames();
        for (StringList::iterator i= children.begin() ; i!=children.end() ; ++i)
        {
            bRes = bRes && DeleteSubKey(*i);
        }
        return bRes;
    }

    bool DeleteValue(const std::string& valuename)
    {
        if (!Open())
            return false;
        LONG res= __RegDeleteValue(m_hKey, ToWString(valuename).c_str());
        if (res != ERROR_SUCCESS)
        {
            error(res, "Failed to delete value %hs:'%hs'\\'%hs'", GetRootName().c_str(), m_path.c_str(), valuename.c_str());
            return false;
        }
        return true;
    }

    //-------------------------------------------------------------------
    // deletes this key and all it's children
    bool DeleteKey()
    {
        if (!DeleteChildren())
            return false;

        Close();

        if (!GetParent().DeleteSubKey(GetName()))
            return false;

        return true;
    }

    //-------------------------------------------------------------------
    // creates registry key
    bool CreateKey()
    {
        if (Open()) {
            Close();
            return true;
        }

        if (!GetParent().CreateKey())
            return false;

        if (!GetParent().CreateSubKey(GetName()))
            return false;

        debug("created key %hs\\%hs\n", GetRootName().c_str(), GetPath().c_str());
        return true;
    }


    void DumpKey()
    {
        debug("[%hs\\%hs]\n", GetRootName().c_str(), GetPath().c_str());
        DumpValues();
        debug("\n");
        DumpChildren();
    }

    //-------------------------------------------------------------------
    void DumpValues()
    {
        RegistryValueMap map= GetValues();

        for (RegistryValueMap::iterator i= map.begin() ; i!=map.end() ; ++i)
        {
            debug("%hs=%hs\n", (*i).first.c_str(), (*i).second.AsString().c_str());
        }
    }

    void DumpChildren()
    {
        RegistryKeyList list= GetSubKeys();

        for (RegistryKeyList::iterator i= list.begin() ; i!=list.end() ; ++i)
        {
            (*i).DumpKey();
        }
    }

    //-------------------------------------------------------------------
    // returns parent to this key
    RegistryKey GetParent()
    {
        size_t pos= m_path.find_last_of("\\/");
        if (pos!=m_path.npos)
            return RegistryKey(m_hRoot, m_path.substr(0, pos));
        else
            return RegistryKey(m_hRoot);
    }
    StringList GetSubKeyNames()
    {
        StringList list;
        if (!Open())
            { debug("GetSubKeyNames: could not open"); exit(1); }

        for (int i=0 ; true ; i++)
        {
            DWORD cbName = m_dwMaxSubKeyNameLength+1;
            std::wstring wname; wname.resize(cbName);
            LONG res =__RegEnumKeyEx(m_hKey, i, stringptr(wname), &cbName, NULL, NULL, NULL, NULL);
            if (res==ERROR_NO_MORE_ITEMS)
                break;
            if (res!=ERROR_SUCCESS)
            {
                error(res, "__RegEnumKeyEx(%hs:'%hs', %d)", GetRootName().c_str(), GetPath().c_str(), i);
                continue;
            }
            wname.resize(cbName);
            list.push_back(ToString(wname));
        }
        return list;
    }
    RegistryKeyList GetSubKeys()
    {
        RegistryKeyList list;
        if (!Open())
            { debug("GetSubkeys: could not open"); exit(1); }

        for (int i=0 ; true ; i++)
        {
            DWORD cbName = m_dwMaxSubKeyNameLength+1;
            std::wstring wname; wname.resize(cbName);
            LONG res =__RegEnumKeyEx(m_hKey, i, stringptr(wname), &cbName, NULL, NULL, NULL, NULL);
            if (res==ERROR_NO_MORE_ITEMS)
                break;
            if (res!=ERROR_SUCCESS)
            {
                error(res, "__RegEnumKeyEx(%hs:'%hs', %d)", GetRootName().c_str(), GetPath().c_str(), i);
                continue;
            }
            wname.resize(cbName);
            list.push_back(RegistryKey(*this, ToString(wname)));
        }
        return list;
    }
    StringList GetValueNames()
    {
        StringList list;
        if (!Open())
            { debug("GetValueNames: could not open"); exit(1); }

        for (int i=0 ; true ; i++)
        {
            DWORD cbName = m_dwMaxValueNameLength+1;
            std::wstring wname; wname.resize(cbName);
            LONG res= __RegEnumValue(m_hKey, i, stringptr(wname), &cbName, NULL, NULL, NULL, NULL);
            if (res==ERROR_NO_MORE_ITEMS)
                break;
            if (res!=ERROR_SUCCESS)
            {
                error(res, "__RegEnumValue(%hs:'%hs', %d)", GetRootName().c_str(), GetPath().c_str(), i);
                continue;
            }
            wname.resize(cbName);
            list.push_back(ToString(wname));
        }
        return list;
    }
    RegistryValueMap GetValues()
    {
        RegistryValueMap map;
        if (!Open())
            { debug("GetValues: could not open"); exit(1); }

        for (int i=0 ; true ; i++)
        {
            DWORD cbName = m_dwMaxValueNameLength+1;
            std::wstring wname; wname.resize(cbName);
            DWORD cbData = m_dwMaxValueLength;
            ByteVector data; data.resize(cbData);
            DWORD dwType;
            LONG res= __RegEnumValue(m_hKey, i, stringptr(wname), &cbName, NULL, &dwType, vectorptr(data), &cbData);
            if (res==ERROR_NO_MORE_ITEMS)
                break;
            if (res!=ERROR_SUCCESS)
            {
                error(res, "__RegEnumValue(%hs:'%hs', %d)", GetRootName().c_str(), GetPath().c_str(), i);
                continue;
            }
            wname.resize(cbName);
            data.resize(cbData);

            map[ToString(wname)]= RegistryValue(dwType, data);
        }
        return map;
    }
    RegistryValue GetValue(const std::string& valuename)
    {
        if (!Open())
            { debug("GetValue: could not open"); exit(1); }
        DWORD cbData = m_dwMaxValueLength;
        ByteVector data; data.resize(cbData);
        DWORD dwType;
        LONG res= __RegQueryValueEx(m_hKey, ToWString(valuename).c_str(), NULL, &dwType, vectorptr(data), &cbData);
        if (res!=ERROR_SUCCESS)
        {
            error(res, "__RegQueryValueEx(%hs:'%hs' '%hs')", GetRootName().c_str(), GetPath().c_str(), valuename.c_str());

            { debug("GetValue: value does not exist"); exit(1); }
        }
        data.resize(cbData);
        return RegistryValue(dwType, data);
    }
    bool SetValue(const std::string& valuename, const RegistryValue& value)
    {
        if (!Open())
            { debug("SetValue: could not open"); exit(1); }

        //debug("setting [%s] %s to %s\n", m_path.c_str(), valuename.c_str(), value.AsString().c_str());
        ByteVector data= value.GetData();
        LONG res= __RegSetValueEx(m_hKey, ToWString(valuename).c_str(), NULL, value.GetType(), vectorptr(data), data.size());
        if (res!=ERROR_SUCCESS)
        {
            error(res, "__RegSetValueEx(%hs:'%hs' '%hs')", GetRootName().c_str(), GetPath().c_str(), valuename.c_str());
            { debug("SetValue: could not create"); exit(1); }
        }
        debug("set value %hs\\%hs  %hs : %hs\n", GetRootName().c_str(), GetPath().c_str(), valuename.c_str(), value.AsString().c_str());
        return true;
    }

private:
        HKEY m_hRoot;
        std::string m_path;

        HKEY m_hKey;
        DWORD m_dwMaxSubKeyNameLength;
        DWORD m_dwMaxValueNameLength;
        DWORD m_dwMaxValueLength;
};

size_t findendquote(const std::string& str, size_t pos, char quotechar)
{
    bool bEscaped= false;
    while (pos<str.size())
    {
        if (bEscaped)
        {
            bEscaped= false;
        }
        else if (str[pos]==quotechar)
        {
            return pos+1;
        }
        else if (str[pos]=='\\')
        {
            bEscaped= true;
        }
        ++pos;
    }
    return std::string::npos;
}
std::string unescape(const std::string& str)
{
    std::string result;
    size_t pos= 0;
    bool bEscaped= false;
    while (pos<str.size())
    {
        if (bEscaped)
        {
            switch(str[pos])
            {
                case 'n': result+='\n'; break;
                case 'r': result+='\r'; break;
                case 't': result+='\t'; break;
                case 'x': if (pos+2<str.size())
                          {
                              result += (char)strtoul(str.substr(pos+1,2).c_str(),0,16); 
                              pos+=2;
                          }
                          break;
                case '0':
                case '1':
                case '2': if (pos+2<str.size())
                          {
                              result += (char)strtoul(str.substr(pos,3).c_str(),0,8); 
                              pos+=2;
                          }
                          break;
                default:
                          result+=str[pos];
            }
            bEscaped= false;
        }
        else if (str[pos]=='\\')
        {
            bEscaped= true;
        }
        else {
            result += str[pos];
        }
        ++pos;
    }

    return result;
}
bool ReadLine(FILE *f, std::string& line)
{
    line.clear();
    line.resize(1024);

    char *p= fgets(stringptr(line), 1024, f);
    if (p==NULL) {
        line.clear();
        return false;
    }
    line.resize(strlen(p));
    // todo: handle long lines ( > 1024 )
    //  - read until line has EOLN

    while (line.size() && (line[line.size()-1]=='\r' || line[line.size()-1]=='\n')) {
        line.resize(line.size()-1);
    }
    return true;
}

bool IsValueNameSpec(const std::string& spec)
{
    if (spec[0]!=':')
        return false;
    if (spec.find('=')==std::string::npos)
        return true;

    if (spec[1]=='\'' || spec[1]=='\"')
    {
        size_t endquote= findendquote(spec, 2, spec[1]);

        if (endquote==std::string::npos)
        {
            { debug("IsValueNameSpec: missing endquote"); exit(1); }
        }
        return endquote==spec.size();
    }
    return false;
}

std::string GetNameFromValueNameSpec(const std::string& spec, size_t start)
{
    if (spec[start]=='\'' || spec[start]=='\"')
    {
        size_t endquote= findendquote(spec, start+1, spec[start]);
        return unescape(spec.substr(start, endquote-start-1));
    }
    return spec.substr(start);
}

bool IsSetSpec(const std::string& spec)
{
    if (spec[0]!=':')
        return false;
    if (spec.find('=')==std::string::npos)
        return false;

    if (spec[1]=='\'' || spec[1]=='\"')
    {
        size_t endquote= findendquote(spec, 2, spec[1]);

        if (endquote==std::string::npos)
        {
            { debug("IsSetSpec: missing endquote\n"); exit(1); }
        }
        //debug("isset: endq=%d size=%d c=%c\n", endquote, spec.size(), spec[endquote]);
        if (endquote==spec.size())
            return false;

        return (spec[endquote]=='=');
    }
    return true;
}
std::string GetNameFromSetSpec(const std::string& spec, size_t start)
{
    if (spec[start]=='\'' || spec[start]=='\"')
    {
        size_t endquote= findendquote(spec, start+1, spec[start]);

        //todo: check for missing quote.
        return unescape(spec.substr(start+1, endquote-start-2));
    }
    size_t eqpos= spec.find('=');

    return spec.substr(start,eqpos-start);
}

std::string GetValueSpecFromSetSpec(const std::string& spec, size_t start)
{
    if (spec[start]=='\'' || spec[start]=='\"')
    {
        size_t endquote= findendquote(spec, start+1, spec[start]);

        //todo: check for missing quote.
        size_t eqpos= spec.find('=', endquote);
        return spec.substr(eqpos+1);
    }
    size_t eqpos= spec.find('=');

    return spec.substr(eqpos+1);
}

bool IsRegFile(const std::string& filespec)
{
    return filespec[0]=='@';
}
std::string GetFilenameFromRegfileSpec(const std::string& filespec)
{
    return filespec.substr(1);
}

void checkparameters(RegAction action, const RegistryKeyList& keys, const StringList& names, const RegistryValueMap& values)
{
    if (action==DO_CREATE && names.size())
        { debug("checkparameters: create with name"); exit(1); }
    if (action==DO_CREATE && values.size())
        { debug("checkparameters: create with 'name=value'"); exit(1); }
    if (action==DO_CREATE && keys.size()==0)
        { debug("checkparameters: create without keys"); exit(1); }
    if (action==DO_DELETE && values.size())
        { debug("checkparameters: delete with 'name=value'"); exit(1); }
    if (action==DO_DELETE && keys.size()==0)
        { debug("checkparameters: delete without keys"); exit(1); }
    if (action==DO_LIST && values.size())
        { debug("checkparameters: list with 'name=value'"); exit(1); }
    if (action==DO_LIST && keys.size()==0)
        { debug("checkparameters: list without keys"); exit(1); }
}

void usage()
{
    debug("Usage: pregutl [keys] [values] [@regfile]\n");
    debug("  -d   delete value or key\n");
    debug("  -c   create key\n");
    debug("  -s   set value\n");
    debug("  -r   recurse\n");
    debug("\n");
    debug("key-values are specified as follows:\n");
    debug("    :valuename  or :'value name'\n");
    debug("    :valuename=value\n");
    debug("\n");
    debug("values are specified as follows:\n");
    debug("    dword:01234567  or \"str\\n\\\"str\" or\n");
    debug("    multi_sz:\"...\",\"...\" or  hex:ff,11,22\n");
    debug("    note that you may also use single (') quotes instead\n");
    debug("    note also, that backslashes in strings must be escaped: \"\\\\windows\"\n");
    debug("    otherwise it would be almost impossible to enter quotes on the dos cmdline\n");
    debug("\n");
    debug("keys are specified as follows:\n");
    debug("    HKLM\\Software\\HTC\n");
    debug("\n");
    debug("if multiple keys and keyvals are specified pregutl\n");
    debug("operates on all combinations\n");
}

int WINAPI WinMain( HINSTANCE hInstance,
                   HINSTANCE hPrevInstance,
                   LPTSTR    lpCmdLine,
                   int       nCmdShow)
{
    DebugSetLogfile("\\regbk.log");

    RegAction action= DO_LIST;
    bool fRecurse= false;

    RegistryKeyList keys;
    StringList names;
    RegistryValueMap values;
    StringList regfiles;

    RegistryKey::FromKeySpec("HKCU").DumpKey();
    RegistryKey::FromKeySpec("HKLM").DumpKey();

    return 0;
}


