#include "debug.h"
#include <Tlhelp32.h>

//#include "cenk.h"
#include "types.h"
#include "PKFUNCS.H"
#include "kernel.h"
#include "schedule.h"
/*
HDATA
    DList       linkage;    // 00: links for active handle list
        DList *fwd;             // 00: forward link
        DList *back;            // 04: backward link
    HANDLE      hValue;     // 08: Current value of handle (nonce)
    ACCESSLOCK  lock;       // 0C: access information
    REFINFO     ref;        // 10: reference information
        ulong   count;          // 10:
        FULLREF *pFr;           // 10: ptr to ushort  usRefs[MAX_PROCESSES];
    const CINFO *pci;       // 14: ptr to object class description structure
    PVOID       pvObj;      // 18: ptr to object (in server address space )
    DWORD       dwInfo;     // 1C: extra handle info

CINFO
    char        acName[4];  // 00: object type ID string
    uchar       disp;       // 04: type of dispatch
    uchar       type;       // 05: api handle type
    ushort      cMethods;   // 06: # of methods in dispatch table
    const PFNVOID *ppfnMethods;// 08: ptr to array of methods (in server address space)
    const DWORD *pdwSig;    // 0C: ptr to array of method signatures
    PPROCESS    pServer;    // 10: ptr to server process

 */
//--------------- SOCK_INFO is in wstype.h ----------------
// but it seems impossible to combine it with the kernel headers.
//#include "sockinfo.h"
#include "sockinfo420.h"

#include "device420.h"			// fsopendev_t

#if _WIN32_WCE==300
#include "filehandle300.h"
#elif _WIN32_WCE==420
// this is only defined in the wce4.x sdk, not for 3.0
#include "filehandle300.h"
#else
#error "unknown windows ce version"
#endif

#include <map>
#include <set>
#include <string>
#include "stringutils.h"

typedef std::set<const CINFO*> CInfoSet;
typedef std::map<std::string,CInfoSet> CInfoMap;

//HANDLE g_hFilesysProc;
//HANDLE g_hDeviceProc;

CInfoMap cinfomap;

bool dumpmem(HANDLE hProc, DWORD dwOfs, DWORD dwLen, const std::string& msg)
{
    ByteVector dw; dw.resize(dwLen);
    DWORD nRead;
    if (!ReadProcessMemory(hProc,  (void*)dwOfs, vectorptr(dw), dw.size(), &nRead)) {
        error("reading %08lx : %08lx\n", hProc, dwOfs);
        return false;
    }
    bighexdump(dwOfs, dw, hexdumpflags(DUMPUNIT_DWORD, 4, DUMP_HEX)|HEXDUMP_WITH_OFFSET);

    return true;
}

HANDLE GetProcessHandle(WCHAR *wszProcessName)
{
    HANDLE hTH= CreateToolhelp32Snapshot(TH32CS_SNAPPROCESS, 0);

    PROCESSENTRY32 pe;
    pe.dwSize= sizeof(PROCESSENTRY32);

    HANDLE hProc= INVALID_HANDLE_VALUE;
    if (Process32First(hTH, &pe))
    {
        do {

            if (wcsicmp(wszProcessName, pe.szExeFile)==0)
            {
                hProc= OpenProcess(0, 0, pe.th32ProcessID);
                if (hProc != INVALID_HANDLE_VALUE && hProc!=NULL)
                    break;
            }
        } while (Process32Next(hTH, &pe));
    }

    CloseToolhelp32Snapshot(hTH);

    return hProc;
}
char *GetCInfoName(const CINFO *ci)
{
	static char name[5];
	for (int i=0 ; i<4 ; i++)
		name[i]= ci->acName[i];
	name[4]=0;
	return name;
}
void DumpCInfo(const CINFO *pci);
void RegisterCInfo(const CINFO *pci)
{
    std::string name(GetCInfoName(pci));

	cinfomap[name].insert(pci);
}


void DumpCInfoMap()
{
	for (CInfoMap::iterator mi= cinfomap.begin() ; mi!=cinfomap.end() ; ++mi)
	{
		debug("....%hs....\n", (*mi).first.c_str());
		for (CInfoSet::iterator si= (*mi).second.begin() ; si!=(*mi).second.end() ; ++si)
		{
			DumpCInfo(*si);
		}
	}
}
char *GetMethodSigString(DWORD dwSig)
{
	static char str[256];
	char *p= str;

	p+=sprintf(p, "(");
	for (int i=0 ; i<16 && dwSig ; i++)
	{
		if (i)	p+=sprintf(p, ", ");
		switch(dwSig&3)
		{
		case 0: p+=sprintf(p, "DWORD"); break;
		case 1: p+=sprintf(p, "PTR"); break;
		case 2: p+=sprintf(p, "I64"); break;
		case 3: p+=sprintf(p, "???"); break;
		}
		dwSig>>=4;
	}
	p+=sprintf(p, ")");
	return str;
}

void DumpCInfo(const CINFO *pci)
{
	debug("CINFO  %-4s  n=%d disp=%02x type=%02x  methods=%08lx sigs=%08lx server=%08lx %ls\n", 
		GetCInfoName(pci), pci->cMethods, pci->disp, pci->type,
		pci->ppfnMethods, pci->pdwSig,
		pci->pServer, IsBadReadPtr(pci->pServer, sizeof(PROCESS))?L"-":pci->pServer->lpszProcName);

	const PFNVOID *ppfnMethods= IsBadReadPtr(pci->pServer, sizeof(PROCESS))
			?pci->ppfnMethods
			:(const PFNVOID *)MapPtrUnsecure((void*)pci->ppfnMethods, pci->pServer->hProc);
	const DWORD *pdwSig= IsBadReadPtr(pci->pServer, sizeof(PROCESS))
			?pci->pdwSig
			:(const DWORD *)MapPtrUnsecure((void*)pci->pdwSig, pci->pServer->hProc);
	for (int i=0 ; i<pci->cMethods ; i++)
	{
		debug("apiset  %-4s %3d %08lx %hs\n", GetCInfoName(pci), 
			i, ppfnMethods[i], (pdwSig==NULL || IsBadReadPtr(pdwSig, sizeof(DWORD)*i))?"-":GetMethodSigString(pdwSig[i]));
	}
}
void DumpThreadTimes(HANDLE hThread)
{
	FILETIME tCreate, tExit, tKernel, tUser;
    if (!GetThreadTimes(hThread, &tCreate, &tExit, &tKernel, &tUser))
	{
		error("GetThreadTimes");
        return;
	}

	debug("      cr= %08lx%08lx  ex=%08lx%08lx  k=%08lx%08lx  u=%08lx%08lx\n",
		tCreate.dwHighDateTime, tCreate.dwLowDateTime,
		tExit.dwHighDateTime, tExit.dwLowDateTime,
		tKernel.dwHighDateTime, tKernel.dwLowDateTime,
		tUser.dwHighDateTime, tUser.dwLowDateTime);
}

void DumpHeapEntry(HEAPENTRY32 *he)
{
	debug("    %4d hBlk=%08lx addr=%08lx size=%08lx fl=%08lx lck=%d ?=%08lx pid=%08lx hid=%08lx\n",
		he->dwSize, he->hHandle, he->dwAddress, he->dwBlockSize,
		he->dwFlags, he->dwLockCount, he->dwResvd, 
		he->th32ProcessID, he->th32HeapID); 

}

void DumpHeap(HANDLE hTH, DWORD nProcessId, DWORD nHeapId)
{
	HEAPENTRY32 he;
	

	DWORD dwTotal=0;

	__try {
		he.dwSize= sizeof(HEAPENTRY32);
		if (Heap32First(hTH, &he, nProcessId, nHeapId))
		{
			do {
				//DumpHeapEntry(&he);

				dwTotal += he.dwBlockSize;

				he.dwSize= sizeof(HEAPENTRY32);
			} while (Heap32Next(hTH, &he));
		}
	}
	__except(1) {
		debug("exception reading heap\n");
	}

	error("done looping over heap - structsize=%d  heapsize=%ld\n", sizeof(HEAPENTRY32), dwTotal);
}


void DumpHeapListEntry(HEAPLIST32 *le)
{
	debug("%4d pid=%08lx hid=%08lx fl=%08lx\n", 
		le->dwSize, le->th32ProcessID, le->th32HeapID, le->dwFlags);
}

void DumpHeapLists(HANDLE hTH)
{
	HEAPLIST32 le;
	le.dwSize= sizeof(HEAPLIST32);

	if (Heap32ListFirst(hTH, &le))
	{
		do {
			DumpHeapListEntry(&le);

			DumpHeap(hTH, le.th32ProcessID, le.th32HeapID);
			le.dwSize= sizeof(HEAPLIST32);
		} while (Heap32ListNext(hTH, &le));
	}

	error("done looping over heap lists - structsize=%d\n", sizeof(HEAPLIST32));
}

void DumpModuleEntry(MODULEENTRY32 *me)
{
	debug("%4d fl=%08lx mid=%08lx pid=%08lx gusg=%3d pusg=%08lx base=%08lx size=%08lx hmod=%08lx mod=%ls exe=%ls\n",
		me->dwSize, me->dwFlags, me->th32ModuleID, me->th32ProcessID, 
		me->GlblcntUsage, me->ProccntUsage, 
		me->modBaseAddr, me->modBaseSize, 
		me->hModule, 
		me->szModule, me->szExePath);

}

void DumpModules(HANDLE hTH)
{
	MODULEENTRY32 me;
	me.dwSize= sizeof(MODULEENTRY32);

	if (Module32First(hTH, &me))
	{
		do {
			DumpModuleEntry(&me);

			me.dwSize= sizeof(MODULEENTRY32);
		} while (Module32Next(hTH, &me));
	}

	error("done looping over modules - structsize=%d\n", sizeof(MODULEENTRY32));
}

void DumpProcessEntry(PROCESSENTRY32 *pe)
{
	debug("%2d usg=%d pid=%08lx hid=%08lx mid=%08lx #t=%2d ppid=%08lx pri=%08lx fl=%08lx mem=%08lx acc=%08lx %ls\n",
		pe->dwSize, pe->cntUsage, pe->th32ProcessID, pe->th32DefaultHeapID, 
		pe->th32ModuleID, pe->cntThreads, pe->th32ParentProcessID, 
		pe->pcPriClassBase, pe->dwFlags, 

		pe->th32MemoryBase,	pe->th32AccessKey,
		pe->szExeFile); 
}

void DumpProcesses(HANDLE hTH)
{
	PROCESSENTRY32 pe;
	pe.dwSize= sizeof(PROCESSENTRY32);

	if (Process32First(hTH, &pe))
	{
		do {
			DumpProcessEntry(&pe);

			if (pe.th32DefaultHeapID)
				DumpHeap(hTH, pe.th32ProcessID, pe.th32DefaultHeapID);

			pe.dwSize= sizeof(PROCESSENTRY32);
		} while (Process32Next(hTH, &pe));
	}

	error("done looping over processes - structsize=%d\n", sizeof(PROCESSENTRY32));
}

void DumpThreadEntry(THREADENTRY32 *te)
{
	debug("%2d usg=%d tid=%08lx opid=%08lx pri=%3d dpri=%3d fl=%08lx acc=%08lx cpid=%08lx\n",
		te->dwSize, te->cntUsage, te->th32ThreadID, te->th32OwnerProcessID, 
		te->tpBasePri, te->tpDeltaPri, te->dwFlags,te->th32AccessKey,
		te->th32CurrentProcessID);
}
void DumpThreads(HANDLE hTH)
{
	THREADENTRY32 te;
	te.dwSize= sizeof(THREADENTRY32);

	if (Thread32First(hTH, &te))
	{
		do {
			DumpThreadEntry(&te);

			DumpThreadTimes((HANDLE)te.th32ThreadID);

			te.dwSize= sizeof(THREADENTRY32);
		} while (Thread32Next(hTH, &te));
	}

	error("done looping over threads - structsize=%d\n", sizeof(THREADENTRY32));
}

void test()
{
	HANDLE hTH= CreateToolhelp32Snapshot(TH32CS_SNAPALL | TH32CS_GETALLMODS, 0);


	debug("-------------threads-------------\n");
	DumpThreads(hTH);

	debug("-------------process-------------\n");
	DumpProcesses(hTH);

	debug("-------------modules-------------\n");
	DumpModules(hTH);

	debug("-------------heaplists-------------\n");
	DumpHeapLists(hTH);

	CloseToolhelp32Snapshot(hTH);

}

HDATA *cvHandle2HDataPtr(HANDLE h)
{
	struct KDataStruct *k= (struct KDataStruct *)0xffffc800;

    return (HDATA *)(k->handleBase+((DWORD)h&0x1ffffffc));
}
DWORD SetAccess(HANDLE hProc, DWORD aky)
{
	//HANDLE hProc= (HANDLE)GetCurrentProcessId();

	HDATA *pHandle= cvHandle2HDataPtr(hProc);

	PROCESS *proc= (PROCESS*)pHandle->pvObj;

	DWORD oldaky= proc->aky;
	proc->aky= 0xffffffff;
	return oldaky;
}

DWORD GetHandleObject(HANDLE h)
{
	HDATA *hi= cvHandle2HDataPtr(h);
	return (DWORD)hi->pvObj;
}

TCHAR *GetProcessNameByHandle(HANDLE h)
{
	return ((PROCESS*)GetHandleObject(h))->lpszProcName;
}
void DumpProcess(HANDLE hProc, PROCESS *p)
{
	debug("  num=%d vmbase=%08lx aky=%08lx base=%08lx e32.base=%08lx  %ls %ls", 
		p->procnum, p->dwVMBase, p->aky, p->BasePtr,
		p->e32.e32_vbase, p->lpszProcName, p->pcmdline);
/*
	if (p->o32_ptr)
	{
		for (int i=0 ; i<p->e32.e32_objcnt ; i++)
			debug("    o32-%d : data=%08lx real=%08lx\n", p->o32_ptr[i].o32_dataptr, p->o32_ptr[i].o32_realaddr);
	}
*/
}

void DumpThread(HANDLE hProc, THREAD *t)
{
	debug("  utime=%08lx ktime=%08lx proc=%08lx %ls",
		t->dwUserTime, t->dwKernTime, t->pOwnerProc, t->pOwnerProc->lpszProcName);
}

void DumpEvent(HANDLE hProc, EVENT *e)
{
	if (e->name)
		debug(" %ls", e->name->name);
}
void DumpApiSet(HANDLE hProc, APISET *a)
{
	debug("  reg=%2d  %-4s  n=%d disp=%02x type=%02x  methods=%08lx sigs=%08lx server=%08lx %ls", 
            a->iReg, 
		GetCInfoName(&a->cinfo), a->cinfo.cMethods, a->cinfo.disp, a->cinfo.type,
		a->cinfo.ppfnMethods, a->cinfo.pdwSig,
		a->cinfo.pServer, a->cinfo.pServer->lpszProcName);

	RegisterCInfo(&a->cinfo);
}
void DumpMutex(HANDLE hProc, MUTEX* m)
{
	debug(" %d locks", m->LockCount);
	if (m->name)
		debug(" %ls", m->name->name);
}
void DumpSemaphore(HANDLE hProc, SEMAPHORE* s)
{
	debug(" cur=%d max=%d pend=%d", s->lCount, s->lMaxCount, s->lPending);	
	if (s->name)
		debug(" %ls", s->name->name);
}
void DumpFSMap(HANDLE hProc, FSMAP* f)
{
	if (f->name)
		debug(" %ls", f->name->name);
}
std::string dwptrvaluestr(HANDLE hProc, DWORD dwOfs)
{
    DWORD value;
    DWORD nRead;
    if (dwOfs==NULL) return "(null)";

    if (!ReadProcessMemory(hProc,  (void*)dwOfs, &value, sizeof(value), &nRead)) {
        error("reading %08lx : %08lx\n", hProc, dwOfs);
        return false;
    }
    return stringformat("%08lx", value);
}
std::string procstring(HANDLE hProc, DWORD dwOfs)
{
    WCHAR str[260];
    DWORD nRead;
    if (dwOfs==NULL) return "(null)";

    if (!ReadProcessMemory(hProc,  (void*)dwOfs, str, sizeof(str), &nRead)) {
        error("reading %08lx : %08lx\n", hProc, dwOfs);
        return false;
    }
    return ToString(str);
}
void FindModuleNameForProcessAddress(DWORD pid, DWORD dwAddress, TCHAR *szModname, int maxmodname)
{
	HANDLE hTH= CreateToolhelp32Snapshot(TH32CS_SNAPMODULE, pid);

	MODULEENTRY32 me;
	me.dwSize= sizeof(MODULEENTRY32);

    szModname[0]= 0;

	if (Module32First(hTH, &me))
	{
		do {
			if ((DWORD)me.modBaseAddr<=dwAddress && dwAddress<(DWORD)me.modBaseAddr+me.modBaseSize)
            {
                if (me.szExePath[0])
                    _snwprintf(szModname, maxmodname, L"%ls\\%ls", me.szExePath, me.szModule);
                else
                    wcsncpy(szModname, me.szModule, maxmodname);
                break;
            }

			me.dwSize= sizeof(MODULEENTRY32);
		} while (Module32Next(hTH, &me));
	}
    CloseToolhelp32Snapshot(hTH);
}
bool DumpHFSD(HANDLE hProc, void *p)
{
    // 48 byte struct.
    struct filedesc {
        DWORD dw1[2];
        DWORD pPartition;
        DWORD dw2;
        DWORD pfileinfo;
        DWORD dw3[2];
        DWORD pgtgtinfo;
        DWORD dw4[4];
    } fd;
    DWORD nRead;
    if (!ReadProcessMemory(hProc,  (void*)p, &fd, sizeof(fd), &nRead)) {
        error("reading filesys.exe : %08lx\n", p);
        return false;
    }
    debug(" fileinfo: part=%08lx  n=%d", fd.pPartition, fd.dw2);

    struct fileinfo {
        DWORD pfninfo;       // points to { dw[16], name }
        DWORD dw1[7];
    } fi;
    if (!ReadProcessMemory(hProc,  (void*)fd.pfileinfo, &fi, sizeof(fi), &nRead)) {
        error("reading filesys.exe : %08lx\n", fd.pfileinfo);
        return false;
    }
    struct fninfo {
        DWORD dw[16];
        WCHAR name[260];
    } fni;
    if (!ReadProcessMemory(hProc,  (void*)fi.pfninfo, &fni, sizeof(fni), &nRead)) {
        error("reading filesys.exe : %08lx\n", fi.pfninfo);
        return false;
    }
    debug("\nfninfo(%08lx):%ls\n", fi.pfninfo, fni.name);

    struct gtgtinfo {
        DWORD magic;
        DWORD pdgdginfo;
        DWORD dw2;
        DWORD pwstrName;
        DWORD pNext;
        DWORD pPrev;
        DWORD dw6;
        DWORD dw7;
    } gtgt;
    if (!ReadProcessMemory(hProc,  (void*)fd.pgtgtinfo, &gtgt, sizeof(gtgt), &nRead)) {
        error("reading filesys.exe : %08lx\n", fd.pgtgtinfo);
        return false;
    }
    debug("gtgt(%08lx): n=%08lx p=%08lx  dgdg=%08lx  %hs\n", fd.pgtgtinfo, gtgt.pNext, gtgt.pPrev, gtgt.pdgdginfo,
            procstring(hProc, gtgt.pwstrName).c_str());

    return true;
}

bool DumpSTRG(HANDLE hProc, void *p)
{
    DWORD nRead;
    struct storagedesc {
        DWORD magic;        // 0x00596164 -> store handle,  0x64615900 -> partition handle
        DWORD pstorageinfo;
        DWORD ppartitioninfo;
        DWORD dw1[5];
        DWORD pdw1[2];
    } sd;
    if (!ReadProcessMemory(hProc,  (void*)p, &sd, sizeof(sd), &nRead)) {
        error("reading filesys.exe : %08lx\n", p);
        return false;
    }
    struct storageinfo {
        DWORD dw1[3];
        WCHAR devname[16];
        WCHAR desc1[32];
        WCHAR desc2[32];
        WCHAR partdll[260];
        WCHAR profilekey[260];
        WCHAR fsname[260];
        DWORD dw2[165];
        WCHAR desc3[32];
        DWORD dw3[29];
        WCHAR driverkey[260];
        DWORD dw4[11];
    } si;
    if (!ReadProcessMemory(hProc,  (void*)sd.pstorageinfo, &si, sizeof(si), &nRead)) {
        error("reading filesys.exe : %08lx\n", sd.pstorageinfo);
        return false;
    }
    debug(" storagedev %ls %ls %ls", si.devname, si.desc1, si.desc2);

    return true;
}
bool DumpFFSD(HANDLE hProc, void *p)
{
    // 48 byte struct.
    struct filedesc {
        DWORD dw1[2];
        DWORD pPartition;
        DWORD dw2;
        DWORD pfilenameinfo;
        DWORD dw3[2];
        DWORD pgtgtinfo;
        DWORD dw4[4];
    } fd;
    DWORD nRead;
    if (!ReadProcessMemory(hProc,  (void*)p, &fd, sizeof(fd), &nRead)) {
        error("reading filesys.exe : %08lx\n", p);
        return false;
    }
    debug(" fileinfo: part=%08lx  n=%d", fd.pPartition, fd.dw2);

    struct filenameinfo {
        DWORD dw;
        WCHAR name[260];
    } fi;
    if (!ReadProcessMemory(hProc,  (void*)fd.pfilenameinfo, &fi, sizeof(fi), &nRead)) {
        error("reading filesys.exe : %08lx\n", fd.pfilenameinfo);
        return false;
    }
    debug(" %08lx  %ls", fi.dw, fi.name);
    return true;
}
bool DumpPartition(HANDLE hProc, void *p)
{
    struct partitioninfo {
        DWORD dw1[8];
        WCHAR name[66];
        DWORD dw2[2];
    } pi;
    DWORD nRead;
    if (!ReadProcessMemory(hProc,  (void*)p, &pi, sizeof(pi), &nRead)) {
        error("reading filesys.exe : %08lx\n", p);
        return false;
    }
    debug(" partition: %ls", pi.name);
    return true;
}
bool DumpBlockDevice(HANDLE hProc, void *p)
{
    struct blockdev {
        DWORD dw1[11];
        WCHAR name1[256];
        DWORD dw2[13];
        WCHAR name2[28];
        DWORD dw3[2];
    } bdev;
    DWORD nRead;
    if (!ReadProcessMemory(hProc,  (void*)p, &bdev, sizeof(bdev), &nRead)) {
        error("reading filesys.exe : %08lx\n", p);
        return false;
    }
    debug(" blockdev: %ls  %ls", bdev.name1, bdev.name2);
    return true;
}
bool DumpW32Device(HANDLE hProc, fsopendev_t *d)
{
    fsopendev_t odev;
    DWORD nRead;
	// W32D struct is 0x28 bytes
	// this handle can only be read from the context of device.exe
    if (!ReadProcessMemory(hProc,  (void*)d, &odev, sizeof(odev), &nRead)) {
        error("reading device.exe : %08lx\n", d);
        return false;
    }
    debug(" opendev next=%08lx data=%08lx dev=%08lx devref=%08lx->(%s) openref=%08lx  h=%08lx hproc=%08lx",
            odev.nextptr, odev.dwOpenData, odev.lpDev, 
            odev.lpdwDevRefCnt, dwptrvaluestr(hProc, (DWORD)odev.lpdwDevRefCnt).c_str(), 
            odev.dwOpenRefCnt, odev.KHandle, odev.hProc);

    fsdev_t dev;
    if (!ReadProcessMemory(hProc,  (void*)odev.lpDev, &dev, sizeof(dev), &nRead)) {
        error("reading device.exe : %08lx\n", d);
        return false;
    }
    debug("   %c%c%c%d:", dev.type[0], dev.type[1], dev.type[2], dev.index);
    return true;
}

std::string FindNameForOid(CEOID oid);
bool GetNameForOid(CEOID oid, std::string& name)
{
    CEOIDINFO info;
    if (!CeOidGetInfo(oid, &info))
    {
        return false;
    }
    switch(info.wObjType)
    {
        case OBJTYPE_INVALID: name="invalid objectid"; break;
        case OBJTYPE_FILE: name=stringformat("file: %ls", info.infFile.szFileName); break;
        case OBJTYPE_DIRECTORY: name=stringformat("dir: %ls", info.infDirectory.szDirName); break;
        case OBJTYPE_DATABASE: name=stringformat("db: %ls", info.infDatabase.szDbaseName); break;
        case OBJTYPE_RECORD: name=stringformat("rec from %hs", FindNameForOid(info.infRecord.oidParent).c_str()); break;
        default: name=stringformat("unknown objtype %d", info.wObjType); 
    }
    return true;
}
std::string FindNameForOid(CEOID oid)
{
    oid &= 0xffffff;
    std::string name;
    for (int i=0 ; i<768 ; i++)
    {
        //  upto 768, since CeOidGetInfo crashes when called with oid's over 0xe0000000
        if (GetNameForOid(oid | i<<22, name))
            return name;
    }

    return "not found";
}

bool DumpW32File(HANDLE hProc, FHANDLE *f)
{
    struct {
        DWORD d[6];
    } fh;
    DWORD nRead;
	// W32H struct is 0x18 bytes
	// this handle can only be read from the context of filesys.exe
    if (!ReadProcessMemory(hProc,  (void*)f, &fh, sizeof(fh), &nRead)) {
        error("reading filesys.exe : %08lx\n", f);
        return false;
    }
    debug(" %hs", hexdump((BYTE*)&fh, sizeof(fh)/sizeof(DWORD), 4).c_str());
    debug("  %hs", FindNameForOid(fh.d[3]).c_str());

    return true;
}

#define DBHANDLE void
bool DumpDatabase(HANDLE hProc, DBHANDLE *d)
{
    struct {
        DWORD d[18];
    } dbh;
    DWORD nRead;
	// DBOA struct is 0x48 bytes
	// this handle can only be read from the context of filesys.exe
    if (!ReadProcessMemory(hProc,  (void*)d, &dbh, sizeof(dbh), &nRead)) {
        error("reading filesys.exe : %08lx\n", d);
        return false;
    }
    debug(" %hs", hexdump((BYTE*)&dbh, sizeof(dbh)/sizeof(DWORD), 4).c_str());
    debug("  %hs", FindNameForOid(dbh.d[5]).c_str());
    return true;
}
bool DumpSocket(HANDLE hProc, DWORD s)
{
	// Sock struct is ? bytes
	// this handle can only be read from the context of device.exe
    struct socketdesc {
        DWORD dw1[10];
        DWORD pnext;
        DWORD dw2[2];
        DWORD ptr1;
        DWORD pproto;
        DWORD dw3[44];
        DWORD ptr2;
        DWORD dw4[1];
        DWORD ptr3;
        DWORD ptr4;
        DWORD dw5[1];
        DWORD ptr5;
        DWORD dw6[1];
        DWORD ptr6;
        DWORD ptr7;
        DWORD dw7[6];
        DWORD ptr8;
        DWORD dw8[5];
        DWORD ptr9;
        DWORD ptr10;
        DWORD ptr11;
        DWORD dw9[2];
        DWORD ptr12;
        DWORD ptr13;
        DWORD ptr14;
        DWORD dw10[43];
        DWORD ptr15;
    } sd;
    DWORD nRead;
    if (!ReadProcessMemory(hProc,  (void*)s, &sd, sizeof(sd), &nRead)) {
        error("reading device.exe : %08lx\n", s);
        return false;
    }
    struct socketproto {
        DWORD dw1[7];
        WCHAR name[20];
        DWORD dw2[4];
    } sp;
    if (!ReadProcessMemory(hProc,  (void*)sd.pproto, &sp, sizeof(sp), &nRead)) {
        error("reading device.exe : %08lx\n", sd.pproto);
        return false;
    }
    debug(" sock: next=%08lx proto=%ls", sd.pnext, sp.name);
/*
    dumpmem(hProc, sd.ptr1,  0x40, "socketmem ptr1");
    dumpmem(hProc, sd.ptr2,  0x40, "socketmem ptr2");
    dumpmem(hProc, sd.ptr3,  0x40, "socketmem ptr3");
    dumpmem(hProc, sd.ptr4,  0x40, "socketmem ptr4");
    dumpmem(hProc, sd.ptr5,  0x40, "socketmem ptr5");
    dumpmem(hProc, sd.ptr6,  0x40, "socketmem ptr6");
    dumpmem(hProc, sd.ptr7,  0x40, "socketmem ptr7");
    dumpmem(hProc, sd.ptr8,  0x40, "socketmem ptr8");
    dumpmem(hProc, sd.ptr9,  0x40, "socketmem ptr9");
    dumpmem(hProc, sd.ptr10, 0x40, "socketmem ptr10");
    dumpmem(hProc, sd.ptr11, 0x40, "socketmem ptr11");
    dumpmem(hProc, sd.ptr12, 0x40, "socketmem ptr12");
    dumpmem(hProc, sd.ptr13, 0x40, "socketmem ptr13");
    dumpmem(hProc, sd.ptr14, 0x40, "socketmem ptr14");
    dumpmem(hProc, sd.ptr15, 0x40, "socketmem ptr15");
*/
    return true;

}
bool isStartPtr(HDATA *ha) {
    return ha->pci==(const CINFO*)0x8005d0a8 || ha->pci==(const CINFO*)0x8004ba18;
}
/*
 *   ha+00    8d94551c linkage.fwd
 *   ha+04    8dddfdd4 linkage.back
 *   ha+08    8d945542 hValue
 *   ha+0c    00040000 lock
 *   ha+10    0000     ref.pFr
 *   ha+12    0001     ref.count
 *   ha+14    8df5ba78 pci
 *     pci+00    6b636f53  acName[4];
 *     pci+04    04        disp;     
 *     pci+05    0b        type;     
 *     pci+06    000f      cMethods; 
 *     pci+08    01c910d8  ppfnMethods
 *     pci+0c    01c91118  pdwSig
 *     pci+10    8c0c3840  pServer
 *   ha+18    00361868 pvObj
 *   ha+1c    00000000 dwInfo
 */
void testhandle(HANDLE h)
{
	HDATA *hi= cvHandle2HDataPtr(h);

	HDATA *ha;

	ha=hi;
	do {

        if (isStartPtr(ha))
        {
            debug("Start of Handle list: %08lx\n", ha);
            ha= (HDATA*)ha->linkage.fwd;
            continue;
        }
		debug("%-4s  %08lx %08lx", GetCInfoName(ha->pci), ha->hValue, ha->pvObj);

		RegisterCInfo(ha->pci);

        HANDLE hProc= ha->pci->pServer? ha->pci->pServer->hProc : NULL;

		if (ha->pvObj && strncmp(ha->pci->acName, "PROC", 4)==0)
			DumpProcess(hProc, (PROCESS*)ha->pvObj);
		else if (ha->pvObj && strncmp(ha->pci->acName, "THRD", 4)==0)
			DumpThread(hProc, (THREAD*)ha->pvObj);
		else if (ha->pvObj && strncmp(ha->pci->acName, "APIS", 4)==0)
			DumpApiSet(hProc, (APISET*)ha->pvObj);
		else if (ha->pvObj && strncmp(ha->pci->acName, "EVNT", 4)==0)
			DumpEvent(hProc, (EVENT*)ha->pvObj);
		else if (ha->pvObj && strncmp(ha->pci->acName, "MUTX", 4)==0)
			DumpMutex(hProc, (MUTEX*)ha->pvObj);
		else if (ha->pvObj && strncmp(ha->pci->acName, "SEMP", 4)==0)
			DumpSemaphore(hProc, (SEMAPHORE*)ha->pvObj);
		else if (ha->pvObj && strncmp(ha->pci->acName, "FMAP", 4)==0)
			DumpFSMap(hProc, (FSMAP*)ha->pvObj);
		else if (ha->pvObj && strncmp(ha->pci->acName, "W32D", 4)==0)
			DumpW32Device(hProc, (fsopendev_t*)ha->pvObj);
		else if (ha->pvObj && strncmp(ha->pci->acName, "W32H", 4)==0)
			DumpW32File(hProc, (FHANDLE*)ha->pvObj);
		else if (ha->pvObj && strncmp(ha->pci->acName, "DBOA", 4)==0)
			DumpDatabase(hProc, (DBHANDLE*)ha->pvObj);
		else if (ha->pvObj && strncmp(ha->pci->acName, "Sock", 4)==0)
			DumpSocket(hProc, (DWORD)ha->pvObj);
		else if (ha->pvObj && strncmp(ha->pci->acName, "BDEV", 4)==0)
			DumpBlockDevice(hProc, ha->pvObj);
		else if (ha->pvObj && strncmp(ha->pci->acName, "PFSD", 4)==0)
			DumpPartition(hProc, ha->pvObj);
		else if (ha->pvObj && strncmp(ha->pci->acName, "HFSD", 4)==0)
			DumpHFSD(hProc, ha->pvObj);
		else if (ha->pvObj && strncmp(ha->pci->acName, "FFSD", 4)==0)
			DumpFFSD(hProc, ha->pvObj);
		else if (ha->pvObj && strncmp(ha->pci->acName, "STRG", 4)==0)
			DumpSTRG(hProc, ha->pvObj);
		//PFAT  { +0x2c : char[11],   (void*)+0x00 : +0x38 : long, wchar[?] }
		//PRdr  { +0x10 : wchar[12] }

        debug("\n");
		ha= (HDATA*)ha->linkage.fwd;
	} while (ha!=hi && ha!=NULL);
	debug("\n");
}

void DumpMemoryBlock(int j, MEMBLOCK *m)
{
	debug("    M%03x: %08lx -> %08lx: %08lx %02x %02x %04x %04x %04x : ",
		j, m,
		m->alk, m->cUses, m->flags, (WORD)m->ixBase, (WORD)m->hPf, (WORD)m->cLocks);
	for (int i=0 ; i<16 ; i++)
	{
		debug(" %08lx", m->aPages[i]);
	}
	debug("\n");
}
void DumpSection(int i, SECTION *s)
{
	debug("S%02x: %08lx\n", i, s);
	for (int j=0 ; j<0x200 ; j++)
	{
		// weird, this typecast should not be nescesary
		DumpMemoryBlock(j, (MEMBLOCK*)s[j]);
	}
}
void DumpKernelInfo()
{
	struct KDataStruct *k= (struct KDataStruct *)0xffffc800;

int i;
debug("%-64s %08lx\n", "0x000 Current thread local storage pointer", k->lpvTls);
debug("%-64s SH_WIN32     %08lx\n", "", k->ahSys[SH_WIN32    ]);
debug("%-64s SH_CURTHREAD %08lx\n", "", k->ahSys[SH_CURTHREAD]);
debug("%-64s SH_CURPROC   %08lx\n", "", k->ahSys[SH_CURPROC  ]);
debug("ahSys  ");
    for (i=SH_CURPROC+1 ; i<NUM_SYS_HANDLES ; i++) debug(" %08lx", k->ahSys[i]);
debug("\n");

debug("%-64s %d\n", "0x084 reschedule flag", k->bResched);
debug("%-64s %d\n", "0x085 kernel exception nesting", k->cNest);
debug("%-64s %d\n", "0x086 TRUE during power off processing", k->bPowerOff);
debug("%-64s %d\n", "0x087 TRUE if profiling enabled", k->bProfileOn);
#if _WIN32_WCE==300
debug("%-64s %08lx\n", "0x088 Page Table Descriptor", k->ptDesc);
#endif
debug("%-64s %08lx\n", "0x08c was DiffMSec", k->rsvd2);
debug("%-64s %08lx\n", "0x090 ptr to current PROCESS struct", k->pCurPrc);
debug("%-64s %08lx\n", "0x094 ptr to current THREAD struct", k->pCurThd);
debug("%-64s %08lx\n", "0x098 ", k->dwKCRes);
debug("%-64s %08lx\n", "0x09c handle table base address", k->handleBase);
debug("%-64s ", "0x0a0 section table for virtual memory");
    for (i=0 ; i<64 ; i++) debug(" %08lx", k->aSections[i]);
debug("\n");

debug("%-64s ", "0x1a0, 0x220 device events ");
    for (i=0 ; i<SYSINTR_MAX_DEVICES ; i++) debug(" %08lx:%08lx", k->alpvIntrData[i], k->alpeIntrEvents[i]);
debug("\n");

debug("%-64s %08lx\n", "0x2a0 direct API return address for kernel mode", k->pAPIReturn);
debug("%-64s %08lx\n", "0x2a4 ptr to MemoryMap array", k->pMap);
debug("%-64s %08lx\n", "0x2a8 !0 when in debugger", k->dwInDebugger);

//debug("%-64s long\n", "0x2ac - padding", k->alPad[21]);
debug("%-64s %08lx\n", "aInfo[KINX_PROCARRAY]: address of process array", k->aInfo[KINX_PROCARRAY]);
debug("%-64s %08lx\n", "aInfo[KINX_PAGESIZE]: system page size", k->aInfo[KINX_PAGESIZE]);
debug("%-64s %08lx\n", "aInfo[KINX_PFN_SHIFT]: shift for page # in PTE", k->aInfo[KINX_PFN_SHIFT]);
debug("%-64s %08lx\n", "aInfo[KINX_PFN_MASK]: mask for page # in PTE", k->aInfo[KINX_PFN_MASK]);
debug("%-64s %08lx\n", "aInfo[KINX_PAGEFREE]: # of free physical pages", k->aInfo[KINX_PAGEFREE]);
debug("%-64s %08lx\n", "aInfo[KINX_SYSPAGES]: # of pages used by kernel", k->aInfo[KINX_SYSPAGES]);
debug("%-64s %08lx\n", "aInfo[KINX_KHEAP]: ptr to kernel heap array", k->aInfo[KINX_KHEAP]);
debug("%-64s %08lx\n", "aInfo[KINX_SECTIONS]: ptr to SectionTable array", k->aInfo[KINX_SECTIONS]);
debug("%-64s %08lx\n", "aInfo[KINX_MEMINFO]: ptr to system MemoryInfo struct", k->aInfo[KINX_MEMINFO]);
debug("%-64s %08lx\n", "aInfo[KINX_MODULES]: ptr to module list", k->aInfo[KINX_MODULES]);
debug("%-64s %08lx\n", "aInfo[KINX_DLL_LOW]: lower bound of DLL shared space", k->aInfo[KINX_DLL_LOW]);
debug("%-64s %08lx\n", "aInfo[KINX_NUMPAGES]: total # of RAM pages", k->aInfo[KINX_NUMPAGES]);
debug("%-64s %08lx\n", "aInfo[KINX_PTOC]: ptr to ROM table of contents", k->aInfo[KINX_PTOC]);
debug("%-64s %08lx\n", "aInfo[KINX_KDATA_ADDR]: kernel mode version of KData", k->aInfo[KINX_KDATA_ADDR]);
debug("%-64s %08lx\n", "aInfo[KINX_GWESHEAPINFO]: Current amount of gwes heap in use", k->aInfo[KINX_GWESHEAPINFO]);
debug("%-64s %08lx\n", "aInfo[KINX_TIMEZONEBIAS]: Fast timezone bias info", k->aInfo[KINX_TIMEZONEBIAS]);
debug("%-64s %08lx\n", "aInfo[KINX_PENDEVENTS]: bit mask for pending interrupt events", k->aInfo[KINX_PENDEVENTS]);
debug("%-64s %08lx\n", "aInfo[KINX_KERNRESERVE]: number of kernel reserved pages", k->aInfo[KINX_KERNRESERVE]);
debug("%-64s %08lx\n", "aInfo[KINX_API_MASK]: bit mask for registered api sets", k->aInfo[KINX_API_MASK]);

#if _WIN32_WCE==300
debug("%-64s %08lx\n", "aInfo[KINX_NLS_OCP]: Current OEM code page", k->aInfo[KINX_NLS_OCP]);
debug("%-64s %08lx\n", "aInfo[KINX_NLS_ACP]: Current ANSI code page", k->aInfo[KINX_NLS_ACP]);
debug("%-64s %08lx\n", "aInfo[KINX_NLS_LOC]: Current NLS locale", k->aInfo[KINX_NLS_LOC]);
#endif

debug("%-64s %08lx\n", "aInfo[KINX_HEAP_WASTE]: Kernel heap wasted space", k->aInfo[KINX_HEAP_WASTE]);
debug("%-64s %08lx\n", "aInfo[KINX_DEBUGGER]: For use by debugger for protocol communication", k->aInfo[KINX_DEBUGGER]);
debug("%-64s %08lx\n", "aInfo[KINX_APISETS]: APIset pointers", k->aInfo[KINX_APISETS]);

CINFO **apisets= (CINFO**)k->aInfo[KINX_APISETS];
for (i=0 ; i<32 ; i++)
	if (apisets[i])
		RegisterCInfo(apisets[i]);
debug("k->aInfo  ");
    for (i=KINX_APISETS+1 ; i<32 ; i++) debug(" %08lx", k->aInfo[i]);
debug("\n");

}
int WINAPI WinMain(	HINSTANCE hInstance,
					HINSTANCE hPrevInstance,
					LPTSTR    lpCmdLine,
					int       nCmdShow)
{
    DebugSetLogfile("\\testpi.log");
    DebugOutputDebugString();

    // create some open handles
    HANDLE hFile= CreateFile(L"\\windows\\spcore.exe", GENERIC_READ, 0,NULL, OPEN_EXISTING, 0, NULL);
    HKEY hkey;
    DWORD disp;
    LONG res= RegCreateKeyEx(HKEY_LOCAL_MACHINE, L"Software", 0, L"??", 0, KEY_WRITE, NULL, &hkey, &disp);
    debug("filehandle=%08lx reghandle=%08lx\n", hFile, hkey);


    BOOL bMode = SetKMode(TRUE);
    DWORD dwPerm = SetProcPermissions(0xFFFFFFFF);

    //g_hFilesysProc= GetProcessHandle(L"filesys.exe");
    //g_hDeviceProc= GetProcessHandle(L"device.exe");

	DumpKernelInfo();

	testhandle(hInstance);

	DumpCInfoMap();

//	test();

    SetProcPermissions(dwPerm);
    SetKMode(bMode);

/*
	ThreadInfo ti(GetCurrentThread());

	DWORD tMax= 2048;

	ti.GetThreadInfo();

	for (DWORD iSleep= 0 ; iSleep < tMax ; iSleep = iSleep?2*iSleep:1)
	{
		DWORD t0= GetTickCount();

		while(GetTickCount()-t0 < tMax) Sleep(iSleep);

		ti.GetThreadInfo();

		debug("%6d : k%5d u%5d - k%5d u%5d\n", 
			iSleep,
			ti.m_promileKernelRecent, ti.m_promileKernelTotal, ti.m_promileUserRecent, ti.m_promileUserTotal);
	}
*/
	return 0;
}
