/* (C) 2003 XDA Developers  itsme@xs4all.nl
 *
 * Copyright 2001,2002,2003 NAH6 BV
 * All Rights Reserved
 *
 * $Header: /var/lib/cvs/xdautils/rilhook/RilClass.cpp,v 1.11 2004/09/24 13:19:01 itsme Exp $
 */

#include <windows.h>

#include "debug.h"
#include "stringutils.h"
#include "ril.h"
#include "RilClass.h"

extern RilClass *g_ril;

std::string GetLockedStateString(DWORD dwLockedState);
std::string GetCallTypeString(int calltype);
std::string GetDisconnectInitString(int di);
std::string GetServiceInfoString(RILSERVICEINFO *si);
std::string GetRingInfoString(RILRINGINFO* ri);
std::string GetConnectInfoString(RILCONNECTINFO* ci);
std::string GetCallInfoString(RILCALLINFO* ci);
void HandleCallControlRing(RILRINGINFO* ri, DWORD cbdata);
void HandleCallControlConnect(RILCONNECTINFO* ci, DWORD cbdata);
void HandleCallControlDisconnect(DWORD* d, DWORD cbdata);
void HandleCallControlDataServiceNegotiated(RILSERVICEINFO* si, DWORD cbdata);
std::string GetRilAddressString(RILADDRESS* ra);
std::string GetRilSubAddressString(RILSUBADDRESS* ra);
std::string GetRemotePartyInfoString(RILREMOTEPARTYINFO* pi);
void HandleSupServiceCallerId(RILREMOTEPARTYINFO* pi, DWORD cbdata);
void HandleSimtoolkitCommand(BYTE* lpData, DWORD cbdata);
std::string GetRegistrationStatusString(int rs);
void HandleNetworkGprsStatusChanged(DWORD* d, DWORD cbdata);
std::string GetRadioPresenceString(int rp);
std::string GetEquipmentStateString(RILEQUIPMENTSTATE* es);
std::string GetEquipmentInfoString(RILEQUIPMENTINFO* ei);
std::string GetRilSignalQuality(RILSIGNALQUALITY *sq);
void HandleRadioPresenceChanged(DWORD* d, DWORD cbdata);
std::string GetRilMessageString(RILMESSAGE* m);
void HandleMessageMessage(RILMESSAGE* m, DWORD cbdata);
void HandleCallStateChanged();


std::string RilResultString(HRESULT res)
{
    switch(res)
    {

case RIL_RESULT_OK:                               return "FUNCRESULT::OK";
case RIL_RESULT_NOCARRIER:                        return "FUNCRESULT::NOCARRIER";
case RIL_RESULT_ERROR:                            return "FUNCRESULT::ERROR";
case RIL_RESULT_NODIALTONE:                       return "FUNCRESULT::NODIALTONE";
case RIL_RESULT_BUSY:                             return "FUNCRESULT::BUSY";
case RIL_RESULT_NOANSWER:                         return "FUNCRESULT::NOANSWER";
case RIL_RESULT_CALLABORTED:                      return "FUNCRESULT::CALLABORTED";

case RIL_NOTIFY_RING:                             return "CALLCTRL::RING";
case RIL_NOTIFY_CONNECT:                          return "CALLCTRL::CONNECT";
case RIL_NOTIFY_DISCONNECT:                       return "CALLCTRL::DISCONNECT";
case RIL_NOTIFY_DATASVCNEGOTIATED:                return "CALLCTRL::DATASVCNEGOTIATED";
case RIL_NOTIFY_CALLSTATECHANGED:                 return "CALLCTRL::CALLSTATECHANGED";
case RIL_NOTIFY_EMERGENCYMODEENTERED:             return "CALLCTRL::EMERGENCYMODEENTERED";
case RIL_NOTIFY_EMERGENCYMODEEXITED:              return "CALLCTRL::EMERGENCYMODEEXITED";
case RIL_NOTIFY_EMERGENCYHANGUP:                  return "CALLCTRL::EMERGENCYHANGUP";
case RIL_NOTIFY_HSCSDPARAMSNEGOTIATED:            return "CALLCTRL::HSCSDPARAMSNEGOTIATED";
case RIL_NOTIFY_DIAL:                             return "CALLCTRL::DIAL";
case RIL_NOTIFY_CALLPROGRESSINFO:                 return "CALLCTRL::CALLPROGRESSINFO";
case RIL_NOTIFY_CURRENTLINECHANGED:               return "CALLCTRL::CURRENTLINECHANGED";

case RIL_NOTIFY_MESSAGE:                          return "MESSAGE::MESSAGE";
case RIL_NOTIFY_BCMESSAGE:                        return "MESSAGE::BCMESSAGE";
case RIL_NOTIFY_STATUSMESSAGE:                    return "MESSAGE::STATUSMESSAGE";
case RIL_NOTIFY_MSGSTORED:                        return "MESSAGE::MSGSTORED";
case RIL_NOTIFY_MSGDELETED:                       return "MESSAGE::MSGDELETED";
case RIL_NOTIFY_MSGSTORAGECHANGED:                return "MESSAGE::MSGSTORAGECHANGED";
case RIL_NOTIFY_MESSAGE_IN_SIM:                   return "MESSAGE::MESSAGE_IN_SIM";
case RIL_NOTIFY_BCMESSAGE_IN_SIM:                 return "MESSAGE::BCMESSAGE_IN_SIM";
case RIL_NOTIFY_STATUSMESSAGE_IN_SIM:             return "MESSAGE::STATUSMESSAGE_IN_SIM";

case RIL_NOTIFY_REGSTATUSCHANGED:                 return "NETWORK::REGSTATUSCHANGED";
case RIL_NOTIFY_CALLMETER:                        return "NETWORK::CALLMETER";
case RIL_NOTIFY_CALLMETERMAXREACHED:              return "NETWORK::CALLMETERMAXREACHED";
case RIL_NOTIFY_GPRSREGSTATUSCHANGED:             return "NETWORK::GPRSREGSTATUSCHANGED";

case RIL_NOTIFY_CALLERID:                         return "SUPSERVICE::CALLERID";
case RIL_NOTIFY_DIALEDID:                         return "SUPSERVICE::DIALEDID";
case RIL_NOTIFY_CALLWAITING:                      return "SUPSERVICE::CALLWAITING";
case RIL_NOTIFY_SUPSERVICEDATA:                   return "SUPSERVICE::SUPSERVICEDATA";

case RIL_NOTIFY_PHONEBOOKENTRYSTORED:             return "PHONEBOOK::PHONEBOOKENTRYSTORED";
case RIL_NOTIFY_PHONEBOOKENTRYDELETED:            return "PHONEBOOK::PHONEBOOKENTRYDELETED";
case RIL_NOTIFY_PHONEBOOKSTORAGECHANGED:          return "PHONEBOOK::PHONEBOOKSTORAGECHANGED";

case RIL_NOTIFY_SIMTOOLKITCMD:                    return "SIMTOOLKIT::SIMTOOLKITCMD";
case RIL_NOTIFY_SIMTOOLKITCALLSETUP:              return "SIMTOOLKIT::SIMTOOLKITCALLSETUP";
case RIL_NOTIFY_SIMTOOLKITEVENT:                  return "SIMTOOLKIT::SIMTOOLKITEVENT";
case RIL_NOTIFY_SIMTOOLKITSESSIONEND:             return "SIMTOOLKIT::SIMTOOLKITSESSIONEND";

case RIL_NOTIFY_RADIOEQUIPMENTSTATECHANGED:       return "RADIOSTATE::RADIOEQUIPMENTSTATECHANGED";
case RIL_NOTIFY_RADIOPRESENCECHANGED:             return "RADIOSTATE::RADIOPRESENCECHANGED";

case RIL_NOTIFY_SIMNOTACCESSIBLE:                 return "MISC::SIMNOTACCESSIBLE";
case RIL_NOTIFY_DTMFSIGNAL:                       return "MISC::DTMFSIGNAL";
case RIL_NOTIFY_GPRSCLASS_NETWORKCHANGED:         return "MISC::GPRSCLASS_NETWORKCHANGED";
case RIL_NOTIFY_GPRSCLASS_RADIOCHANGED:           return "MISC::GPRSCLASS_RADIOCHANGED";
case RIL_NOTIFY_SIGNALQUALITY:                    return "MISC::SIGNALQUALITY";

case RIL_NOTIFY_CELLID:                           return "DEVSPECIFIC::CELLID";

default:
	if ((res&0xffff0000) == RIL_NCLASS_DEVSPECIFIC)
		return stringformat("DEVSPECIFIC::%04x", res&0xffff);
	else
		return stringformat("RIL_%04x::%04x", (res>>16)&0xffff, res&0xffff);
    }
}

std::string RilErrorClassString(DWORD dwErrorClass)
{
    switch(dwErrorClass)
    {
case RIL_ERRORCLASS_NONE: return "RIL_ERRORCLASS_NONE";
case RIL_ERRORCLASS_PASSWORD: return "RIL_ERRORCLASS_PASSWORD";
case RIL_ERRORCLASS_SIM: return "RIL_ERRORCLASS_SIM";
case RIL_ERRORCLASS_NETWORKACCESS: return "RIL_ERRORCLASS_NETWORKACCESS";
case RIL_ERRORCLASS_NETWORK: return "RIL_ERRORCLASS_NETWORK";
case RIL_ERRORCLASS_MOBILE: return "RIL_ERRORCLASS_MOBILE";
case RIL_ERRORCLASS_NETWORKUNSUPPORTED: return "RIL_ERRORCLASS_NETWORKUNSUPPORTED";
case RIL_ERRORCLASS_MOBILEUNSUPPORTED: return "RIL_ERRORCLASS_MOBILEUNSUPPORTED";
case RIL_ERRORCLASS_BADPARAM: return "RIL_ERRORCLASS_BADPARAM";
case RIL_ERRORCLASS_STORAGE: return "RIL_ERRORCLASS_STORAGE";
case RIL_ERRORCLASS_SMSC: return "RIL_ERRORCLASS_SMSC";
case RIL_ERRORCLASS_DESTINATION: return "RIL_ERRORCLASS_DESTINATION";
case RIL_ERRORCLASS_DESTINATIONUNSUPPORTED: return "RIL_ERRORCLASS_DESTINATIONUNSUPPORTED";
case RIL_ERRORCLASS_RADIOUNAVAILABLE: return "RIL_ERRORCLASS_RADIOUNAVAILABLE";

default:
    return stringformat("RIL_ERRORCLASS_%02x", dwErrorClass);
    }
}


std::string RilErrorString(DWORD dwError)
{
    switch(dwError)
    {
case RIL_E_PHONEFAILURE:              return "RIL_E_MOBILE::PHONEFAILURE";
case RIL_E_NOCONNECTION:              return "RIL_E_MOBILE::NOCONNECTION";
case RIL_E_LINKRESERVED:              return "RIL_E_MOBILE::LINKRESERVED";
case RIL_E_OPNOTALLOWED:              return "RIL_E_MOBILEUNSUPPORTED::OPNOTALLOWED";
case RIL_E_OPNOTSUPPORTED:            return "RIL_E_MOBILEUNSUPPORTED::OPNOTSUPPORTED";
case RIL_E_PHSIMPINREQUIRED:          return "RIL_E_PASSWORD::PHSIMPINREQUIRED";
case RIL_E_PHFSIMPINREQUIRED:         return "RIL_E_PASSWORD::PHFSIMPINREQUIRED";
case RIL_E_PHFSIMPUKREQUIRED:         return "RIL_E_PASSWORD::PHFSIMPUKREQUIRED";
case RIL_E_SIMNOTINSERTED:            return "RIL_E_SIM::SIMNOTINSERTED";
case RIL_E_SIMPINREQUIRED:            return "RIL_E_PASSWORD::SIMPINREQUIRED";
case RIL_E_SIMPUKREQUIRED:            return "RIL_E_PASSWORD::SIMPUKREQUIRED";
case RIL_E_SIMFAILURE:                return "RIL_E_SIM::SIMFAILURE";
case RIL_E_SIMBUSY:                   return "RIL_E_SIM::SIMBUSY";
case RIL_E_SIMWRONG:                  return "RIL_E_SIM::SIMWRONG";
case RIL_E_INCORRECTPASSWORD:         return "RIL_E_PASSWORD::INCORRECTPASSWORD";
case RIL_E_SIMPIN2REQUIRED:           return "RIL_E_PASSWORD::SIMPIN2REQUIRED";
case RIL_E_SIMPUK2REQUIRED:           return "RIL_E_PASSWORD::SIMPUK2REQUIRED";
case RIL_E_MEMORYFULL:                return "RIL_E_STORAGE::MEMORYFULL";
case RIL_E_INVALIDINDEX:              return "RIL_E_STORAGE::INVALIDINDEX";
case RIL_E_NOTFOUND:                  return "RIL_E_STORAGE::NOTFOUND";
case RIL_E_MEMORYFAILURE:             return "RIL_E_STORAGE::MEMORYFAILURE";
case RIL_E_TEXTSTRINGTOOLONG:         return "RIL_E_BADPARAM::TEXTSTRINGTOOLONG";
case RIL_E_INVALIDTEXTSTRING:         return "RIL_E_BADPARAM::INVALIDTEXTSTRING";
case RIL_E_DIALSTRINGTOOLONG:         return "RIL_E_BADPARAM::DIALSTRINGTOOLONG";
case RIL_E_INVALIDDIALSTRING:         return "RIL_E_BADPARAM::INVALIDDIALSTRING";
case RIL_E_NONETWORKSVC:              return "RIL_E_NETWORKACCESS::NONETWORKSVC";
case RIL_E_NETWORKTIMEOUT:            return "RIL_E_NETWORK::NETWORKTIMEOUT";
case RIL_E_EMERGENCYONLY:             return "RIL_E_NETWORKACCESS::EMERGENCYONLY";
case RIL_E_NETWKPINREQUIRED:          return "RIL_E_PASSWORD::NETWKPINREQUIRED";
case RIL_E_NETWKPUKREQUIRED:          return "RIL_E_PASSWORD::NETWKPUKREQUIRED";
case RIL_E_SUBSETPINREQUIRED:         return "RIL_E_PASSWORD::SUBSETPINREQUIRED";
case RIL_E_SUBSETPUKREQUIRED:         return "RIL_E_PASSWORD::SUBSETPUKREQUIRED";
case RIL_E_SVCPINREQUIRED:            return "RIL_E_PASSWORD::SVCPINREQUIRED";
case RIL_E_SVCPUKREQUIRED:            return "RIL_E_PASSWORD::SVCPUKREQUIRED";
case RIL_E_CORPPINREQUIRED:           return "RIL_E_PASSWORD::CORPPINREQUIRED";
case RIL_E_CORPPUKREQUIRED:           return "RIL_E_PASSWORD::CORPPUKREQUIRED";
case RIL_E_TELEMATICIWUNSUPPORTED:    return "RIL_E_NETWORKUNSUPPORTED::TELEMATICIWUNSUPPORT";
case RIL_E_SMTYPE0UNSUPPORTED:        return "RIL_E_SMSC::SMTYPE0UNSUPPORTED";
case RIL_E_CANTREPLACEMSG:            return "RIL_E_SMSC::CANTREPLACEMSG";
case RIL_E_PROTOCOLIDERROR:           return "RIL_E_SMSC::PROTOCOLIDERROR";
case RIL_E_DCSUNSUPPORTED:            return "RIL_E_SMSC::DCSUNSUPPORTED";
case RIL_E_MSGCLASSUNSUPPORTED:       return "RIL_E_SMSC::MSGCLASSUNSUPPORTED";
case RIL_E_DCSERROR:                  return "RIL_E_SMSC::DCSERROR";
case RIL_E_CMDCANTBEACTIONED:         return "RIL_E_SMSC::CMDCANTBEACTIONED";
case RIL_E_CMDUNSUPPORTED:            return "RIL_E_SMSC::CMDUNSUPPORTED";
case RIL_E_CMDERROR:                  return "RIL_E_SMSC::CMDERROR";
case RIL_E_MSGBODYHEADERERROR:        return "RIL_E_SMSC::MSGBODYHEADERERROR";
case RIL_E_SCBUSY:                    return "RIL_E_SMSC::SCBUSY";
case RIL_E_NOSCSUBSCRIPTION:          return "RIL_E_SMSC::NOSCSUBSCRIPTION";
case RIL_E_SCSYSTEMFAILURE:           return "RIL_E_SMSC::SCSYSTEMFAILURE";
case RIL_E_INVALIDADDRESS:            return "RIL_E_SMSC::INVALIDADDRESS";
case RIL_E_DESTINATIONBARRED:         return "RIL_E_SMSC::DESTINATIONBARRED";
case RIL_E_REJECTEDDUPLICATE:         return "RIL_E_SMSC::REJECTEDDUPLICATE";
case RIL_E_VPFUNSUPPORTED:            return "RIL_E_SMSC::VPFUNSUPPORTED";
case RIL_E_VPUNSUPPORTED:             return "RIL_E_SMSC::VPUNSUPPORTED";
case RIL_E_SIMMSGSTORAGEFULL:         return "RIL_E_STORAGE::SIMMSGSTORAGEFULL";
case RIL_E_NOSIMMSGSTORAGE:           return "RIL_E_SIM::NOSIMMSGSTORAGE";
case RIL_E_SIMTOOLKITBUSY:            return "RIL_E_SIM::SIMTOOLKITBUSY";
case RIL_E_SIMDOWNLOADERROR:          return "RIL_E_SIM::SIMDOWNLOADERROR";
case RIL_E_MSGSVCRESERVED:            return "RIL_E_NETWORKUNSUPPORTED::MSGSVCRESERVED";
case RIL_E_INVALIDMSGPARAM:           return "RIL_E_BADPARAM::INVALIDMSGPARAM";
case RIL_E_UNKNOWNSCADDRESS:          return "RIL_E_SMSC::UNKNOWNSCADDRESS";
case RIL_E_UNASSIGNEDNUMBER:          return "RIL_E_DESTINATION::UNASSIGNEDNUMBER";
case RIL_E_MSGBARREDBYOPERATOR:       return "RIL_E_NETWORKACCESS::MSGBARREDBYOPERATOR";
case RIL_E_MSGCALLBARRED:             return "RIL_E_NETWORKACCESS::MSGCALLBARRED";
case RIL_E_MSGXFERREJECTED:           return "RIL_E_DESTINATION::MSGXFERREJECTED";
case RIL_E_DESTINATIONOUTOFSVC:       return "RIL_E_DESTINATION::DESTINATIONOUTOFSVC";
case RIL_E_UNIDENTIFIEDSUBCRIBER:     return "RIL_E_NETWORKACCESS::UNIDENTIFIEDSUBCRIBE";
case RIL_E_SVCUNSUPPORTED:            return "RIL_E_NETWORKUNSUPPORTED::SVCUNSUPPORTED";
case RIL_E_UNKNOWNSUBSCRIBER:         return "RIL_E_NETWORKACCESS::UNKNOWNSUBSCRIBER";
case RIL_E_NETWKOUTOFORDER:           return "RIL_E_NETWORK::NETWKOUTOFORDER";
case RIL_E_NETWKTEMPFAILURE:          return "RIL_E_NETWORK::NETWKTEMPFAILURE";
case RIL_E_CONGESTION:                return "RIL_E_NETWORK::CONGESTION";
case RIL_E_RESOURCESUNAVAILABLE:      return "RIL_E_NONE::RESOURCESUNAVAILABLE";
case RIL_E_SVCNOTSUBSCRIBED:          return "RIL_E_NETWORKUNSUPPORTED::SVCNOTSUBSCRIBED";
case RIL_E_SVCNOTIMPLEMENTED:         return "RIL_E_NETWORKUNSUPPORTED::SVCNOTIMPLEMENTED";
case RIL_E_INVALIDMSGREFERENCE:       return "RIL_E_BADPARAM::INVALIDMSGREFERENCE";
case RIL_E_INVALIDMSG:                return "RIL_E_BADPARAM::INVALIDMSG";
case RIL_E_INVALIDMANDATORYINFO:      return "RIL_E_BADPARAM::INVALIDMANDATORYINFO";
case RIL_E_MSGTYPEUNSUPPORTED:        return "RIL_E_NETWORKUNSUPPORTED::MSGTYPEUNSUPPORTED";
case RIL_E_ICOMPATIBLEMSG:            return "RIL_E_NETWORKUNSUPPORTED::ICOMPATIBLEMSG";
case RIL_E_INFOELEMENTUNSUPPORTED:    return "RIL_E_NETWORKUNSUPPORTED::INFOELEMENTUNSUPPORT";
case RIL_E_PROTOCOLERROR:             return "RIL_E_NETWORK::PROTOCOLERROR";
case RIL_E_NETWORKERROR:              return "RIL_E_NETWORK::NETWORKERROR";
case RIL_E_MESSAGINGERROR:            return "RIL_E_NETWORK::MESSAGINGERROR";
case RIL_E_NOTREADY:                  return "RIL_E_NONE::NOTREADY";
case RIL_E_TIMEDOUT:                  return "RIL_E_NONE::TIMEDOUT";
case RIL_E_CANCELLED:                 return "RIL_E_NONE::CANCELLED";
case RIL_E_NONOTIFYCALLBACK:          return "RIL_E_NONE::NONOTIFYCALLBACK";
case RIL_E_OPFMTUNAVAILABLE:          return "RIL_E_NETWORKUNSUPPORTED::OPFMTUNAVAILABLE";
case RIL_E_NORESPONSETODIAL:          return "RIL_E_NETWORKACCESS::NORESPONSETODIAL";
case RIL_E_SECURITYFAILURE:           return "RIL_E_NONE::SECURITYFAILURE";
case RIL_E_RADIOFAILEDINIT:           return "RIL_E_NONE::RADIOFAILEDINIT";
case RIL_E_DRIVERINITFAILED:          return "RIL_E_RADIOUNAVAILABLE::DRIVERINITFAILED";
case RIL_E_RADIONOTPRESENT:           return "RIL_E_RADIOUNAVAILABLE::RADIONOTPRESENT";
case RIL_E_RADIOOFF:                  return "RIL_E_RADIOUNAVAILABLE::RADIOOFF";

default:
    return stringformat("%s::%02x", RilErrorClassString(RILERRORCLASS(dwError)).c_str(), dwError&0xff);
    }
}

RilClass::RilClass()
{
    m_hRil= 0;
    m_hrCallListRequest= 0;
    m_hrEquipmentInfoRq= 0;
    m_hrEquipmentStateRequest= 0;
    m_hrCurrentAddressIdRequest= 0;
    m_hrSignalQualityRequest= 0;
    m_hrSimRecordStatusRequest= 0;
    m_hrAudioDevicesRequest= 0;
    m_hrAudioMutingRequest= 0;
    m_hrSubscriberNumbersRequest= 0;
    m_hrLockedStateRequest= 0;
	m_hrSetAudioMutingRequest= 0;

    m_hrCBSRequest= 0;
	m_hrCellIdRequest= 0;
}
RilClass::~RilClass()
{
    close();
}
bool RilClass::open(DWORD dwIndex)
{
#ifdef _HAVE_RIL
    HRESULT res= RIL_Initialize(dwIndex, &ResultProc, &NotifyProc, -1, (DWORD)this, &m_hRil);
    if (res)
    {
        error("RIL_Initialize: %08lx\n", res);
        return false;
    }
#endif
    return true;
}
bool RilClass::close()
{
    bool bRes= true;
#ifdef _HAVE_RIL
    if (m_hRil)
    {
        HRESULT res= RIL_Deinitialize(m_hRil);
        if (res)
        {
            error("RIL_Deinitialize: %08lx\n", res);
            bRes= false;
        }
        m_hRil= 0;
    }
#endif
    return bRes;
}
std::string GetSerialPortStatisticsString(RILSERIALPORTSTATS *p)
{
    std::string info= "PORTSTATS:";
    if (p->dwParams&RIL_PARAM_SPS_READBITSPERSECOND)    info += stringformat(" read:%d bps",  p->dwReadBitsPerSecond);
    if (p->dwParams&RIL_PARAM_SPS_WRITTENBITSPERSECOND) info += stringformat(" write:%d bps", p->dwWrittenBitsPerSecond);
    return info;
}
void RilClass::GetDriverInfo()
{
    HRESULT hr;

    HANDLE hSerial;
    hr= RIL_GetSerialPortHandle(m_hRil, &hSerial);
    if (IS_ERROR(hr))
        error(hr, "RIL_GetSerialPortHandle");
    debug("RIL_GetSerialPortHandle: handle=%08lx\n", hSerial);

    RILSERIALPORTSTATS portstats;
    hr= RIL_GetSerialPortStatistics(m_hRil, &portstats);
    if (IS_ERROR(hr))
        error(hr, "RIL_GetSerialPortStatistics");
    debug("RIL_GetSerialPortStatistics: %hs\n", GetSerialPortStatisticsString(&portstats).c_str());

    DWORD dwVersion;
    hr= RIL_GetDriverVersion(m_hRil, &dwVersion);
    if (IS_ERROR(hr))
        error(hr, "RIL_GetDriverVersion");
    debug("RIL_GetDriverVersion: version=%08lx\n", dwVersion);

    DWORD dwRadioPresence;
    hr= RIL_GetRadioPresence(m_hRil, &dwRadioPresence);
    if (IS_ERROR(hr))
        error(hr, "RIL_GetRadioPresence");
    debug("RIL_GetRadioPresence: presence=%hs\n", GetRadioPresenceString(dwRadioPresence).c_str());


}
void RilClass::ResultProc(DWORD dwCode, HRESULT hrCmdID, const void *lpData, DWORD cbdata, DWORD dwParam)
{
    ((RilClass*)dwParam)->HandleResult(dwCode, hrCmdID, lpData, cbdata);
}
void RilClass::NotifyProc(DWORD dwCode, const void *lpData, DWORD cbdata, DWORD dwParam)
{
    ((RilClass*)dwParam)->HandleNotify(dwCode, lpData, cbdata);
}
void RilClass::HandleResult(DWORD dwCode, HRESULT hrCmdID, const void *lpData, DWORD cbdata)
{
	DWORD dwRilError= 0;

    SYSTEMTIME now;
    GetLocalTime(&now);

    debug("%04d-%02d-%02d %02d:%02d:%02d.%03d  RIL_Result: %hs id=%08lx %d bytes\n", 
            now.wYear, now.wMonth, now.wDay,
            now.wHour, now.wMinute, now.wSecond, now.wMilliseconds,
            RilResultString(dwCode).c_str(), hrCmdID, cbdata);
    //debug("raw: %hs\n", hexdump((BYTE*)lpData, cbdata).c_str());

    if (dwCode==RIL_RESULT_ERROR) {
		if (cbdata!=sizeof(DWORD)) {
            debug("ril-error - missing error code\n");
			return;
		}
        else {
			dwRilError= *(DWORD*)lpData;
            debug("ril-error: %hs\n", RilErrorString(dwRilError).c_str());
		}
    }


	//  find command the requested this result.

	if (hrCmdID==m_hrEquipmentInfoRq) {
		m_hrEquipmentInfoRq= 0;
		if (dwCode==RIL_RESULT_OK) {
			if (cbdata!=sizeof(RILEQUIPMENTINFO))
				return;
			HandleEquipmentInfoAnswer((RILEQUIPMENTINFO*)lpData);
		}
	}
    else if (hrCmdID==m_hrEquipmentStateRequest) {
		m_hrEquipmentStateRequest= 0;
		if (dwCode==RIL_RESULT_OK) {
			if (cbdata!=sizeof(RILEQUIPMENTSTATE))
				return;
			HandleEquipmentStateAnswer((RILEQUIPMENTSTATE*)lpData);
		}
	}
    else if (hrCmdID==m_hrCurrentAddressIdRequest) {
		m_hrCurrentAddressIdRequest= 0;
		if (dwCode==RIL_RESULT_OK) {
			HandleCurrentAddressIdAnswer(*(DWORD*)lpData);
		}
	}
    else if (hrCmdID==m_hrSimRecordStatusRequest) {
		m_hrSimRecordStatusRequest= 0;

		if (dwCode==RIL_RESULT_OK) {
			HandleSimRecordStatusAnswer((RILSIMRECORDSTATUS*)lpData, cbdata);
		}
	}

    else if (hrCmdID==m_hrSignalQualityRequest) {
		m_hrSignalQualityRequest= 0;

		if (dwCode==RIL_RESULT_OK) {
			if (cbdata!=sizeof(RILSIGNALQUALITY))
				return;
			HandleSignalQualityAnswer((RILSIGNALQUALITY *)lpData);
		}
		else {
			HandleSignalQualityError(dwRilError);
		}
	}
	else if (hrCmdID==m_hrCallListRequest) {
		m_hrCallListRequest= 0;

		if (dwCode==RIL_RESULT_OK) {
			HandleCallListAnswer((RILCALLINFO*)lpData, cbdata/sizeof(RILCALLINFO));
		}
	}
	else if (hrCmdID==m_hrAudioDevicesRequest) {
		m_hrAudioDevicesRequest= 0;

		if (dwCode==RIL_RESULT_OK) {
			HandleAudioDevicesAnswer((RILAUDIODEVICEINFO*)lpData);
		}
	}
    else if (hrCmdID==m_hrLockedStateRequest) {
		m_hrLockedStateRequest= 0;

		if (dwCode==RIL_RESULT_OK) {
			HandleLockedStateAnswer(*(DWORD*)lpData);
		}
	}
    else if (hrCmdID==m_hrSubscriberNumbersRequest) {
		m_hrSubscriberNumbersRequest= 0;

        if (cbdata%sizeof(RILSUBSCRIBERINFO))
            debug("subscribernumbers: noninteger nr of structs\n");
		if (dwCode==RIL_RESULT_OK) {
			HandleSubscriberNumbersAnswer((RILSUBSCRIBERINFO*)lpData, cbdata/sizeof(RILSUBSCRIBERINFO));
		}
	}
    else if (hrCmdID==m_hrAudioMutingRequest) {
		m_hrAudioMutingRequest= 0;

		if (dwCode==RIL_RESULT_OK) {
			HandleAudioMutingAnswer(*(BOOL*)lpData);
		}
	}
    else if (hrCmdID==m_hrSetAudioMutingRequest) {
		m_hrSetAudioMutingRequest= 0;
        debug("ril-set muting reply\n");
	}
    else if (hrCmdID==m_hrCBSRequest) {
		m_hrCBSRequest= 0;

        debug("ril-req CBS reply\n");
	}
    else if (hrCmdID==m_hrCellIdRequest) {
		m_hrCellIdRequest= 0;

        debug("ril-req CellId reply\n");
	}
    else {
        debug("ril::HandleResult unknown\n");
    }
}
void RilClass::HandleSignalQualityAnswer(RILSIGNALQUALITY *sq)
{
	debug("ril-HandleSignalQualityAnswer %hs\n", GetRilSignalQuality(sq).c_str());
}
void RilClass::HandleSignalQualityError(DWORD dwRilError)
{
	debug("ril-HandleSignalQualityError %08lx\n", dwRilError);
}
std::string GetSimRecordTypeString(DWORD dwType)
{
    switch(dwType)
    {
case RIL_SIMRECORDTYPE_UNKNOWN: return "RIL_SIMRECORDTYPE_UNKNOWN";
case RIL_SIMRECORDTYPE_TRANSPARENT: return "RIL_SIMRECORDTYPE_TRANSPARENT";
case RIL_SIMRECORDTYPE_CYCLIC: return "RIL_SIMRECORDTYPE_CYCLIC";
case RIL_SIMRECORDTYPE_LINEAR: return "RIL_SIMRECORDTYPE_LINEAR";
case RIL_SIMRECORDTYPE_MASTER: return "RIL_SIMRECORDTYPE_MASTER";
case RIL_SIMRECORDTYPE_DEDICATED: return "RIL_SIMRECORDTYPE_DEDICATED";

default: return stringformat("simrecordtype-%d", dwType);
    }
}
std::string GetSimRecordStatusString(RILSIMRECORDSTATUS *ei, DWORD dwSize)
{
    std::string info= "RILSIMRECORDSTATUS : ";

    if (ei->dwParams&RIL_PARAM_SRS_RECORDTYPE) info += stringformat(" rectype=%hs", GetSimRecordTypeString(ei->dwRecordType).c_str());
    if (ei->dwParams&RIL_PARAM_SRS_ITEMCOUNT ) info += stringformat(" count=%d", ei->dwItemCount);
    if (ei->dwParams&RIL_PARAM_SRS_SIZE      ) info += stringformat(" itemsize=%hs", ei->dwSize);

    info += hexdump((BYTE*)(ei+1), dwSize-sizeof(RILSIMRECORDSTATUS));

    return info;
}
void RilClass::HandleSimRecordStatusAnswer(RILSIMRECORDSTATUS *sq, DWORD dwSize)
{
	debug("ril-HandleSimRecordStatusAnswer: %hs\n", GetSimRecordStatusString(sq, dwSize).c_str());
}

void RilClass::HandleEquipmentInfoAnswer(RILEQUIPMENTINFO* info)
{
    debug("ril-HandleEquipmentInfoAnswer: %hs\n", GetEquipmentInfoString(info).c_str());
}

void RilClass::HandleEquipmentStateAnswer(RILEQUIPMENTSTATE* state)
{
    debug("ril-HandleEquipmentStateAnswer: %hs\n", GetEquipmentStateString(state).c_str());
}
void RilClass::HandleCurrentAddressIdAnswer(DWORD dwId)
{
    debug("ril-HandleCurrentAddressIdAnswer: %08lx\n", dwId);
}

void RilClass::HandleCallListAnswer(RILCALLINFO* cilist, int ncalls)
{
    debug("RilClass::HandleCallListAnswer ");
	for (int i=0 ; i<ncalls ; i++)
	{
		debug("%d: %hs\n", i, GetCallInfoString(&cilist[i]).c_str());
	}
}

void RilClass::HandleAudioMutingAnswer(BOOL bMuted)
{
    debug("ril audio muting: %d\n", bMuted);
}

std::string GetSubscriberInfoString(RILSUBSCRIBERINFO *p)
{
    std::string info= "SUBINFO: ";
    if (p->dwParams&RIL_PARAM_SI_ADDRESS     ) info += stringformat(" address    = %hs", GetRilAddressString(&p->raAddress).c_str());
    if (p->dwParams&RIL_PARAM_SI_DESCRIPTION ) info += stringformat(" description= %ls", p->wszDescription);
    if (p->dwParams&RIL_PARAM_SI_SPEED       ) info += stringformat(" speed      = %08x", p->dwSpeed);
    if (p->dwParams&RIL_PARAM_SI_SERVICE     ) info += stringformat(" service    = %08x", p->dwService);
    if (p->dwParams&RIL_PARAM_SI_ITC         ) info += stringformat(" itc        = %08x", p->dwITC);
    if (p->dwParams&RIL_PARAM_SI_ADDRESSID   ) info += stringformat(" addressid  = %08x", p->dwAddressId);

    return info;
}
void RilClass::HandleSubscriberNumbersAnswer(RILSUBSCRIBERINFO *pList, int n)
{
    for (int i=0 ; i<n ; i++)
        debug("ril subscriberinfo - %d : %s\n", i, GetSubscriberInfoString(pList+i).c_str());
}


//--------------------------------------------------------------------
std::string GetLockedStateString(DWORD dwLockedState)
{
    switch(dwLockedState)
    {
case RIL_LOCKEDSTATE_UNKNOWN: return "RIL_LOCKEDSTATE_UNKNOWN";
case RIL_LOCKEDSTATE_READY: return "RIL_LOCKEDSTATE_READY";
case RIL_LOCKEDSTATE_SIM_PIN: return "RIL_LOCKEDSTATE_SIM_PIN";
case RIL_LOCKEDSTATE_SIM_PUK: return "RIL_LOCKEDSTATE_SIM_PUK";
case RIL_LOCKEDSTATE_PH_SIM_PIN: return "RIL_LOCKEDSTATE_PH_SIM_PIN";
case RIL_LOCKEDSTATE_PH_FSIM_PIN: return "RIL_LOCKEDSTATE_PH_FSIM_PIN";
case RIL_LOCKEDSTATE_PH_FSIM_PUK: return "RIL_LOCKEDSTATE_PH_FSIM_PUK";
case RIL_LOCKEDSTATE_SIM_PIN2: return "RIL_LOCKEDSTATE_SIM_PIN2";
case RIL_LOCKEDSTATE_SIM_PUK2: return "RIL_LOCKEDSTATE_SIM_PUK2";
case RIL_LOCKEDSTATE_PH_NET_PIN: return "RIL_LOCKEDSTATE_PH_NET_PIN";
case RIL_LOCKEDSTATE_PH_NET_PUK: return "RIL_LOCKEDSTATE_PH_NET_PUK";
case RIL_LOCKEDSTATE_PH_NETSUB_PIN: return "RIL_LOCKEDSTATE_PH_NETSUB_PIN";
case RIL_LOCKEDSTATE_PH_NETSUB_PUK: return "RIL_LOCKEDSTATE_PH_NETSUB_PUK";
case RIL_LOCKEDSTATE_PH_SP_PIN: return "RIL_LOCKEDSTATE_PH_SP_PIN";
case RIL_LOCKEDSTATE_PH_SP_PUK: return "RIL_LOCKEDSTATE_PH_SP_PUK";
case RIL_LOCKEDSTATE_PH_CORP_PIN: return "RIL_LOCKEDSTATE_PH_CORP_PIN";
case RIL_LOCKEDSTATE_PH_CORP_PUK: return "RIL_LOCKEDSTATE_PH_CORP_PUK";
default: return stringformat("lockedstate-%d", dwLockedState);
    }
}

void RilClass::HandleLockedStateAnswer(DWORD dwLockedState)
{
    debug("ril lockedstate : %hs\n", GetLockedStateString(dwLockedState).c_str());
}

void RilClass::HandleAudioDevicesAnswer(RILAUDIODEVICEINFO* devinfo)
{
    debug("ril audio devices: tx=%d rx=%d\n", devinfo->dwTxDevice, devinfo->dwRxDevice);
}

//--------------------------------------------------------------------
std::string GetCallTypeString(int calltype)
{
    switch(calltype)
    {
case RIL_CALLTYPE_UNKNOWN: return "RIL_CALLTYPE_UNKNOWN";
case RIL_CALLTYPE_VOICE: return "RIL_CALLTYPE_VOICE";
case RIL_CALLTYPE_DATA: return "RIL_CALLTYPE_DATA";
case RIL_CALLTYPE_FAX: return "RIL_CALLTYPE_FAX";
default: return stringformat("calltype-%d", calltype);
    }
}

std::string GetDisconnectInitString(int di)
{
    switch(di)
    {
case RIL_DISCINIT_LOCAL: return "RIL_DISCINIT_LOCAL";
case RIL_DISCINIT_REMOTE: return "RIL_DISCINIT_REMOTE";
default: return stringformat("discinit-%d", di);
    }
}
std::string GetServiceInfoString(RILSERVICEINFO *si)
{
    std::string info= "SVCINFO: ";
    if (si->dwParams&RIL_PARAM_SVCI_SYNCHRONOUS)
        info += stringformat("  synchronous=%d", si->fSynchronous);
    if (si->dwParams&RIL_PARAM_SVCI_TRANSPARENT)
        info += stringformat("  transparent=%d", si->fTransparent);

    return info;
}

std::string GetRingInfoString(RILRINGINFO* ri)
{
    std::string info= "RINGINFO: ";
    if (ri->dwParams&RIL_PARAM_RI_CALLTYPE)
        info += stringformat(" calltype= %hs", GetCallTypeString(ri->dwCallType).c_str());
    if (ri->dwParams&RIL_PARAM_RI_ADDRESSID)
        info += stringformat(" addressid= %d", ri->dwAddressId);
    if (ri->dwParams&RIL_PARAM_RI_SERVICEINFO)
        info += stringformat(" serviceinfo= %hs", GetServiceInfoString(&ri->rsiServiceInfo).c_str());

    return info;
}
std::string GetConnectInfoString(RILCONNECTINFO* ci)
{
    std::string info= "CONNECTINFO: ";
    if (ci->dwParams&RIL_PARAM_CNI_CALLTYPE)
        info += stringformat("  calltype=%hs", GetCallTypeString(ci->dwCallType).c_str());
    if (ci->dwParams&RIL_PARAM_CNI_BAUDRATE)
        info += stringformat("  baudrate=%d", ci->dwBaudRate);

    return info;
}

std::string GetCallInfoString(RILCALLINFO* ci)
{
    std::string info= "CALLINFO: ";
    if (ci->dwParams&RIL_PARAM_CI_ID)
        info += stringformat("  id=%d", ci->dwID);
    if (ci->dwParams&RIL_PARAM_CI_DIRECTION)
        info += stringformat("  dir=%d", ci->dwDirection);
    if (ci->dwParams&RIL_PARAM_CI_STATUS)
        info += stringformat("  stat=%d", ci->dwStatus);
    if (ci->dwParams&RIL_PARAM_CI_TYPE)
        info += stringformat("  type=%d", ci->dwType);
    if (ci->dwParams&RIL_PARAM_CI_MULTIPARTY)
        info += stringformat("  multi=%d", ci->dwMultiparty);
    if (ci->dwParams&RIL_PARAM_CI_ADDRESS)
        info += stringformat("  addr=%hs", GetRilAddressString(&ci->raAddress).c_str());
    if (ci->dwParams&RIL_PARAM_CI_DESCRIPTION)
        info += stringformat("  desc=%d", ci->wszDescription);
    if (ci->dwParams&RIL_PARAM_CI_CPISTATUS)
        info += stringformat("  cpistat=%d", ci->dwStatus);

    return info;
}

void HandleCallControlRing(RILRINGINFO* ri, DWORD cbdata)
{
	if (ri->cbSize!=cbdata)
		debug("ril-HandleCallControlRing: size mismatch ( rec:%d not:%d )\n", ri->cbSize, cbdata);
    debug("ril-callcontrolring: %hs  %hs\n", GetRingInfoString(ri).c_str(), GetCallTypeString(ri->dwCallType).c_str());
}
void HandleCallControlConnect(RILCONNECTINFO* ci, DWORD cbdata)
{
	if (ci->cbSize!=cbdata)
		debug("ril-HandleCallControlConnect: size mismatch ( rec:%d not:%d )\n", ci->cbSize, cbdata);
    debug("ril-callcontrolconnect: %hs  %hs\n", GetConnectInfoString(ci).c_str(), GetCallTypeString(ci->dwCallType).c_str());

    if (g_ril) g_ril->RequestCallList();
}

void HandleCallControlDisconnect(DWORD* d, DWORD cbdata)
{
	if (cbdata!=sizeof(DWORD))
		debug("ril-HandleCallControlDisconnect: size mismatch ( not:%d )\n", cbdata);
    debug("ril-Disconnect: %hs\n", GetDisconnectInitString(*d).c_str());
}

void HandleCallControlDataServiceNegotiated(RILSERVICEINFO* si, DWORD cbdata)
{
	if (si->cbSize!=cbdata)
		debug("ril-HandleCallControlDataServiceNegotiated: size mismatch ( rec:%d not:%d )\n", si->cbSize, cbdata);
    debug("ril-callcontroldatasvc: %hs\n", GetServiceInfoString(si).c_str());
}

std::string GetRilAddressString(RILADDRESS* ra)
{
    std::string info= "ADDRESS: ";
    if (ra->dwParams&RIL_PARAM_A_TYPE)
        info += stringformat(" type=%d", ra->dwType);
    if (ra->dwParams&RIL_PARAM_A_NUMPLAN)
        info += stringformat(" numplan=%d", ra->dwNumPlan);
    if (ra->dwParams&RIL_PARAM_A_ADDRESS)
        info += stringformat(" address=%ls", ra->wszAddress);

    return info;
}
std::string GetRilSubAddressString(RILSUBADDRESS* ra)
{
    std::string info= "SUBADDRESS: ";
    if (ra->dwParams&RIL_PARAM_SA_TYPE)
        info += stringformat(" type=%d", ra->dwType);
    if (ra->dwParams&RIL_PARAM_SA_SUBADDRESS)
        info += stringformat(" subaddress=%ls", ra->wszSubAddress);

    return info;
}
std::string GetRemotePartyInfoString(RILREMOTEPARTYINFO* pi)
{
    std::string info= "REMOTEPARTYINFO: ";

    if (pi->dwParams&RIL_PARAM_RPI_ADDRESS    )
        info += stringformat(" calleraddress=%hs", GetRilAddressString(&pi->raAddress).c_str());
    if (pi->dwParams&RIL_PARAM_RPI_SUBADDRESS )
        info += stringformat(" callersubaddress=%hs", GetRilSubAddressString(&pi->rsaSubAddress).c_str());
    if (pi->dwParams&RIL_PARAM_RPI_DESCRIPTION)
        info += stringformat(" desc=%ls", pi->wszDescription);
    if (pi->dwParams&RIL_PARAM_RPI_VALIDITY   )
        info += stringformat(" validity=%d", pi->dwValidity);

    return info;
}

void HandleSupServiceCallerId(RILREMOTEPARTYINFO* pi, DWORD cbdata)
{
	if (pi->cbSize!=cbdata)
		debug("ril-HandleSupServiceCallerId: size mismatch ( rec:%d not:%d )\n", pi->cbSize, cbdata);
    debug("ril-supservice: %hs\n", GetRemotePartyInfoString(pi).c_str());
}
void HandleSimtoolkitCommand(BYTE* lpData, DWORD cbdata)
{
    debug("ril-SIMTOOLKITCMD: %hs\n", lpData==NULL?"NULL":hexdump(lpData,cbdata).c_str());
}

std::string GetRegistrationStatusString(int rs)
{
    switch(rs)
    {
case RIL_REGSTAT_UNKNOWN: return "RIL_REGSTAT_UNKNOWN";
case RIL_REGSTAT_UNREGISTERED: return "RIL_REGSTAT_UNREGISTERED";
case RIL_REGSTAT_HOME: return "RIL_REGSTAT_HOME";
case RIL_REGSTAT_ATTEMPTING: return "RIL_REGSTAT_ATTEMPTING";
case RIL_REGSTAT_DENIED: return "RIL_REGSTAT_DENIED";
case RIL_REGSTAT_ROAMING: return "RIL_REGSTAT_ROAMING";
default: return stringformat("regstat-%d", rs);
    }
}

void HandleNetworkGprsStatusChanged(DWORD* d, DWORD cbdata)
{
	if (cbdata!=sizeof(DWORD))
		debug("ril-HandleNetworkGprsStatusChanged: size mismatch ( not:%d )\n", cbdata);
    debug("ril-GprsStatus: %hs\n", GetRegistrationStatusString(*d).c_str());
}
void RilClass::HandleNetworkRegistrationStatusChanged(DWORD* d, DWORD cbdata)
{
	if (cbdata!=sizeof(DWORD))
		debug("ril-HandleNetworkRegistrationStatusChanged: size mismatch ( not:%d )\n", cbdata);
    debug("ril-RegistrationStatus: %hs\n", GetRegistrationStatusString(*d).c_str());
}

std::string GetRadioPresenceString(int rp)
{
    switch(rp)
    {
case RIL_RADIOPRESENCE_NOTPRESENT: return "RIL_RADIOPRESENCE_NOTPRESENT";
case RIL_RADIOPRESENCE_PRESENT: return "RIL_RADIOPRESENCE_PRESENT";
default: return stringformat("radioprecense-%d", rp);
    }
}
std::string GetRadioSupportString(int rp)
{
    switch(rp)
    {
case RIL_RADIOSUPPORT_UNKNOWN: return "RIL_RADIOSUPPORT_UNKNOWN";
case RIL_RADIOSUPPORT_OFF: return "RIL_RADIOSUPPORT_OFF";
case RIL_RADIOSUPPORT_ON: return "RIL_RADIOSUPPORT_ON";
default: return stringformat("radiosupport-%d", rp);
    }
}
std::string GetEqStateString(int rp)
{
    switch(rp)
    {
case RIL_EQSTATE_UNKNOWN: return "RIL_EQSTATE_UNKNOWN";
case RIL_EQSTATE_MINIMUM: return "RIL_EQSTATE_MINIMUM";
case RIL_EQSTATE_FULL: return "RIL_EQSTATE_FULL";
case RIL_EQSTATE_DISABLETX: return "RIL_EQSTATE_DISABLETX";
case RIL_EQSTATE_DISABLERX: return "RIL_EQSTATE_DISABLERX";
case RIL_EQSTATE_DISABLETXANDRX: return "RIL_EQSTATE_DISABLETXANDRX";
default: return stringformat("eqstate-%d", rp);
    }
}
std::string GetPbLocString(int rp)
{
    switch(rp)
    {
case RIL_PBLOC_UNKNOWN: return "RIL_PBLOC_UNKNOWN";
case RIL_PBLOC_SIMEMERGENCY: return "RIL_PBLOC_SIMEMERGENCY";
case RIL_PBLOC_SIMFIXDIALING: return "RIL_PBLOC_SIMFIXDIALING";
case RIL_PBLOC_SIMLASTDIALING: return "RIL_PBLOC_SIMLASTDIALING";
case RIL_PBLOC_OWNNUMBERS: return "RIL_PBLOC_OWNNUMBERS";
case RIL_PBLOC_SIMPHONEBOOK: return "RIL_PBLOC_SIMPHONEBOOK";
default: return stringformat("pbloc-%d", rp);
    }
}

std::string GetReadyStateString(int rp)
{
    std::string state;
    if (rp==RIL_READYSTATE_NONE)       state += " RIL_READYSTATE_NONE";
    if (rp&RIL_READYSTATE_INITIALIZED) state += " RIL_READYSTATE_INITIALIZED";
    if (rp&RIL_READYSTATE_SIM)         state += " RIL_READYSTATE_SIM";
    if (rp&RIL_READYSTATE_SMS)         state += " RIL_READYSTATE_SMS";
    if (rp&RIL_READYSTATE_UNLOCKED)    state += " RIL_READYSTATE_UNLOCKED";
    if (rp&~(RIL_READYSTATE_INITIALIZED|RIL_READYSTATE_SIM|RIL_READYSTATE_SMS|RIL_READYSTATE_UNLOCKED))
        state += stringformat("readystate-0x%x", rp&~(RIL_READYSTATE_INITIALIZED|RIL_READYSTATE_SIM|RIL_READYSTATE_SMS|RIL_READYSTATE_UNLOCKED));
    return state;
}
std::string GetEquipmentStateString(RILEQUIPMENTSTATE* es)
{
    std::string info= "EQUIPMENTSTATE: ";

    if (es->dwParams&RIL_PARAM_EQUIPMENTSTATE_RADIOSUPPORT)
        info += stringformat(" %hs", GetRadioSupportString(es->dwRadioSupport).c_str());
    if (es->dwParams&RIL_PARAM_EQUIPMENTSTATE_EQSTATE     )
        info += stringformat(" %hs", GetEqStateString(es->dwEqState).c_str());
    if (es->dwParams&RIL_PARAM_EQUIPMENTSTATE_READYSTATE  )
        info += stringformat(" %hs", GetReadyStateString(es->dwReadyState).c_str());

    return info;
}
std::string GetEquipmentInfoString(RILEQUIPMENTINFO* ei)
{
    std::string info= "EQUIPMENTINFO: ";

    if (ei->dwParams&RIL_PARAM_EI_MANUFACTURER) info += stringformat(" manufacturer=%hs", ei->szManufacturer);
    if (ei->dwParams&RIL_PARAM_EI_MODEL)        info += stringformat(" model=%hs", ei->szModel);
    if (ei->dwParams&RIL_PARAM_EI_REVISION)     info += stringformat(" revision=%hs", ei->szRevision);
    if (ei->dwParams&RIL_PARAM_EI_SERIALNUMBER) info += stringformat(" serial=%hs", ei->szSerialNumber);

    return info;
}
std::string GetRilSignalQuality(RILSIGNALQUALITY *sq)
{
    std::string info= "RILSIGNALQUALITY: ";
    if (sq->dwParams&RIL_PARAM_SQ_SIGNALSTRENGTH    ) info += stringformat(" ss=%4d", sq->nSignalStrength);
    if (sq->dwParams&RIL_PARAM_SQ_MINSIGNALSTRENGTH ) info += stringformat(" mins=%4d", sq->nMinSignalStrength);
    if (sq->dwParams&RIL_PARAM_SQ_MAXSIGNALSTRENGTH ) info += stringformat(" maxs=%4d", sq->nMaxSignalStrength);
    if (sq->dwParams&RIL_PARAM_SQ_BITERRORRATE      ) info += stringformat(" ber=%4d", sq->dwBitErrorRate);
    if (sq->dwParams&RIL_PARAM_SQ_LOWSIGNALSTRENGTH ) info += stringformat(" los=%4d", sq->nLowSignalStrength);
    if (sq->dwParams&RIL_PARAM_SQ_HIGHSIGNALSTRENGTH) info += stringformat(" his=%4d", sq->nHighSignalStrength);

    return info;
}

void RilClass::HandleRadioStateChanged(RILEQUIPMENTSTATE* es, DWORD cbdata)
{
	if (es->cbSize!=cbdata)
		debug("ril-HandleRadioStateChanged: size mismatch ( rec:%d not:%d )\n", es->cbSize, cbdata);

    debug("ril-RadioState: %hs\n", GetEquipmentStateString(es).c_str());
}
void HandleRadioPresenceChanged(DWORD* d, DWORD cbdata)
{
	if (cbdata!=sizeof(DWORD))
		debug("ril-HandleRadioPresenceChanged: size mismatch ( not:%d )\n", cbdata);

    debug("ril-RadioPresence: %hs\n", GetRadioPresenceString(*d).c_str());
}

std::string GetRilMessageString(RILMESSAGE* m)
{
    std::string info= "RILMESSAGE: ";
    if (m->dwParams&RIL_PARAM_M_SVCCTRADDRESS)
        info += stringformat(" servicecenter=%hs", GetRilAddressString(&m->raSvcCtrAddress).c_str());
    if (m->dwParams&RIL_PARAM_M_FLAGS)
        info += stringformat(" flags=%08lx", m->dwFlags);

    if (m->dwParams&RIL_PARAM_M_TYPE)
    switch(m->dwType)
    {
    case RIL_MSGTYPE_IN_DELIVER : 
        info += stringformat(" msgInDeliver");
        break;
    case RIL_MSGTYPE_IN_STATUS  : 
        info += stringformat(" msgInStatus");
        break;
    case RIL_MSGTYPE_OUT_SUBMIT : 
        info += stringformat(" msgOutSubmit");
        break;
    case RIL_MSGTYPE_OUT_COMMAND: 
        info += stringformat(" msgOutCommand");
        break;
    case RIL_MSGTYPE_OUT_RAW    : 
        info += stringformat(" msgOutRaw");
        break;
    case RIL_MSGTYPE_BC_GENERAL : 
        info += stringformat(" msgBcGeneral");
        break;
    default:
        info += stringformat("UNKNOWN RILMSGTYPE: %d", m->dwType);
    }
    else {
        info += " !!! NO TYPE !!!";
    }
    return info;
}

void RilClass::HandleCellIdMessage(RILCELLID *cellid)
{
	debug("ril - lac=%08lx  cellid=%08lx\n", cellid->lac, cellid->cellid);
}

void HandleMessageMessage(RILMESSAGE* m, DWORD cbdata)
{
	if (m->cbSize!=cbdata)
		debug("ril-HandleMessageMessage: size mismatch ( rec:%d not:%d )\n", m->cbSize, cbdata);

    debug("RilMessage: %hs\n", GetRilMessageString(m).c_str());
}
void HandleCallStateChanged()
{
	debug("ril-HandleCallStateChanged\n");
    if (g_ril) g_ril->RequestCallList();
}

void HandlePhonebookStorageChanged(DWORD dwPbLoc)
{
	debug("ril-HandlePhonebookStorageChanged %hs\n", GetPbLocString(dwPbLoc).c_str());
}
void HandleCurrentLineChanged(DWORD dwAddressId)
{
	debug("ril-HandleCurrentLineChanged curaddressid=%d\n", dwAddressId);
}
void HandleSignalQuality(RILSIGNALQUALITY *sq, DWORD cbdata)
{
    if (cbdata!=sq->cbSize)
		debug("ril-HandleSignalQuality: size mismatch ( rec:%d not:%d )\n", sq->cbSize, cbdata);

    debug("ril-HandleSignalQuality: %hs\n", GetRilSignalQuality(sq).c_str());
}
void RilClass::HandleNotify(DWORD dwCode, const void *lpData, DWORD cbdata)
{
    SYSTEMTIME now;
    GetLocalTime(&now);
    debug("%04d-%02d-%02d %02d:%02d:%02d.%03d RIL_Notification: %hs %d bytes\n", 
            now.wYear, now.wMonth, now.wDay,
            now.wHour, now.wMinute, now.wSecond, now.wMilliseconds,
            RilResultString(dwCode).c_str(), cbdata);

    switch(dwCode) {
        case RIL_NOTIFY_RING: HandleCallControlRing((RILRINGINFO*)lpData, cbdata); break;
        case RIL_NOTIFY_CONNECT: HandleCallControlConnect((RILCONNECTINFO*)lpData, cbdata); break;
        case RIL_NOTIFY_DISCONNECT: HandleCallControlDisconnect((DWORD*)lpData, cbdata); break;
        case RIL_NOTIFY_DATASVCNEGOTIATED: HandleCallControlDataServiceNegotiated((RILSERVICEINFO*)lpData, cbdata); break;
        case RIL_NOTIFY_CALLSTATECHANGED: HandleCallStateChanged(); break;
        case RIL_NOTIFY_CALLERID: HandleSupServiceCallerId((RILREMOTEPARTYINFO*)lpData, cbdata); break;
        case RIL_NOTIFY_SIMTOOLKITCMD: HandleSimtoolkitCommand((BYTE*)lpData, cbdata); break;
        case RIL_NOTIFY_GPRSREGSTATUSCHANGED: HandleNetworkGprsStatusChanged((DWORD*)lpData, cbdata); break;
        case RIL_NOTIFY_REGSTATUSCHANGED: HandleNetworkRegistrationStatusChanged((DWORD*)lpData, cbdata); break;
        case RIL_NOTIFY_RADIOEQUIPMENTSTATECHANGED: HandleRadioStateChanged((RILEQUIPMENTSTATE*)lpData, cbdata); break;
        case RIL_NOTIFY_RADIOPRESENCECHANGED: HandleRadioPresenceChanged((DWORD*)lpData, cbdata); break;
        case RIL_NOTIFY_MESSAGE: HandleMessageMessage((RILMESSAGE*)lpData, cbdata); break;
		case RIL_NOTIFY_CELLID: HandleCellIdMessage((RILCELLID*)lpData); break;
        case RIL_NOTIFY_PHONEBOOKSTORAGECHANGED: HandlePhonebookStorageChanged(*(DWORD*)lpData); break;
        case RIL_NOTIFY_CURRENTLINECHANGED: HandleCurrentLineChanged(*(DWORD*)lpData); break;
        case RIL_NOTIFY_SIGNALQUALITY: HandleSignalQuality((RILSIGNALQUALITY *)lpData, cbdata); break;

        default: debug("ril- unknown notification: %08lx %hs\n", dwCode, hexdump((const BYTE*)lpData, cbdata).c_str()); 
    }
}

bool RilClass::RequestSignalQuality()
{
#ifdef _HAVE_RIL
    if (m_hRil==NULL)
    {
        debug("ril-ERROR: ril not initialized\n");
        return false;
    }
	if (m_hrSignalQualityRequest)
	{
		debug("ril-RequestSignalQuality: still busy\n");
		return true;
	}
    m_hrSignalQualityRequest= RIL_GetSignalQuality(m_hRil);
    if (IS_ERROR(m_hrSignalQualityRequest))
    {
        error(m_hrSignalQualityRequest, "RIL_GetSignalQuality");
		m_hrSignalQualityRequest= 0;
        return false;
    }
#endif
    return true;
}
bool RilClass::RequestSimRecordStatus(DWORD dwFile)
{
#ifdef _HAVE_RIL
    if (m_hRil==NULL)
    {
        debug("ril-ERROR: ril not initialized\n");
        return false;
    }
	if (m_hrSimRecordStatusRequest)
	{
		debug("ril-RequestSimRecordStatus: still busy\n");
		return true;
	}
    m_hrSimRecordStatusRequest= RIL_GetSimRecordStatus(m_hRil, dwFile);
    if (IS_ERROR(m_hrSimRecordStatusRequest))
    {
        error(m_hrSimRecordStatusRequest, "RIL_GetSimRecordStatus");
		m_hrSimRecordStatusRequest= 0;
        return false;
    }
#endif
    return true;
}


bool RilClass::RequestEquipmentInfo()
{
#ifdef _HAVE_RIL
	if (m_hrEquipmentInfoRq)
	{
		debug("ril-RequestEquipmentInfo: still busy\n");
		return true;
	}
    m_hrEquipmentInfoRq= RIL_GetEquipmentInfo(m_hRil);
    if (IS_ERROR(m_hrEquipmentInfoRq))
	{
        error(m_hrEquipmentInfoRq, "RIL_GetEquipmentInfo");
		m_hrEquipmentInfoRq= 0;
        return false;
    }
#endif
	return true;
}

bool RilClass::RequestCallList()
{
#ifdef _HAVE_RIL
	if (m_hrCallListRequest)
	{
		debug("ril-RequestCallList: still busy\n");
		return true;
	}
	m_hrCallListRequest= RIL_GetCallList(m_hRil);
    if (IS_ERROR(m_hrCallListRequest))
	{
        error(m_hrCallListRequest, "RIL_GetCallList");
		m_hrCallListRequest= 0;
        return false;
    }
#endif
	return true;
}
bool RilClass::RequestAudioDevices()
{
#ifdef _HAVE_RIL
	if (m_hrAudioDevicesRequest)
	{
		debug("ril-RequestAudioDevices: still busy\n");
		return true;
	}
	m_hrAudioDevicesRequest= RIL_GetAudioDevices(m_hRil);
    if (IS_ERROR(m_hrAudioDevicesRequest))
	{
        error(m_hrAudioDevicesRequest, "RIL_GetAudioDevices");
		m_hrAudioDevicesRequest= 0;
        return false;
    }
#endif
	return true;
}
bool RilClass::RequestLockedState()
{
#ifdef _HAVE_RIL
	if (m_hrLockedStateRequest)
	{
		debug("ril-RequestLockedState: still busy\n");
		return true;
	}
	m_hrLockedStateRequest= RIL_GetPhoneLockedState(m_hRil);
    if (IS_ERROR(m_hrLockedStateRequest))
	{
        error(m_hrLockedStateRequest, "RIL_GetPhoneLockedState");
		m_hrLockedStateRequest= 0;
        return false;
    }
#endif
	return true;
}
bool RilClass::RequestCurrentAddressId()
{
#ifdef _HAVE_RIL
	if (m_hrCurrentAddressIdRequest)
	{
		debug("ril-RequestCurrentAddressId: still busy\n");
		return true;
	}
	m_hrCurrentAddressIdRequest= RIL_GetCurrentAddressId(m_hRil);
    if (IS_ERROR(m_hrCurrentAddressIdRequest))
	{
        error(m_hrCurrentAddressIdRequest, "RIL_GetCurrentAddressId");
		m_hrCurrentAddressIdRequest= 0;
        return false;
    }
#endif
	return true;
}

bool RilClass::RequestEquipmentState()
{
#ifdef _HAVE_RIL
	if (m_hrEquipmentStateRequest)
	{
		debug("ril-RequestEquipmentState: still busy\n");
		return true;
	}
	m_hrEquipmentStateRequest= RIL_GetEquipmentState(m_hRil);
    if (IS_ERROR(m_hrEquipmentStateRequest))
	{
        error(m_hrEquipmentStateRequest, "RIL_GetEquipmentState");
		m_hrEquipmentStateRequest= 0;
        return false;
    }
#endif
	return true;
}
bool RilClass::RequestSubscriberNumbers()
{
#ifdef _HAVE_RIL
	if (m_hrSubscriberNumbersRequest)
	{
		debug("ril-RequestSubscriberNumbers: still busy\n");
		return true;
	}
    
	m_hrSubscriberNumbersRequest= RIL_GetSubscriberNumbers(m_hRil);
    if (IS_ERROR(m_hrSubscriberNumbersRequest))
	{
        error(m_hrSubscriberNumbersRequest, "RIL_GetSubscriberInfo");
		m_hrSubscriberNumbersRequest= 0;
        return false;
    }
#endif
	return true;
}


bool RilClass::RequestAudioMuting()
{
#ifdef _HAVE_RIL
	if (m_hrAudioMutingRequest)
	{
		debug("ril-RequestAudioMuting: still busy\n");
		return true;
	}
	m_hrAudioMutingRequest= RIL_GetAudioMuting(m_hRil);
    if (IS_ERROR(m_hrAudioMutingRequest))
	{
        error(m_hrAudioMutingRequest, "RIL_GetAudioMuting");
		m_hrAudioMutingRequest= 0;
        return false;
    }
#endif
	return true;
}
bool RilClass::SetAudioMuting(bool bFlag)
{
#ifdef _HAVE_RIL
	if (m_hrSetAudioMutingRequest)
	{
		debug("RilClass::SetAudioMuting: still busy\n");
		return true;
	}
	m_hrSetAudioMutingRequest= RIL_SetAudioMuting(m_hRil, bFlag);
    if (IS_ERROR(m_hrSetAudioMutingRequest))
	{
        error(m_hrSetAudioMutingRequest, "RIL_SetAudioMuting");
		m_hrSetAudioMutingRequest= 0;
        return false;
    }
#endif
	return true;
}
// wait for up to 5 seconds for answer from ril.
bool WaitForRilAnswer(HRESULT &hrCmdId)
{
    DWORD t0= GetTickCount();
    while (hrCmdId && GetTickCount()-t0 < 5000) {
        Sleep(100);
    }
    if (hrCmdId) {
        hrCmdId= 0;
        return false;
    }
    return true;
}

bool RilClass::EnableCellBroadcasts(bool bFlag)
{
#ifdef _HAVE_RIL
	if (m_hrCBSRequest)
	{
		debug("RilClass::EnableCellBroadcasts: CBSRequest: still busy\n");
		return true;
	}
    DWORD req= bFlag? 24 : 25;
	m_hrCBSRequest= RIL_DevSpecific(m_hRil, (BYTE*)&req, sizeof(req));
    if (IS_ERROR(m_hrCBSRequest))
	{
        error(m_hrCBSRequest, "RIL_DevSpecific(%d)", req);
		m_hrCBSRequest= 0;
        return false;
    }

    if (!WaitForRilAnswer(m_hrCBSRequest))
    {
        debug("ril-ERROR: RIL_DevSpecific(%d) timeout\n", req);
        return false;
    }

#endif
	return true;
}
bool RilClass::EnableCellId(bool bFlag)
{
#ifdef _HAVE_RIL
	if (m_hrCellIdRequest)
	{
		debug("RilClass::EnableCellId: CellIdRequest: still busy\n");
	}
    DWORD req= bFlag? 26 : 27;
	m_hrCellIdRequest= RIL_DevSpecific(m_hRil, (BYTE*)&req, sizeof(req));
    if (IS_ERROR(m_hrCellIdRequest))
	{
        error(m_hrCellIdRequest, "RIL_DevSpecific(%d)", req);
		m_hrCellIdRequest= 0;
        return false;
    }
    if (!WaitForRilAnswer(m_hrCellIdRequest))
    {
        debug("ril-ERROR: RIL_DevSpecific(%d) timeout\n", req);
        return false;
    }

#endif
	return true;
}
bool RilClass::SetEqState(DWORD dwState)
{
    HRESULT hr= RIL_SetEquipmentState(m_hRil, dwState);
    if (IS_ERROR(hr)) {
        error(hr, "RIL_SetEquipmentState");
        return false;
    }
    return true;
}
/*
 * ril functions not yet implemented:

RIL_DevSpecific(HRIL hRil, const BYTE* lpbParams, DWORD dwSize);    -> BYTE*
RIL_GetDevCaps(HRIL hRil, DWORD dwCapsType);
    RIL_CAPSTYPE_DIAL                    points to an <t RILCAPSDIAL> structure
    RIL_CAPSTYPE_DTMFDURATIONRANGE       points to an <t RILRANGE> structure (values in milliseconds)
    RIL_CAPSTYPE_BEARERSERVICE           points to an <t RILCAPSBEARERSVC> structure
    RIL_CAPSTYPE_PBINDEXRANGE            points to an <t RILRANGE> structure
    RIL_CAPSTYPE_PBENTRYLENGTH           points to an <t RILCAPSPBENTRYLENGTH> strcuture
    RIL_CAPSTYPE_MSGMEMORYLOCATIONS      points to an <t RILCAPSMSGMEMORYLOCATIONS> structure
    RIL_CAPSTYPE_MSGCONFIGINDEXRANGE     points to an <t RILRANGE> structure
    RIL_CAPSTYPE_PREFOPINDEXRANGE        points to an <t RILRANGE> structure
    RIL_CAPSTYPE_SIMTOOLKITNOTIFICATIONS points to an <t TBD> SIMTOOLKIT structure 
    RIL_CAPSTYPE_HSCSD                   points to an <t RILCAPSHSCSD> structure
    RIL_CAPSTYPE_GPRS                    points to an <t RILCAPSGPRS> structure
    RIL_CAPSTYPE_RLP                     points to an array of <t RILAPSRLP> structures
    RIL_CAPSTYPE_LOCKINGPWDLENGTHS       points to an array of <t RILCAPSLOCKINGPWDLENGTH> structures
    RIL_CAPSTYPE_BARRINGPWDLENGTHS       points to an array of <t RILCAPSBARRINGPWDLENGTH> structures
    RIL_CAPSTYPE_CALLMGTCMDS             points to <t DWORD> containing a combination of <def RIL_CAPS_CALLCMD_> constants
    RIL_CAPSTYPE_EQUIPMENTSTATES         points to <t DWORD> containing a combination of <def RIL_CAPS_EQSTATE_> constants
    RIL_CAPSTYPE_PBSTORELOCATIONS        points to <t DWORD> containing a combination of <def RIL_CAPS_PBLOC_> constants
    RIL_CAPSTYPE_MSGSERVICETYPES         points to <t DWORD> containing a combination of <def RIL_CAPS_MSGSVCTYPE_> constants
    RIL_CAPSTYPE_BROADCASTMSGLANGS       points to <t DWORD> containing a combination of <def RIL_CAPS_DCSLANG_> constants
    RIL_CAPSTYPE_MSGSTATUSVALUES         points to <t DWORD> containing a combination of <def RIL_CAPS_MSGSTATUS_> constants
    RIL_CAPSTYPE_LOCKFACILITIES          points to <t DWORD> containing a combination of <def RIL_CAPS_LOCKFACILITY_> constants
    RIL_CAPSTYPE_BARRTYPES 		         points to <t DWORD> containing a combination of <def RIL_CAPS_BARRTYPE_> constants
    RIL_CAPSTYPE_FORWARDINGREASONS       points to <t DWORD> containing a combination of <def RIL_CAPS_FWDREASON_> constants
    RIL_CAPSTYPE_INFOCLASSES             points to <t DWORD> containing a combination of <def RIL_CAPS_INFOCLASS_> constants


RIL_GetGPRSAddress(HRIL hRil, DWORD dwContextID);   -> WCHAR*
RIL_GetSimRecordStatus(HRIL hRil, DWORD dwFileID);  -> RILSIMRECORDSTATUS

RIL_GetCCBSStatus(HRIL hRil, DWORD dwCCBSIndex);    -> char*
RIL_GetCallBarringStatus(HRIL hRil, DWORD dwType, DWORD dwInfoClass, LPCSTR lpszPassword);  -> RIL_BARRINGSTATUS
RIL_GetCallForwardingSettings(HRIL hRil, DWORD dwReason, DWORD dwInfoClass);    -> RILCALLFORWARDINGSETTINGS
RIL_GetCallWaitingSettings(HRIL hRil, DWORD dwInfoClass);   -> RIL_INFOCLASS_
RIL_GetCurrentOperator(HRIL hRil, DWORD dwFormat);  -> RILOPERATORNAMES
RIL_GetLockingStatus(HRIL hRil, DWORD dwFacility, LPCSTR lpszPassword); -> RIL_LOCKINGSTATUS_
RIL_GetPreferredOperatorList(HRIL hRil, DWORD dwFormat);    -> RILOPERATORINFO
RIL_ReadMsg(HRIL hRil, DWORD dwIndex);      -> RILMESSAGEINFO
RIL_ReadPhonebookEntries(HRIL hRil, DWORD dwStartIndex, DWORD dwEndIndex);  -> RILPHONEBOOKENTRY

RIL_GetCellBroadcastMsgConfig(HRIL hRil);   -> RILCBMSGCONFIG
RIL_GetAudioGain(HRIL hRil);                -> RILGAININFO
RIL_GetBearerServiceOptions(HRIL hRil);     -> RILBEARERSVCINFO
RIL_GetCallerIdSettings(HRIL hRil);         -> RILCALLERIDSETTINGS
RIL_GetCellTowerInfo(HRIL hRil);            -> RILCELLTOWERINFO
RIL_GetCostInfo(HRIL hRil);                 -> RILCOSTINFO
RIL_GetDataCompression(HRIL hRil);          -> RILDATACOMPINFO
RIL_GetDialedIdSettings(HRIL hRil);         -> RILDIALEDIDSETTINGS
RIL_GetErrorCorrection(HRIL hRil);          -> RILERRORCORRECTIONINFO
                                           
RIL_GetClosedGroupSettings(HRIL hRil);     
RIL_GetMsgConfig(HRIL hRil);                -> RILMSGCONFIG
RIL_GetMsgServiceOptions(HRIL hRil);        -> RILMSGSERVICEINFO
RIL_GetRLPOptions(HRIL hRil);               -> RILRLPINFO
RIL_GetSystemTime(HRIL hRil);               -> SYSTEMTIME
RIL_GetUserIdentity(HRIL hRil);             -> char*
RIL_GetLineStatus(HRIL hRil);               -> RIL_LINESTAT_
RIL_GetHideIdSettings(HRIL hRil);           -> RILHIDEIDSETTINGS
RIL_GetHideConnectedIdSettings(HRIL hRil);  -> RILHIDECONNECTEDIDSETTINGS


RIL_GetGPRSAttached(HRIL hRil);             -> BOOL
RIL_GetGPRSClass(HRIL hRil);                -> RIL_GPRSCLASS_
RIL_GetGPRSContextActivatedList(HRIL hRil); -> RILGPRSCONTEXTACTIVATED
RIL_GetGPRSContextList(HRIL hRil);          -> RILGPRSCONTEXT
RIL_GetGPRSRegistrationStatus(HRIL hRil);   -> RIL_REGSTAT_
RIL_GetHSCSDCallSettings(HRIL hRil);        -> RILCALLHSCSDINFO
RIL_GetHSCSDOptions(HRIL hRil);             -> RILHSCSDINFO
RIL_GetMOSMSService(HRIL hRil);             -> RIL_MOSMSSERVICE_
RIL_GetOperatorList(HRIL hRil);             -> RILOPERATORINFO
RIL_GetPhonebookOptions(HRIL hRil);         -> RILPHONEBOOKINFO
RIL_GetRegistrationStatus(HRIL hRil);       -> RIL_REGSTAT_
RIL_GetSimToolkitProfile(HRIL hRil);        -> BYTE*

RIL_GetMinimumQualityOfServiceList(HRIL hRil);  -> RILGPRSQOSPROFILE
RIL_GetRequestedQualityOfServiceList(HRIL hRil);    -> RILGPRSQOSPROFILE
*/
