/* (C) 2003 XDA Developers  itsme@xs4all.nl
 *
 * $Header: /var/lib/cvs/nah6utils/src/pdebug.cpp,v 1.7 2004/01/05 18:01:47 itsme Exp $
 */
#include <windows.h>
#include <tchar.h>
#include <stdio.h>
#include <string.h>
#include <rapi.h>
#include "args.h"
#include "ItsUtils.h"
#include "debug.h"
#include "dllversion.h"

#include <string>
using namespace std;

// !!! I am mixing up handles and processid's quite often.
//     on WinCE this is not a problem, but this needs to be
//     fixed before porting to another platform.


bool DebugProcess(DWORD dwProcessId, const WCHAR *szExeName, const WCHAR *szCmdLine);
DWORD ITGetProcessId(const char *szProcessName);
HANDLE ITGetProcessHandle(const char *szProcessName);

bool ITStartDebuggingProcess(DWORD dwProcessId, const WCHAR *wszExeName, const WCHAR *wszCmdLine, DWORD *pdwDebugger);
bool ITWaitForDebugEvent(DWORD dwDebugger, DebugEvent *ev, DWORD *pdwMissedEvents);
bool ITStopDebuggingProcess(DWORD dwDebugger);
void WriteDebugEvent(DebugEvent *ev);

wstring StringToWString(const string &str)
{
    WCHAR *ws= new WCHAR[str.length()+1];
    _snwprintf(ws, str.length()+1, L"%hs", str.c_str());
    wstring wstr(ws);
    delete ws;

    return wstr;
}

void usage()
{
    printf("Usage: pdebug [-x] [-h PID|HANDLE ] <processname> [cmdline args]\n");
}

int main(int argc, char *argv[])
{
    DebugStdOut();

    bool bStartProcess= false;
    DWORD dwProcessId= 0;
    string exename;
    string cmdline;

    int argsfound=0;
    for (int i=1 ; i<argc ; i++)
    {
        if (argv[i][0]=='-') switch(argv[i][1])
        {
            case 'h': HANDLEULOPTION(dwProcessId, DWORD); break;
            case 'x': bStartProcess= true; break;

            default:
                usage();
                return 1;
        }
        else if (argsfound==0)
        {
            exename= argv[i];
            argsfound++;
        }
        else
        {
            if (cmdline.size())
                cmdline += ' ';
            cmdline += argv[i];

            argsfound++;
        }
    }

    if ((argsfound==0 && dwProcessId==0)
            || (argsfound>1 && !bStartProcess))
    {
        usage();
        return 1;
    }

    CheckITSDll();

    if (dwProcessId==0 && !bStartProcess)
    {
        string processname;
        string::size_type ls= exename.find_last_of('\\');
        if (ls==string::npos)
            processname= exename;
        else
            processname= exename.substr(ls+1);

        dwProcessId= ITGetProcessId(processname.c_str());
    }
    DebugProcess(dwProcessId, StringToWString(exename).c_str(), StringToWString(cmdline).c_str());

    CeRapiUninit();

    return 0;
}


bool DebugProcess(DWORD dwProcessId, const WCHAR *szExeName, const WCHAR *szCmdLine)
{
    DWORD debugger;
    if (!ITStartDebuggingProcess(dwProcessId, szExeName, szCmdLine, &debugger))
    {
        debug("ERROR: ITStartDebuggingProcess failed\n");
        return false;
    }

    DebugEvent ev;
    DWORD missed= 0;
    DWORD prevmissed= 0;
    while (ITWaitForDebugEvent(debugger, &ev, &missed))
    {
        if (missed>prevmissed)
        {
            debug("MISSED %d events\n", prevmissed, missed);
            prevmissed= missed;
        }
        WriteDebugEvent(&ev);
    }
    if (!ITStopDebuggingProcess(debugger))
    {
        debug("ERROR: ITStopDebuggingProcess failed\n");
        return false;
    }

    return true;
}
bool DebugProcess(HANDLE hProc)
{
	return DebugProcess((DWORD)hProc, NULL, NULL);
}

bool ITStartDebuggingProcess(DWORD dwProcessId, const WCHAR *wszExeName, const WCHAR *wszCmdLine, DWORD *pdwDebugger)
{
    StartDebuggingProcessParams inbuf;
    memset(&inbuf, 0, sizeof(inbuf));

    inbuf.dwProcessId= dwProcessId;
    if (dwProcessId==0)
    {
        if (wszExeName)
            wcsncpy(inbuf.wszExeName, wszExeName, sizeof(inbuf.wszExeName)/sizeof(WCHAR));
        if (wszCmdLine)
            wcsncpy(inbuf.wszCmdLine, wszCmdLine, sizeof(inbuf.wszCmdLine)/sizeof(WCHAR));
    }

    StartDebuggingProcessResult *outbuf;
    DWORD outsize= 0;

    HRESULT res= CeRapiInvoke(L"\\Windows\\ItsUtils.dll", L"ITStartDebuggingProcess",
            sizeof(StartDebuggingProcessParams), (BYTE*)&inbuf,
            &outsize, (BYTE**)&outbuf, NULL, 0);
    if (res)
    {
        error(res, "ITStartDebuggingProcess(%08lx, %ls, %ls)", dwProcessId, wszExeName, wszCmdLine);
        return false;
    }

    *pdwDebugger= outbuf->debugger;

    LocalFree(outbuf);

    return true;
}
bool ITWaitForDebugEvent(DWORD dwDebugger, DebugEvent *ev, DWORD *pdwMissedEvents)
{
    WaitForDebugEventParams inbuf;
    WaitForDebugEventResult *outbuf=NULL;
    DWORD outsize=0;

    inbuf.debugger= dwDebugger;

    HRESULT res= CeRapiInvoke(L"\\Windows\\ItsUtils.dll", L"ITWaitForDebugEvent",
            sizeof(WaitForDebugEventParams), (BYTE*)&inbuf,
            &outsize, (BYTE**)&outbuf, NULL, 0);
    if (res)
    {
        error(res, "ITWaitForDebugEvent");
        return false;
    }

    memcpy(ev, &outbuf->ev, sizeof(DebugEvent));
    *pdwMissedEvents= outbuf->dwMissedEvents;

    LocalFree(outbuf);

    return true;
}

bool ITStopDebuggingProcess(DWORD dwDebugger)
{
    StopDebuggingProcessParams inbuf;

    inbuf.debugger= dwDebugger;

    HRESULT res= CeRapiInvoke(L"\\Windows\\ItsUtils.dll", L"ITStopDebuggingProcess",
            sizeof(StopDebuggingProcessParams), (BYTE*)&inbuf,
            NULL, NULL, NULL, 0);
    if (res)
    {
        error(res, "ITStopDebuggingProcess");
        return false;
    }

    return true;
}

void WriteDebugEvent(DebugEvent *ev)
{
    debug("%08lx:%08lx %d %hs\n", ev->dwProcessId, ev->dwThreadId, ev->dwDebugEventCode, ev->description);
}

DWORD ITGetProcessId(const char *szProcessName)
{
    return (DWORD)ITGetProcessHandle(szProcessName);
}
HANDLE ITGetProcessHandle(const char *szProcessName)
{
    DWORD insize= (strlen(szProcessName)+1)*sizeof(WCHAR);
    WCHAR *inbuf= (WCHAR*)LocalAlloc(LPTR, insize);

	_snwprintf((WCHAR*)inbuf, insize/sizeof(WCHAR), L"%hs", szProcessName);

    DWORD outsize=0;
    HANDLE *outbuf=NULL;

    HRESULT res= CeRapiInvoke(L"\\Windows\\ItsUtils.dll", L"ITGetProcessHandle",
            insize, (BYTE*)inbuf,
            &outsize, (BYTE**)&outbuf, NULL, 0);
    if (res) {
        error(res, "ITGetProcessHandle");
        return INVALID_HANDLE_VALUE;
    }

    HANDLE hproc= *outbuf;

    LocalFree(outbuf);

    return hproc;
}


