/* (C) 2003 XDA Developers  itsme@xs4all.nl
 *
 * $Header: /var/lib/cvs/xdautils/src/pdir.cpp,v 1.9 2004/02/10 22:03:09 itsme Exp $
 */
#include <windows.h>
#include <stdio.h>
#include <string.h>
#include <rapi.h>

#include "debug.h"
#include "args.h"

#include "vectorutils.h"
#include "stringutils.h"
#include <map>

bool g_bRecurse= false;

typedef std::map<std::string,std::string> StringStringMap;

#define AT_NONEXISTANT      1
#define AT_ISDIRECTORY      2
#define AT_ISFILE           3

int getCeAttributes(const std::string& name)
{
    DWORD dwAttr = CeGetFileAttributes( ToWString(name).c_str() );
    if (0xFFFFFFFF == dwAttr)
        return AT_NONEXISTANT;

    if (dwAttr & FILE_ATTRIBUTE_DIRECTORY)
        return AT_ISDIRECTORY;

    return AT_ISFILE;
}
bool isCeDirectory(const std::string& name)
{
    return getCeAttributes(name)==AT_ISDIRECTORY;
}
bool isCeFile(const std::string& name)
{
    return getCeAttributes(name)==AT_ISFILE;
}
bool isCeNonExistant(const std::string& name)
{
    return getCeAttributes(name)==AT_NONEXISTANT;
}
std::string concat_path(const std::string& p1, const std::string& p2)
{
    std::string result(p1);

    if (result.size() && result[result.size()-1]!='/' && result[result.size()-1]!='\\')
        result += '\\';

    result += p2;

    return result;
}

std::string GetBasePath(const std::string& name)
{
    size_t lastslash= name.find_last_of("\\/");
    if (lastslash==name.npos)
        return "";

    return name.substr(0, lastslash);
}

std::string MakeFileLine(const CE_FIND_DATA& wfd, const std::string& name)
{
    std::string line;
    SYSTEMTIME systime;
    FILETIME lfiletime;
    FileTimeToLocalFileTime(&wfd.ftLastWriteTime, &lfiletime);
    FileTimeToSystemTime(&lfiletime, &systime);
    line= stringformat("%04d-%02d-%02d %02d:%02d:%02d.%03d",
            systime.wYear, systime.wMonth, systime.wDay,
            systime.wHour, systime.wMinute, systime.wSecond,systime.wMilliseconds);

    if(wfd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
        line += stringformat(" <DIR>   %s\\", name.c_str());
    else {
// ignoring the possibility of 4G+ files on portable devices for now.
        line += stringformat(" %7d %s", wfd.nFileSizeLow, name.c_str());
    }

    return line;
}

bool CeFileGlob(const std::string& name, StringStringMap& filelist, StringStringMap& dirlist)
{
    CE_FIND_DATA wfd;
    HANDLE hFind = CeFindFirstFile( ToWString(name).c_str(), &wfd);
    if (INVALID_HANDLE_VALUE == hFind)
        return false;

    std::string rootpath= GetBasePath(name);
    do {
        std::string fullpath= concat_path(rootpath, ToString(wfd.cFileName));
        if (wfd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
            dirlist[fullpath]= MakeFileLine(wfd, fullpath);
        else
            filelist[fullpath]= MakeFileLine(wfd, fullpath);
    } while (CeFindNextFile(hFind, &wfd));

    CeFindClose(hFind);

    return true;
}

void ProcessDirectory(const std::string& name, int level= 0)
{
    StringStringMap files;
    StringStringMap subdirs;

    if (!CeFileGlob(concat_path(name,"*"), files, subdirs))
        return;

    for (StringStringMap::iterator i= files.begin() ; i!=files.end() ; ++i)
        printf("%s\n", (*i).second.c_str());

    for (StringStringMap::iterator i= subdirs.begin() ; i!=subdirs.end() ; ++i) {
        printf("%s\n", (*i).second.c_str());
        if (g_bRecurse) {
            ProcessDirectory((*i).first, level+1);
        }
    }
}
bool ProcessFile(const std::string& name)
{
    CE_FIND_DATA wfd;
    HANDLE hFind = CeFindFirstFile( ToWString(name).c_str(), &wfd);
    if (INVALID_HANDLE_VALUE == hFind)
        return false;

    printf("%s\n", MakeFileLine(wfd, name).c_str());

    return true;
}
bool hasWildCards(const std::string& name)
{
    return name.find_first_of("*?")!=name.npos;
}

void ProcessName(const std::string& name)
{

    int type= getCeAttributes(name);
    if (type==AT_ISDIRECTORY)
        ProcessDirectory(name);
    else if (type==AT_ISFILE)
        ProcessFile(name);
    else
        printf("%s not found\n", name.c_str());
}

void usage()
{
    printf("Usage: pdir [-r] dir ...\n");
}
int main( int argc, char *argv[])
{
    DebugStdOut();

    StringList dirlist;

    for (int i=1 ; i<argc ; i++)
    {
        if (argv[i][0]=='-') switch(argv[i][1])
        {
            case 'r': g_bRecurse= true; break;
            default:
                usage();
                return 1;
        }
        else
        {
            dirlist.push_back(argv[i]);
        }
    }
    if (dirlist.empty())
        dirlist.push_back("");      // empty string for root.

    if (FAILED(CeRapiInit()))
    {
        printf( "Failed\n");
        return 1;
    }

    for (StringList::iterator i= dirlist.begin() ; i!=dirlist.end(); ++i)
    {
        if (hasWildCards(*i)) {
            StringStringMap files;
            StringStringMap subdirs;
            if (!CeFileGlob(*i, files, subdirs))
            {
                debug("error matching wildcard %s\n", (*i).c_str());
                continue;
            }
            for (StringStringMap::iterator i= files.begin() ; i!=files.end() ; ++i)
                printf("%s\n", (*i).second.c_str());
            for (StringStringMap::iterator i= subdirs.begin() ; i!=subdirs.end() ; ++i)
                printf("%s\n", (*i).second.c_str());
        }
        else {
            ProcessName(*i);
        }
    }

    CeRapiUninit();
}

