/* (C) 2003 XDA Developers  itsme@xs4all.nl
 *
 * $Header: /var/lib/cvs/xdautils/src/pmemdump.cpp,v 1.23 2004/08/20 12:06:38 itsme Exp $
 */
#include <windows.h>
#include "ItsUtils.h"
#include "dllversion.h"
#include "debug.h"
#include "stringutils.h"
#include "args.h"
#include <stdio.h>
#include <string.h>
#include <rapi.h>

DumpUnitType g_dumpunit=DUMPUNIT_BYTE;
DumpFormat g_dumpformat= DUMP_HEX_ASCII;
int g_nMaxUnitsPerLine=-1;
int g_nStepSize= 0;
DWORD g_blocksize= 0x10000;
bool g_verbose= false;
bool g_fulldump= false;
bool g_showerrors= true;

HANDLE GetRapiProcessHandle();
void CopyProcessMemoryToFile(HANDLE hProc, DWORD dwOffset, DWORD dwLength, char *szOutfile);
void StepProcessMemoryToStdout(HANDLE hProc, DWORD dwOffset, DWORD dwLength);
void DumpProcessMemoryToStdout(HANDLE hProc, DWORD dwOffset, DWORD dwLength);
HANDLE ITGetProcessHandle(char *szProcessName);

void usage()
{
    printf("Usage: pmemdump [ -m | -p procname | -h prochandle] start length [ filename ]\n");
    printf("    numbers can be specified as 0x1234abcd\n");
    printf("    -1 -2 -4 : dump as bytes/words/dwords\n");
    printf("    -w NUM : specify nr of words per line\n");
    printf("    -s SIZE: step with SIZE through memory\n");
    printf("    -a     : ascdump iso hexdump\n");
    printf("    -f     : full - do not summarize identical lines\n");
    printf("    -c     : print raw memory to stdout\n");
    printf("    -x     : print only hex\n");
    printf("    -xx    : print only fixed length ascii dumps\n");
    printf("    -v     : verbose\n");
    printf("\n");
    printf("    -n NAME: view memory in the context of process NAME\n");
    printf("    -h NUM : view memory in the context of process with handle NUM\n");
    printf("    -m     : directly access memory - not using ReadProcessMemory\n");
    printf("    -p     : access physical memory, instead of virtual memory\n");
    printf("        if neither -p, -h or -m is specified, memory is read from the context\n");
    printf("        of rapisrv.exe\n");
    printf("\n");
}

int main( int argc, char *argv[])
{
    DebugStdOut();

    DWORD dwOffset=0;
    DWORD dwLength=0;
    char *szOutfile=NULL;
    char *szProcessName= NULL;
    HANDLE hProc= INVALID_HANDLE_VALUE;
    bool bDirectMemoryAccess= false;    // false: use readprocessmemory, true: use readmemory
    bool bPhysicalMemoryAccess= false;    // false: use readprocessmemory, true: use readmemory
    int nDumpUnitSize= 1;

    int argsfound=0;
    for (int i=1 ; i<argc ; i++)
    {
        if (argv[i][0]=='-') switch(argv[i][1])
        {
            case 'h': HANDLEULOPTION(hProc, HANDLE); break;
            case 'n': HANDLESTROPTION(szProcessName); break;
            case 'm': bDirectMemoryAccess= true; break;
            case 'p': bPhysicalMemoryAccess= true; break;
            case 'v': g_verbose= true; break;
            case 'a': g_dumpformat= DUMP_STRINGS; break;
            case 'c': g_dumpformat= DUMP_RAW; break;
            case 'x': if (argv[i][2]=='x')
                          g_dumpformat= DUMP_ASCII; 
                      else
                          g_dumpformat= DUMP_HEX; 
                      break;
            case 'f': g_fulldump= true; break;
            case 'w': HANDLEULOPTION(g_nMaxUnitsPerLine, DWORD); break;
            case 's': HANDLEULOPTION(g_nStepSize, DWORD); break;
            case 'b': HANDLEULOPTION(g_blocksize, DWORD); break;

            case '1': case '2': case '4':
                nDumpUnitSize= argv[i][1]-'0';
                break;
            default:
                usage();
                return 1;
        }
        else switch (argsfound++)
        {
            case 0: dwOffset= strtoul(argv[i], 0, 0); break;
            case 1: dwLength= strtoul(argv[i], 0, 0); break;
            case 2: szOutfile= argv[i]; break;
        }
    }

    if (argsfound==0 || argsfound>3)
    {
        usage();
        return 1;
    }
    if (argsfound==1)
        dwLength= 0x100;

    if (g_nMaxUnitsPerLine<0) {
        if (g_dumpformat==DUMP_ASCII) 
            g_nMaxUnitsPerLine= 64/nDumpUnitSize;
        else if (g_dumpformat==DUMP_HEX) 
            g_nMaxUnitsPerLine= 32/nDumpUnitSize;
        else
            g_nMaxUnitsPerLine= 16/nDumpUnitSize;
    }

    g_dumpunit= 
        nDumpUnitSize==1?DUMPUNIT_BYTE:
        nDumpUnitSize==2?DUMPUNIT_WORD:
        nDumpUnitSize==4?DUMPUNIT_DWORD:DUMPUNIT_BYTE;



    CheckITSDll();

    if (hProc!=INVALID_HANDLE_VALUE)
    {
        // - do nothing, process handle already there.
    }
    else if (szProcessName==NULL)
    {
        hProc= GetRapiProcessHandle();
        if (hProc==INVALID_HANDLE_VALUE) {
            debug("error getting process context\n");
            return 1;
        }
    }
    else {
        hProc= ITGetProcessHandle(szProcessName);
        if (hProc==INVALID_HANDLE_VALUE) {
            debug("error getting process context\n");
            return 1;
        }
    }
    if (bDirectMemoryAccess)
        hProc= NULL;
    else if (bPhysicalMemoryAccess)
        hProc= INVALID_HANDLE_VALUE;

    if (g_nStepSize)
        StepProcessMemoryToStdout(hProc, dwOffset, dwLength);
    else if (szOutfile==NULL)
        DumpProcessMemoryToStdout(hProc, dwOffset, dwLength);
    else
        CopyProcessMemoryToFile(hProc, dwOffset, dwLength, szOutfile);

    CeRapiUninit();

    return 0;
}

HANDLE GetRapiProcessHandle()
{
    DWORD outsize=0;
    GetContextResult *r=NULL;

    HRESULT res= CeRapiInvoke(L"\\Windows\\ItsUtils.dll", L"ITGetContext",
            0, NULL, &outsize, (BYTE**)&r, NULL, 0);
    if (res)
    {
        error(res, "ITGetContext");
        return INVALID_HANDLE_VALUE;
    }

    HANDLE hProc= r->hProcess;
    LocalFree(r);
    return hProc;
}
HANDLE ITGetProcessHandle(char *szProcessName)
{
    DWORD insize= (strlen(szProcessName)+1)*sizeof(WCHAR);
    WCHAR *inbuf= (WCHAR*)LocalAlloc(LPTR, insize);

	_snwprintf((WCHAR*)inbuf, insize/sizeof(WCHAR), L"%hs", szProcessName);

    DWORD outsize=0;
    HANDLE *outbuf=NULL;

    HRESULT res= CeRapiInvoke(L"\\Windows\\ItsUtils.dll", L"ITGetProcessHandle",
            insize, (BYTE*)inbuf,
            &outsize, (BYTE**)&outbuf, NULL, 0);
    if (res) {
        error(res, "ITGetProcessHandle");
        return INVALID_HANDLE_VALUE;
    }

    HANDLE hproc= *outbuf;

    LocalFree(outbuf);

    return hproc;
}

bool ITReadProcessMemory(HANDLE hProc, DWORD dwOffset, BYTE *buffer, DWORD dwBytesWanted, DWORD *pdwNumberOfBytesRead)
{
    ReadProcessMemoryParams inbuf;
    DWORD outsize=0;
    ReadProcessMemoryResult *outbuf=NULL;

    inbuf.hProcess= hProc;
    inbuf.dwOffset= dwOffset;
    inbuf.nSize= dwBytesWanted;
    outbuf= NULL; outsize= 0;
    HRESULT res= CeRapiInvoke(L"\\Windows\\ItsUtils.dll", L"ITReadProcessMemory",
            sizeof(ReadProcessMemoryParams), (BYTE*)&inbuf,
            &outsize, (BYTE**)&outbuf, NULL, 0);
    if (res)
    {
        if (g_showerrors)
            error(res, "ITReadProcessMemory");
        return false;
    }
    memcpy(buffer, &outbuf->buffer, outbuf->dwNumberOfBytesRead);
    *pdwNumberOfBytesRead= outbuf->dwNumberOfBytesRead;
    LocalFree(outbuf);

    return true;
}

void StepProcessMemoryToStdout(HANDLE hProc, DWORD dwOffset, DWORD dwLength)
{
    ByteVector buffer;
    std::string prevline;
    bool bSamePrinted= false;

    g_showerrors= false;
    while (dwLength)
    {
        buffer.resize(DumpUnitSize(g_dumpunit)*g_nMaxUnitsPerLine);

        DWORD dwBytesWanted= min(dwLength, buffer.size());
        DWORD dwNumberOfBytesRead;
        std::string line;
        if (!ITReadProcessMemory(hProc, dwOffset, vectorptr(buffer), dwBytesWanted, &dwNumberOfBytesRead)) {
            line= " * * * * * *\n"; // indicates invalid memory
        }
        else {
            if (g_dumpformat==DUMP_RAW) {
                line.clear();
            }
            else if (g_dumpformat==DUMP_STRINGS)
                line= ascdump(buffer);
            else if (g_dumpformat==DUMP_ASCII)
                line= asciidump(vectorptr(buffer), dwNumberOfBytesRead)+"\n";
            else
                line= hexdump(dwOffset, vectorptr(buffer), dwNumberOfBytesRead, DumpUnitSize(g_dumpunit), g_nMaxUnitsPerLine).substr(9);
        }

        if (g_dumpformat==DUMP_RAW)
            fwrite(vectorptr(buffer), 1, buffer.size(), stdout);
        else if (!g_fulldump && line == prevline) {
            if (!bSamePrinted && line != " * * * * * *\n")
                printf("*\n", dwOffset);
            bSamePrinted= true;
        }
        else {
            bSamePrinted= false;

            printf("%08lx: %hs", dwOffset, line.c_str());
        }
        prevline= line;
        DWORD dwStep= min(dwLength, g_nStepSize);
        dwLength -= dwStep;
        dwOffset += dwStep;
    }
    g_showerrors= true;
}
void DumpProcessMemoryToStdout(HANDLE hProc, DWORD dwOffset, DWORD dwLength)
{
    ByteVector buffer;

    DWORD flags= hexdumpflags(g_dumpunit, g_nMaxUnitsPerLine, g_dumpformat)
        | (g_fulldump?0:HEXDUMP_SUMMARIZE) | (g_dumpformat==DUMP_RAW?0:HEXDUMP_WITH_OFFSET);

    while (dwLength)
    {
        buffer.resize(g_blocksize);
        DWORD dwWanted= min(dwLength, buffer.size());
        DWORD dwNumberOfBytesRead;
        std::string line;
        if (!ITReadProcessMemory(hProc, dwOffset, vectorptr(buffer), dwWanted, &dwNumberOfBytesRead))
            return;

        buffer.resize(dwNumberOfBytesRead);

        bighexdump(dwOffset, buffer, flags| (dwLength!=dwNumberOfBytesRead ? HEXDUMP_MOREFOLLOWS : 0));

        dwLength -= dwNumberOfBytesRead;
        dwOffset += dwNumberOfBytesRead;
    }
}
std::string hhmmss(DWORD s)
{
    return stringformat("%2d:%02d:%02d", s/3600, (s/60)%60, s%60);
}
void CopyProcessMemoryToFile(HANDLE hProc, DWORD dwStartOffset, DWORD dwLength, char *szOutfile)
{
    debug("CopyProcessMemoryToFile(%08lx, %08lx, %08lx, %s)\n",
            hProc, dwStartOffset, dwLength, szOutfile);
    HANDLE hDest = CreateFile( szOutfile, GENERIC_WRITE, FILE_SHARE_READ,
                NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
    if (INVALID_HANDLE_VALUE == hDest)
    {
        error("Unable to open host/destination file");
        return;
    }

    DWORD t0= GetTickCount();
    DWORD tprev= t0;
    ByteVector buffer;
    DWORD dwOffset= dwStartOffset;
    while (dwLength)
    {
        buffer.resize(g_blocksize);
        DWORD dwWanted= min(dwLength, buffer.size());
        DWORD dwNumberOfBytesRead;
        if (!ITReadProcessMemory(hProc, dwOffset, vectorptr(buffer), dwWanted, &dwNumberOfBytesRead))
            return;

        DWORD dwNumWritten;
        if (!WriteFile(hDest, vectorptr(buffer), dwNumberOfBytesRead, &dwNumWritten, NULL))
        {
            error("Error Writing file");
            return;
        }

        dwLength -= dwNumberOfBytesRead;
        dwOffset += dwNumberOfBytesRead;

        DWORD t= GetTickCount();
        if (g_verbose && t-tprev>2000) {
            double bps= (double)1000.0*(dwOffset-dwStartOffset)/(t-t0);
            printf("read %08lx bytes in %6d msec : %8.0f bytes/sec - timeleft: %hs\r", 
                    dwOffset-dwStartOffset, t-t0, bps, hhmmss(dwLength/bps).c_str());

            tprev= t;
        }
    }
    DWORD t1= GetTickCount();
    if (g_verbose) {
        printf("read %08lx bytes in %6d msec : %8.0f bytes/sec\n", 
                dwOffset-dwStartOffset, t1-t0, (double)1000.0*(dwOffset-dwStartOffset)/(t1-t0));
    }

    CloseHandle (hDest);
}
