/* (C) 2003 XDA Developers  itsme@xs4all.nl
 *
 * $Header: /var/lib/cvs/nah6utils/src/pnewbmp.cpp,v 1.6 2004/01/08 00:29:17 itsme Exp $
 */
#include <windows.h>
#include "ItsUtils.h"
#include "dllversion.h"

#include "debug.h"
#include "args.h"
#include <stdio.h>
#include <string.h>
#include <rapi.h>

#include <vector>
using namespace std;

bool ITFlashROM(DWORD dwOffset, BYTE *buffer, DWORD dwBufLength);
bool ReadFromFile(char *filename, DWORD dwOffset, BYTE *buffer, DWORD dwSize);
DWORD CalcCheckSum(BYTE *buffer, DWORD dwSize);

void usage()
{
    printf("Usage: { -3 | -4 } pnewbmp filename\n");
}

int main( int argc, char *argv[])
{
    DebugStdOut();

    char *filename=NULL;
    DWORD dwBmpOffset= 0;

    int argsfound=0;
    for (int i=1 ; i<argc ; i++)
    {
        if (argv[i][0]=='-') switch(argv[i][1])
        {
            case '4': dwBmpOffset= 0x81ec0000; break;
            case '3': dwBmpOffset= 0x81900000; break;
            case 'o': HANDLEULOPTION(dwBmpOffset, DWORD); break;
            default:
                usage();
                return 1;
        }
        else switch (argsfound++)
        {
            case 0: filename= argv[i]; break;
        }
    }

    if (argsfound!=1)
    {
        usage();
        return 1;
    }
    if (dwBmpOffset==0)
    {
        debug("ERROR: no version specified ( -4 or -3 )\n");
        return 1;
    }

    BYTE buffer[0x25800]; 
    if (!ReadFromFile(filename, 0x46, buffer, 0x25800))
    {
        debug("ERROR reading bitmap\n");
        return 1;
    }


    CheckITSDll();

    if (!ITFlashROM(dwBmpOffset, buffer, 0x25800))
        debug("ERROR in ITFlashROM\n");

    CeRapiUninit();

    return 0;
}

bool ITFlashROM(DWORD dwOffset, BYTE *buffer, DWORD dwBufLength)
{
    DWORD insize= sizeof(FlashROMParams)+dwBufLength;
    FlashROMParams *req= (FlashROMParams*)LocalAlloc(LPTR, insize);

    req->dwOffset= dwOffset;
    req->dwSize= dwBufLength;
    memcpy(req->buffer, buffer, dwBufLength);
    req->dwCheckSum= CalcCheckSum(req->buffer, req->dwSize);

    DWORD outsize=0;
    FlashROMResult *outbuf=NULL;

    HRESULT res= CeRapiInvoke(L"\\Windows\\ItsUtils.dll", L"ITFlashROM",
            insize, (BYTE*)req,
            &outsize, (BYTE**)&outbuf, NULL, 0);
    if (res)
    {
        error(res, "ITFlashROM(%08lx, %08lx)", dwOffset, dwBufLength);
        return false;
    }
    debug("result: %08lx %08lx\n", outbuf->dwResult, outbuf->dwReturned);
    LocalFree(outbuf);

    return true;
}


bool ReadFromFile(char *filename, DWORD dwOffset, BYTE *buffer, DWORD dwSize)
{
    FILE *f= fopen(filename, "rb");
    if (f==NULL)
    {
        error("fopen(%hs)", filename);
        return false;
    }
    if (0!=fseek(f, dwOffset, SEEK_SET))
    {
        error("fseek(%hs, %08lx)", filename, dwOffset);
        fclose(f);
        return false;
    }
    if (1!=fread(buffer, dwSize, 1, f))
    {
        error("fread(%hs, %08lx)", filename, dwSize);
        fclose(f);
        return false;
    }

    fclose(f);
    return true;
}

DWORD CalcCheckSum(BYTE *buffer, DWORD dwSize)
{
    DWORD sum= 0;
    while (dwSize--)
        sum+=*buffer++;

    return sum;
}

