/* (C) 2003 XDA Developers  itsme@xs4all.nl
 *
 * $Header: /var/db/cvs/xdautils/src/pnewbootloader.cpp,v 1.11 2004/12/01 10:07:01 itsme Exp $
 */
#include <windows.h>
#include "ItsUtils.h"
#include "dllversion.h"

#include "debug.h"
#include "stringutils.h"
#include "args.h"
#include <stdio.h>
#include <string.h>
#include <rapi.h>

#include <vector>
using namespace std;

char *msgFlashStartErase= "FLASH: Start erase address %xh\r\n";
char *msgFlashErase= "FLASH: +FlashErase Start=%Xh, Len=%Xh\n";

bool bVerbose= false;

// to flash the bootloader, we have to unprotect it:
//

// look where the strings 'FLASH: Start erase address' and 'FLASH: +FlashErase
// Start' are referenced, between these there should be 4 consequetive RAM
// addresses, the lowest is the bootloader address
// these 4 addresses contain: 
//       bootloaderstart, bootloadersize, kernelstart, kernelsize.
//
//  dwAddressList points to a section just after the subroutine that erases flash blocks.
//       DCD  msgFlashStartErase   ; points to string
//       DCD  pdwBootloaderStart   ; == dwBootloaderAddress from struct below.
//       DCD  pdwBootloaderSize
//       DCD  pdwKernelStart
//       DCD  pdwKernelSize

//  these are all consequetive addresses.
//
// dwBootloaderStart pointed to by pdwBootloaderStart to a list in RAM of the following:
//       DCD 0x80000000
// dwBootloaderSize
//       DCD 0x00040000
// dwKernelStart
//       DCD 0x80040000
// dwKernelSize
//       DCD 0x00150000
// dwRomStart
//       DCD 0x80000000
// dwRomSize
//       DCD 0x02000000
//
struct versiondata {
    DWORD dwVersion;
    WCHAR *versionstring;
    DWORD dwAddressList;        // points to pointer to 'FLASH: ...' text
    DWORD dwBootloaderAddress;  // 2nd address from the above list
} versions[]= {
    // xda1 / wallaby
    { 0x80041918, L"3.04.00 ENG", 0x8005FDCC, 0x8C0C60E8   },
    { 0x80055434, L"3.16.52 ENG", 0x80086610, 0x8C0BA204   },
    { 0x80055434, L"3.17.03 ENG", 0x80086648, 0x8C0BA204   },
    { 0x80055414, L"3.18.04 ENG", 0x80086628, 0x8C0BA204   },
    { 0x80055414, L"3.19.01 GER", 0x80086628, 0x8C0BA204   },
    { 0x80055be4, L"4.00.10 ENG", 0x80096F48, 0x8C0D62D8   },
    { 0x80055be4, L"4.00.01 ENG", 0x80096184, 0x8C0D62D8   },
    { 0x80055be4, L"4.00.05 ENG", 0x80096F48, 0x8C0D62D8   },
    { 0x80055be4, L"4.00.16 ENG", 0x80096F48, 0x8C0D62D8   },

    // xda2 / himalaya
    { 0x800433B0, L"1.60.00WWE",  0x8008DAF8, 0x901C63D0   },
    { 0x800433B0, L"1.52.00WWE",  0x8008DAF8, 0x901C63D0   },
    { 0x800433B0, L"1.66.00WWE",  0x80091734, 0x901c6f68   },
    { 0x800433B0, L"1.66.01GER",  0x80091734, 0x901c6f68   },
    { 0x800433B0, L"1.72.00WWE",  0x80091744, 0x901C6F68   },

    // i-mate jam / magician
    { 0x80043670, L"0.68.00 WWE", 0x8008C498, 0x905C63D0   },
    { 0x800435D0, L"1.00.00 WWE", 0x8008B794, 0x906663D4   },
};
#define N_VERSIONS (sizeof(versions)/sizeof(*versions))

bool ITFlashROM(DWORD dwOffset, BYTE *buffer, DWORD dwBufLength);
bool ITWriteProcessMemory(HANDLE hProc, DWORD dwOffset, BYTE *buffer, DWORD dwBufLength, DWORD *pdwNumberOfBytesWritten);
bool ITReadProcessMemory(HANDLE hProc, DWORD dwOffset, BYTE *buffer, DWORD dwBytesWanted, DWORD *pdwNumberOfBytesRead);

bool DisableWriteProtect(DWORD dwFlashInfoOffset, DwordVector& orig);
bool EnableWriteProtect(DWORD dwFlashInfoOffset, const DwordVector& orig);
bool FindFlashInfoOffset(DWORD& dwFlashInfoOffset);

bool ReadFromFile(char *filename, DWORD dwOffset, DWORD dwSize, ByteVector& buffer);
DWORD CalcCheckSum(BYTE *buffer, DWORD dwSize);

void usage()
{
    printf("Usage: pnewbootloader [-f] [-o offset] [-l length] [-s fileoffset] filename\n");
}

int main( int argc, char *argv[])
{
    DebugStdOut();

    char *filename=NULL;
    bool bForce= false;

    DWORD dwRomOffset= 0x80000000;
    DWORD dwLength= 0x40000;
    DWORD dwFileOffset= 0;

    bool bBootloaderTool= strstr(argv[0], "pnewbootloader")!=0;

    int argsfound=0;
    for (int i=1 ; i<argc ; i++)
    {
        if (argv[i][0]=='-') switch(argv[i][1])
        {
            case 'f': bForce= true; break;
            case 'v': bVerbose= true; break;
            case 'o': HANDLEULOPTION(dwRomOffset, DWORD); break;
            case 's': HANDLEULOPTION(dwFileOffset, DWORD); break;
            case 'l': HANDLEULOPTION(dwLength, DWORD); break;
            default:
                usage();
                return 1;
        }
        else switch (argsfound++)
        {
            case 0: filename= argv[i]; break;
        }
    }

    if (argsfound!=1)
    {
        usage();
        return 1;
    }

    if (bBootloaderTool && MessageBox(0,
"Do not run this tool unless :\n"
"    - you want to downgrade your 5.17 bootloader to 5.15\n"
"    - you have somehow managed to overwrite your bootloader and still\n"
"      have a working ( ce bootable ) device, and still have the operating\n"
"      system running\n"
"\n"
"If the above does not apply to you, press NO\n",
"WARNING", MB_YESNO)!=IDYES)
        return 1;

    ByteVector buffer; 
    if (!ReadFromFile(filename, dwFileOffset, dwLength, buffer))
    {
        debug("ERROR reading bootloader\n");
        return 1;
    }

    // verify that input file really resembles a bootloader
    if (bBootloaderTool && !bForce && !(buffer[3]==0xea 
            && buffer[64]==69 && buffer[65]==67 && buffer[66]==69 && buffer[67]==67     // 'ECEC'
            && ( (buffer[68]==0 && buffer[69]==0 && buffer[70]==10 && buffer[71]==0x8c) // xda1
                || (buffer[68]==0 && buffer[69]==64 && buffer[70]==6 && buffer[71]==0x90)   // xda2
                )
            )) 
    {
        debug("this is possibly not a bootloader!!\n");
        debug("%hs\n", hexdump(dwRomOffset, vectorptr(buffer), 80, 4, 4).c_str());
        return 1;
    }

    CheckITSDll();

    DWORD dwFlashInfoOffset;
    if (!FindFlashInfoOffset(dwFlashInfoOffset))
    {
        debug("Unable to find flash info offset, cannot disable bootloader writeprotect\n");
        return 1;
    }
    debug("protection found at %08lx\n", dwFlashInfoOffset);

    DwordVector protectparams;
    if (!DisableWriteProtect(dwFlashInfoOffset, protectparams))
    {
        debug("unable to disable bootloader writeprotect\n");
        return 1;
    }

    if (!ITFlashROM(dwRomOffset, vectorptr(buffer), buffer.size()))
        debug("ERROR in ITFlashROM\n");

    if (!EnableWriteProtect(dwFlashInfoOffset, protectparams))
    {
        debug("unable to re-enable bootloader writeprotect\n");
        return 1;
    }
    CeRapiUninit();

    return 0;
}

bool FindFlashInfoOffset(DWORD& dwFlashInfoOffset)
{
    for (int i=0 ; i<N_VERSIONS ; i++)
    {
        WCHAR vstr[16];
        DWORD nRead;
        if (!ITReadProcessMemory(NULL, versions[i].dwVersion, (BYTE*)vstr, sizeof(vstr), &nRead))
            return false;
        if (wcscmp(vstr, versions[i].versionstring)) {

            if (bVerbose) {
                printf("%08lx:  no version '%ls' but\n   %hs\n\n", 
                        versions[i].dwVersion, 
                        versions[i].versionstring, 
                        ascdump(ByteVector((BYTE*)vstr, (BYTE*)(vstr+8))).c_str()); 
            }

            continue;
        }

        DWORD addrs[5];
        if (!ITReadProcessMemory(NULL, versions[i].dwAddressList, (BYTE*)addrs, sizeof(addrs), &nRead))
            return false;

        if (bVerbose) {
            printf("%08lx: addrlist: %hs\n", 
                    versions[i].dwAddressList, 
                    hexdump(ByteVector((BYTE*)addrs, (BYTE*)(addrs+5)), 4).c_str());
        }
        char txt[256];
        if (!ITReadProcessMemory(NULL, addrs[0], (BYTE*)txt, sizeof(txt), &nRead))
            continue;
        if (strcmp(txt, msgFlashStartErase)) {

            if (bVerbose) {
                printf("%08lx:  no txt '%hs' but\n   %hs\n\n", 
                        addrs[0],
                        msgFlashStartErase, 
                        ascdump(ByteVector((BYTE*)txt, (BYTE*)(txt+16))).c_str()); 
            }

            continue;
        }

        dwFlashInfoOffset= versions[i].dwBootloaderAddress;
        return true;

    }
    return false;
}

// this changes the protected area to 60000000 ( and thus opening up 80000000 )
bool DisableWriteProtect(DWORD dwFlashInfoOffset, DwordVector& orig)
{
    orig.resize(6);

    DWORD nRead;
    if (!ITReadProcessMemory(NULL, dwFlashInfoOffset, (BYTE*)vectorptr(orig), sizeof(DWORD)*orig.size(), &nRead)) {
        debug("ERROR: reading from %08lx\n", dwFlashInfoOffset);
        return false;
    }

    DwordVector data= orig;
    data[0]= 0x60000000;
    data[2]= 0x60040000;
    data[5]= 0x04000000;    // extend rom size to include extended rom.

    DWORD nWritten;
    return ITWriteProcessMemory(NULL, dwFlashInfoOffset, (BYTE*)vectorptr(data), sizeof(DWORD)*data.size(), &nWritten);
}
// reset it to normal
bool EnableWriteProtect(DWORD dwFlashInfoOffset, const DwordVector& orig)
{
    DWORD nWritten;
    return ITWriteProcessMemory(NULL, dwFlashInfoOffset, (BYTE*)vectorptr(orig), sizeof(DWORD)*orig.size(), &nWritten);
}
bool ITFlashROM(DWORD dwOffset, BYTE *buffer, DWORD dwBufLength)
{
    DWORD insize= sizeof(FlashROMParams)+dwBufLength;
    FlashROMParams *req= (FlashROMParams*)LocalAlloc(LPTR, insize);

    req->dwOffset= dwOffset;
    req->dwSize= dwBufLength;
    memcpy(req->buffer, buffer, dwBufLength);
    req->dwCheckSum= CalcCheckSum(req->buffer, req->dwSize);

    DWORD outsize=0;
    FlashROMResult *outbuf=NULL;

    HRESULT res= CeRapiInvoke(L"\\Windows\\ItsUtils.dll", L"ITFlashROM",
            insize, (BYTE*)req,
            &outsize, (BYTE**)&outbuf, NULL, 0);
    if (res)
    {
        error(res, "ITFlashROM(%08lx, %08lx)", dwOffset, dwBufLength);
        return false;
    }
    debug("result: %08lx %08lx\n", outbuf->dwResult, outbuf->dwReturned);
    LocalFree(outbuf);

    return true;
}


bool ReadFromFile(char *filename, DWORD dwOffset, DWORD dwSize, ByteVector& buffer)
{
    FILE *f= fopen(filename, "rb");
    if (f==NULL)
    {
        error("fopen(%hs)", filename);
        return false;
    }
    if (0!=fseek(f, dwOffset, SEEK_SET))
    {
        error("fseek(%hs, %08lx)", filename, dwOffset);
        fclose(f);
        return false;
    }
    buffer.resize(dwSize);
    if (1!=fread(vectorptr(buffer), dwSize, 1, f))
    {
        error("fread(%hs, %08lx)", filename, dwSize);
        fclose(f);
        return false;
    }

    fclose(f);
    return true;
}

DWORD CalcCheckSum(BYTE *buffer, DWORD dwSize)
{
    DWORD sum= 0;
    while (dwSize--)
        sum+=*buffer++;

    return sum;
}

bool ITReadProcessMemory(HANDLE hProc, DWORD dwOffset, BYTE *buffer, DWORD dwBytesWanted, DWORD *pdwNumberOfBytesRead)
{
    ReadProcessMemoryParams inbuf;
    DWORD outsize=0;
    ReadProcessMemoryResult *outbuf=NULL;

    inbuf.hProcess= hProc;
    inbuf.dwOffset= dwOffset;
    inbuf.nSize= dwBytesWanted;
    outbuf= NULL; outsize= 0;
    HRESULT res= CeRapiInvoke(L"\\Windows\\ItsUtils.dll", L"ITReadProcessMemory",
            sizeof(ReadProcessMemoryParams), (BYTE*)&inbuf,
            &outsize, (BYTE**)&outbuf, NULL, 0);
    if (res)
    {
        error(res, "ITReadProcessMemory");
        return false;
    }
    memcpy(buffer, &outbuf->buffer, outbuf->dwNumberOfBytesRead);
    *pdwNumberOfBytesRead= outbuf->dwNumberOfBytesRead;
    LocalFree(outbuf);

    return true;
}
bool ITWriteProcessMemory(HANDLE hProc, DWORD dwOffset, BYTE *buffer, DWORD dwBufLength, DWORD *pdwNumberOfBytesWritten)
{
    WriteProcessMemoryParams *inbuf;
    DWORD outsize=0;
    WriteProcessMemoryResult *outbuf=NULL;

    DWORD insize= sizeof(WriteProcessMemoryParams)+dwBufLength;
    inbuf= (WriteProcessMemoryParams *)LocalAlloc(LPTR, insize);

    inbuf->hProcess= hProc;
    inbuf->dwOffset= dwOffset;
    inbuf->nSize= dwBufLength;
    memcpy(inbuf->buffer, buffer, dwBufLength);

    outbuf= NULL; outsize= 0;
    HRESULT res= CeRapiInvoke(L"\\Windows\\ItsUtils.dll", L"ITWriteProcessMemory",
            insize, (BYTE*)inbuf,
            &outsize, (BYTE**)&outbuf, NULL, 0);
    if (res)
    {
        error(res, "ITWriteProcessMemory(%08lx, %08lx, %08lx)", hProc, dwOffset, dwBufLength);
        return false;
    }
    *pdwNumberOfBytesWritten= outbuf->dwNumberOfBytesWritten;
    LocalFree(outbuf);

    return true;
}

