/* (C) 2003 XDA Developers  itsme@xs4all.nl
 *
 * $Header: /var/lib/cvs/xdautils/src/ppostmsg.cpp,v 1.1 2004/02/09 17:23:28 itsme Exp $
 *
 *
 * 0x0002= WM_DESTROY (0,0)
 * 0x0010= WM_CLOSE (0,0)
 * 0x000A= WM_ENABLE (fEnable, 0)
 * 0x0012= WM_QUIT (nExitCode, 0)
 * 0x0111= WM_COMMAND (wNotify<<16 | wId, hWndCtl) 
 *     BN_CLICKED = 0
 *
 */
#include <windows.h>
#include "ItsUtils.h"
#include "dllversion.h"
#include "debug.h"
#include "stringutils.h"
#include "args.h"
#include <stdio.h>
#include <string.h>
#include <rapi.h>

HWND ITFindWindow(const std::string& classname, const std::string& windowname);
bool ITPostMessage(HWND hWnd, DWORD nMsg, DWORD wParam, DWORD lParam);

void usage()
{
    printf("Usage: ppostmsg [-h HANDLE] [-w WINDOWNAME] [-c CLASSNAME]  msg wparam lparam\n");
    printf("    numbers can be specified as 0x1234abcd\n");
    printf("    either specify a handle, or a name+class\n");
}

int main( int argc, char *argv[])
{
    DebugStdOut();

    HWND hWnd=0;
    std::string windowname;
    std::string classname;
    DWORD nMsg=0;
    DWORD wParam=0;
    DWORD lParam=0;

    int argsfound=0;
    for (int i=1 ; i<argc ; i++)
    {
        if (argv[i][0]=='-') switch(argv[i][1])
        {
            case 'h': HANDLEULOPTION(hWnd, HWND); break;
            case 'w': HANDLESTROPTION(windowname); break;
            case 'c': HANDLESTROPTION(classname); break;

            default:
                usage();
                return 1;
        }
        else switch (argsfound++)
        {
            case 0: nMsg= strtoul(argv[i], 0, 0); break;
            case 1: wParam= strtoul(argv[i], 0, 0); break;
            case 2: lParam= strtoul(argv[i], 0, 0); break;
        }
    }

    if (argsfound==0 || argsfound>3)
    {
        usage();
        return 1;
    }

    CheckITSDll();

    if (hWnd==0)
        hWnd= ITFindWindow(classname, windowname);

    if (hWnd==0) {
        debug("Window not found\n");
        return 1;
    }

    if (!ITPostMessage(hWnd, nMsg, wParam, lParam)) {
        debug("Error sending message\n");
    }
    CeRapiUninit();

    return 0;
}


HWND ITFindWindow(const std::string& classname, const std::string& windowname)
{
    FindWindowParams inbuf;
    DWORD insize= sizeof(inbuf);
    _snwprintf(inbuf.szClassName, MAX_PATH-1, L"%hs", classname.c_str());
    _snwprintf(inbuf.szWindowName, MAX_PATH-1, L"%hs", windowname.c_str());

    DWORD outsize=0;
    FindWindowResult *outbuf=NULL;

    HRESULT res= CeRapiInvoke(L"\\Windows\\ItsUtils.dll", L"ITFindWindow",
            insize, (BYTE*)&inbuf,
            &outsize, (BYTE**)&outbuf, NULL, 0);
    if (res) {
        error(res, "ITFindWindow");
        return 0;
    }

    HWND hWnd= outbuf->hWnd;

    LocalFree(outbuf);

    return hWnd;
}

bool ITPostMessage(HWND hWnd, DWORD nMsg, DWORD wParam, DWORD lParam)
{
    PostMessageParams inbuf;
    inbuf.hWnd= hWnd;
    inbuf.nMsg= nMsg;
    inbuf.wParam= wParam;
    inbuf.lParam= lParam;
    DWORD insize= sizeof(inbuf);

    DWORD outsize=0;
    BYTE *outbuf=NULL;

    HRESULT res= CeRapiInvoke(L"\\Windows\\ItsUtils.dll", L"ITPostMessage",
            insize, (BYTE*)&inbuf,
            &outsize, (BYTE**)&outbuf, NULL, 0);
    if (res) {
        error(res, "ITPostMessage");
        return false;
    }

    LocalFree(outbuf);

    return true;
}

