/* (C) 2003 XDA Developers  itsme@xs4all.nl
 *
 * $Header: /var/db/cvs/xdautils/src/pps.cpp,v 1.13 2005/01/14 15:54:47 itsme Exp $
 */
#include <windows.h>
#include <stdio.h>
#include <string.h>
#include <rapi.h>

#include <map>

#define MEASUREMENT_MSEC 200

#include "ItsUtils.h"
#include "dllversion.h"

#include "debug.h"
#include "args.h"

class ProcessTimes {
public:
    DWORD tKernel;
    DWORD tUser;
};
typedef std::map<HANDLE,ProcessTimes> ProcessTimeMap;

typedef std::map<HANDLE,CEPROCESSENTRY> ProcessInfoMap;
typedef std::map<HANDLE,ThreadSummaryInfo> ThreadSummaryInfoMap;

void PrintProcessList(const ProcessInfoMap &pinfo, const ProcessTimeMap &ptm0, const ProcessTimeMap &ptm1, DWORD measurementtime);
bool GetProcessInfo(bool bIncludeHeap, ProcessInfoMap &pinfo);
bool GetProcessTimes(ProcessTimeMap &ptimes, DWORD &tQuery);

bool ITGetThreadUsageList(DWORD& tQuery, ThreadSummaryInfoMap& tinfo);
void PrintThreadList(const ProcessInfoMap &pinfo, const ThreadSummaryInfoMap& tsim0, const ThreadSummaryInfoMap& tsim1, DWORD measurementtime);
void PrintThreadEntry(const CEPROCESSENTRY* pe, const ThreadSummaryInfo& ti,  const ThreadSummaryInfo* ti0, const ThreadSummaryInfo* ti1, DWORD measurementtime);

void PrintContext();

void PrintThreadInfo(HANDLE hThread, DWORD tMeasure);
bool ITGetThreadTimes(HANDLE hThread, FILETIME *tCreate, FILETIME *tExit, FILETIME *tKernel, FILETIME *tUser);
BOOL CalcFiletimeDiff(const FILETIME *tFirst, const FILETIME *tLast, FILETIME *tDiff);
DWORD CalcPromile(FILETIME *tA, FILETIME *tB);

void usage()
{
    printf("Usage: pps [options]\n");
    printf("   -t        : list detailed thread info\n");
    printf("   -p HANDLE : list info for process\n");
    printf("   -m msec   : measurement interval\n");
    printf("   -h        : include heap information (slow)\n");
    printf("   -v        : verbose\n");
}
int main( int argc, char *argv[])
{
    DebugStdOut();

    bool bVerbose= false;
    HANDLE hThread= INVALID_HANDLE_VALUE;
    HANDLE hProcess= INVALID_HANDLE_VALUE;
    DWORD tMeasure= MEASUREMENT_MSEC;
    bool bListAllThreads= false;
    bool bIncludeHeap= false;

    int argsfound=0;
    for (int i=1 ; i<argc ; i++)
    {
        if (argv[i][0]=='-') switch(argv[i][1])
        {
            case 't': if (!HANDLEULOPTION(hThread, HANDLE)) bListAllThreads= true; break;
            case 'p': HANDLEULOPTION(hProcess, HANDLE); break;
            case 'm': HANDLEULOPTION(tMeasure, DWORD); break;
            case 'h': bIncludeHeap= true; break;
            case 'v': bVerbose= true; break;
            default:
                usage();
                return 1;
        }
        else 
            argsfound++;
    }
    if (argsfound)
    {
        usage();
        return 1;
    }

    CheckITSDll();
    if (bVerbose)
        PrintContext();
    if (bListAllThreads) {
        ProcessInfoMap pinfo;
        ThreadSummaryInfoMap tsim0, tsim1;
        DWORD tQuery0, tQuery1;

        if (!GetProcessInfo(false, pinfo))
            return 1;
        if (!ITGetThreadUsageList(tQuery0, tsim0))
            return 1;
        Sleep(tMeasure);
        if (!ITGetThreadUsageList(tQuery1, tsim1))
            return 1;

        PrintThreadList(pinfo, tsim0, tsim1, tQuery1-tQuery0);
    }
    else if (hThread!=INVALID_HANDLE_VALUE)
        PrintThreadInfo(hThread, tMeasure);
    else {
        ProcessInfoMap pinfo;
        ProcessTimeMap tm0, tm1;
        DWORD tQuery0, tQuery1;

        if (!GetProcessInfo(bIncludeHeap, pinfo))
            return 1;
        if (!GetProcessTimes(tm0, tQuery0))
            return 1;
        Sleep(tMeasure);
        if (!GetProcessTimes(tm1, tQuery1))
            return 1;

        PrintProcessList(pinfo, tm0, tm1, tQuery1-tQuery0);
    }

    CeRapiUninit();

    return 0;
}

void PrintContext()
{
    DWORD outsize=0;
    GetContextResult *r=NULL;

    HRESULT res= CeRapiInvoke(L"\\Windows\\ItsUtils.dll", L"ITGetContext",
            0, NULL, &outsize, (BYTE**)&r, NULL, 0);

    printf("result=%08lx  buf=%08lx size=%08lx\n", res, r, outsize);
    if (res==0 && r!=NULL)
    {
        printf("pid=%08lx ph=%08lx tid=%08lx th=%08lx trust.caller=%d trust.cur=%d\n",
                r->dwProcessId, r->hProcess, r->dwThreadId, r->hThread,
                r->dwCurrentTrust, r->dwCallerTrust);
        printf("cmd: %ls\n", r->wszCmdLine);
        LocalFree(r);
    }
    printf("\n");

}
const ProcessTimes *GetTime(const ProcessTimeMap &tmap, HANDLE h)
{
    ProcessTimeMap::const_iterator i= tmap.find(h);
    if (i==tmap.end())
        return NULL;

    return &(*i).second;
}
void PrintProcessList(const ProcessInfoMap &pinfo, const ProcessTimeMap &tm0, const ProcessTimeMap &tm1, DWORD measurementtime)
{
    double pmUserTotal=0;
    double pmKernelTotal=0;
    DWORD dwMemTotal=0;
    DWORD dwTotalThreads=0;

    printf("%-8s %2s %-8s %5s %5s %7s %s\n", "handle", "n", "base", "kern", "user", "heap", "exe");
    for (ProcessInfoMap::const_iterator i=pinfo.begin() ; i!=pinfo.end() ; ++i)
    {
        HANDLE h= (*i).first;
        const CEPROCESSENTRY &pe= (*i).second;
        const ProcessTimes *pt0= GetTime(tm0,h);
        const ProcessTimes *pt1= GetTime(tm1,h);

        double pmUser  = -1;
        double pmKernel= -1;

        if (pt0 && pt1) {
            pmUser= ((pt1->tUser-pt0->tUser)*100.0)/measurementtime;
            pmKernel= ((pt1->tKernel-pt0->tKernel)*100.0)/measurementtime;

            pmUserTotal += pmUser;
            pmKernelTotal += pmKernel;
        }
        dwMemTotal += pe.dwDefaultHeapSize;
        printf("%08lx %2d %08lx %5.1f %5.1f %7d %ls%hs%ls\n", h, pe.cntThreads, pe.dwMemoryBase, pmKernel, pmUser, pe.dwDefaultHeapSize, pe.szExeFile, pe.szCmdLine[0]?" ":"", pe.szCmdLine);
    }

    printf("         %2d ........ %5.1f %5.1f %7d total\n", dwTotalThreads, pmKernelTotal, pmUserTotal, dwMemTotal);
}
bool GetProcessInfo(bool bIncludeHeap, ProcessInfoMap &pinfo)
{
    GetProcessListParams p;
    p.bIncludeHeapUsage= bIncludeHeap;

    DWORD outsize=0;
    GetProcessListResult *r=NULL;

    HRESULT res= CeRapiInvoke(L"\\Windows\\ItsUtils.dll", L"ITGetProcessList",
            sizeof(GetProcessListParams), (BYTE*)&p, &outsize, (BYTE**)&r, NULL, 0);

    if (res)
    {
        error(res, "ITGetProcessList");
        return false;
    }

    for (int i=0 ; i<r->nEntries ; i++)
    {
        memcpy(&pinfo[(HANDLE)r->pe[i].dwProcessID], &r->pe[i], sizeof(CEPROCESSENTRY));
    }

    LocalFree(r);
    
    return true;
}


bool GetProcessTimes(ProcessTimeMap &ptimes, DWORD& tQuery)
{
    DWORD outsize=0;
    GetProcessUsageListResult *r=NULL;

    HRESULT res= CeRapiInvoke(L"\\Windows\\ItsUtils.dll", L"ITGetProcessUsageList",
            0, NULL, &outsize, (BYTE**)&r, NULL, 0);

    if (res)
    {
        error(res, "ITGetProcessUsageList");
        return false;
    }

    for (int i=0 ; i<MAX_PROCESSES; i++)
    {
        if (r->list[i].hProc)
            ptimes[r->list[i].hProc].tKernel= r->list[i].tKernel;
            ptimes[r->list[i].hProc].tUser= r->list[i].tUser;
    }

    tQuery= r->tQuery;

    LocalFree(r);
    
    return true;
}

void PrintThreadInfo(HANDLE hThread, DWORD tMeasure)
{
    FILETIME tKernel0, tUser0, t0;
    FILETIME tKernel1, tUser1, t1;


    GetSystemTimeAsFileTime(&t0);
    if (!ITGetThreadTimes(hThread, NULL, NULL, &tKernel0, &tUser0))
        return;
    Sleep(tMeasure);
    GetSystemTimeAsFileTime(&t1);
    if (!ITGetThreadTimes(hThread, NULL, NULL, &tKernel1, &tUser1))
        return;

    FILETIME tKernel, tUser, t;

    CalcFiletimeDiff(&tKernel0, &tKernel1, &tKernel);
    CalcFiletimeDiff(&tUser0, &tUser1, &tUser);
    CalcFiletimeDiff(&t0, &t1, &t);

    printf("user = %4d  kernel = %4d\n", CalcPromile(&tUser, &t), CalcPromile(&tKernel, &t));
}

bool ITGetThreadTimes(HANDLE hThread, FILETIME *tCreate, FILETIME *tExit, FILETIME *tKernel, FILETIME *tUser)
{
    DWORD outsize=0;
    GetThreadTimesResult *r=NULL;
    GetThreadTimesParams p;

    p.hThread= hThread;

    HRESULT res= CeRapiInvoke(L"\\Windows\\ItsUtils.dll", L"ITGetThreadTimes",
            sizeof(p), (BYTE*)&p, &outsize, (BYTE**)&r, NULL, 0);

    if (res) {
        error(res, "ITGetThreadTimes");
        return false;
    }
    if (tCreate) *tCreate= r->tCreate;
    if (tExit) *tExit= r->tExit;
    if (tKernel) *tKernel= r->tKernel;
    if (tUser) *tUser= r->tUser;

    LocalFree(r);

    return true;
}


BOOL CalcFiletimeDiff(const FILETIME *tFirst, const FILETIME *tLast, FILETIME *tDiff)
{
    __int64 a, b, c;

    a= tFirst->dwHighDateTime; a<<=32; a |= tFirst->dwLowDateTime;
    b= tLast->dwHighDateTime; b<<=32; b |= tLast->dwLowDateTime;

    c= b-a;

    tDiff->dwHighDateTime= (DWORD) (c>>32);
    tDiff->dwLowDateTime= (DWORD) (c);

    return (BOOL)(c>>32);
}

DWORD CalcPromile(FILETIME *tA, FILETIME *tB)
{
    __int64 a, b;

    a= tA->dwHighDateTime; a<<=32; a |= tA->dwLowDateTime;
    b= tB->dwHighDateTime; b<<=32; b |= tB->dwLowDateTime;

    if (b==0)
        return (DWORD)-1;

    a *= 1000;
    a /= b;

    return (DWORD)a;
}

bool ITGetThreadUsageList(DWORD& tQuery, ThreadSummaryInfoMap& tinfo)
{
    DWORD outsize=0;
    GetThreadUsageListResult *r=NULL;

    HRESULT res= CeRapiInvoke(L"\\Windows\\ItsUtils.dll", L"ITGetThreadUsageList",
            0, NULL, &outsize, (BYTE**)&r, NULL, 0);

    if (res)
    {
        error(res, "ITGetThreadUsageList");
        return false;
    }

    tinfo.clear();
    for (DWORD i=0 ; i<r->nThreads ; i++)
        tinfo[r->list[i].hThread]= r->list[i];

    tQuery= r->tQuery;

    LocalFree(r);
    
    return true;
}

void PrintThreadEntry(const CEPROCESSENTRY* pe, const ThreadSummaryInfo& ti,  const ThreadSummaryInfo* ti0, const ThreadSummaryInfo* ti1, DWORD measurementtime)
{
    printf("%08lx %08lx %02x %02x", ti.hThread, ti.dwStartAddr, ti.baseprio, ti.curprio);

    if (ti1==NULL && ti0==NULL)
        printf("             ");
    if (ti1==NULL)
        printf(" terminated  ");
    else if (ti0==NULL)
        printf(" created     ");
    else {
        double pmUser= (ti1->tUser-ti0->tUser)*100.0/measurementtime;
        double pmKernel= (ti1->tKernel-ti0->tKernel)*100.0/measurementtime;

        printf(" %5.1f %5.1f", pmUser, pmKernel);
    }
    if (ti.szModname[0])
        printf("  %-12ls", ti.szModname);
    else
        printf("              ");

    printf("   %08lx", ti.hProc);
    if (pe)
        printf(": %ls%hs%ls\n", pe->szExeFile, pe->szCmdLine[0]?" ":"", pe->szCmdLine);
    else
        printf(": ... unknown process\n");
}
const CEPROCESSENTRY* getmapentry(const ProcessInfoMap& map, HANDLE key)
{
    ProcessInfoMap::const_iterator i= map.find(key);
    if (i==map.end())
        return NULL;
    return &(*i).second;
}

void PrintThreadList(const ProcessInfoMap &pinfo, const ThreadSummaryInfoMap& tsim0, const ThreadSummaryInfoMap& tsim1, DWORD measurementtime)
{
    for (ThreadSummaryInfoMap::const_iterator i=tsim0.begin() ; i!=tsim0.end() ; ++i)
    {
        ThreadSummaryInfoMap::const_iterator i1= tsim1.find((*i).first);
        const ThreadSummaryInfo& ti0= (*i).second;
        const CEPROCESSENTRY *pe= getmapentry(pinfo,ti0.hProc);
        if (i1==tsim1.end())
            PrintThreadEntry(pe, ti0, &ti0, NULL, measurementtime);
        else {
            const ThreadSummaryInfo& ti1= (*i1).second;
            PrintThreadEntry(pe, ti0, &ti0, &ti1, measurementtime);
        }
    }
    for (ThreadSummaryInfoMap::const_iterator i=tsim1.begin() ; i!=tsim1.end() ; ++i)
    {
        ThreadSummaryInfoMap::const_iterator i0= tsim0.find((*i).first);
        const ThreadSummaryInfo& ti1= (*i).second;
        const CEPROCESSENTRY *pe= getmapentry(pinfo,ti1.hProc);
        if (i0==tsim0.end())
            PrintThreadEntry(pe, ti1, NULL, &ti1, measurementtime);
    }
}

