/* (C) 2003 XDA Developers  itsme@xs4all.nl
 *
 * $Header: /var/lib/cvs/xdautils/src/pregutl.cpp,v 1.17 2004/10/22 13:59:12 itsme Exp $
 */
#include <windows.h>
#include "debug.h"
#include "args.h"
#include <stdio.h>
#include <rapi.h>
#include <string>
#include <map>
#include <vector>
#include "stringutils.h"
#include "vectorutils.h"
#include <boost/regex.hpp>

#define STD_NAME_BUF_SIZE   256
#define STD_VALUE_BUF_SIZE  1024

#ifdef WIN32REGUTL
#define __RegCloseKey        RegCloseKey
#define __RegCreateKeyEx     RegCreateKeyEx
#define __RegDeleteKey       RegDeleteKey
#define __RegDeleteValue     RegDeleteValue
#define __RegEnumKeyEx       RegEnumKeyEx
#define __RegEnumValue       RegEnumValue
#define __RegOpenKeyEx       RegOpenKeyEx
#define __RegQueryInfoKey    RegQueryInfoKey
#define __RegQueryValueEx    RegQueryValueEx
#define __RegSetValueEx      RegSetValueEx
#endif

#ifdef WINCEREGUTL
#define __RegCloseKey        CeRegCloseKey
#define __RegCreateKeyEx     CeRegCreateKeyEx
#define __RegDeleteKey       CeRegDeleteKey
#define __RegDeleteValue     CeRegDeleteValue
#define __RegEnumKeyEx       CeRegEnumKeyEx
#define __RegEnumValue       CeRegEnumValue
#define __RegOpenKeyEx       CeRegOpenKeyEx
#define __RegQueryInfoKey    CeRegQueryInfoKey
#define __RegQueryValueEx    CeRegQueryValueEx
#define __RegSetValueEx      CeRegSetValueEx
#endif

// todo:
//    - add option to make output dump format compatible with .reg file
//        ( and possibly less readable 'hex' output )
//    - support 'REGEDIT4' fileformat header.
//    - support unicode files
//    - change 'r' option in non-recursive option
//    - add option to choose how to mix keys and values.
//         ( full cross product, or values per key. )

std::string GetNameFromSetSpec(const std::string& spec, size_t start);
std::string GetValueSpecFromSetSpec(const std::string& spec, size_t start);

enum RegAction {
    DO_LOAD,
    DO_LIST,
    DO_DELETE,
    DO_SET,
    DO_CREATE,
};

class RegistryKey;
/*
class RegistryData {
};
class DwordData : public RegistryData {
public:
    DwordData(ByteVector& data)
    {
    }
};
class BinaryData : public RegistryData {
public:
    BinaryData(ByteVector& data)
    {
    }
};
class StringData : public RegistryData {
public:
    StringData(ByteVector& data)
    {
    }
};
class StringListData : public RegistryData {
public:
    StringListData(ByteVector& data)
    {
    }
};
*/
std::string unescape(const std::string& str);
class RegistryValue {
private:
    const static boost::regex reDwordPattern;
    const static boost::regex reStringPattern_s;
    const static boost::regex reStringPattern_d;
    const static boost::regex reMultiPartStringPattern_s;
    const static boost::regex reMultiPartStringPattern_d;
    const static boost::regex reMultiStringPattern;
    const static boost::regex reBinaryPattern;
public:

    // this function takes a hexadecimal string value,  like '81f01234'
    // and converts it to a REG_DWORD : { 0x34, 0x12, 0xf0, 0x81 } value
    static RegistryValue FromDwordValue(const std::string& hexval) {
        return RegistryValue(strtoul(stringptr(hexval), 0, 16));
    }
    // this function takes an escaped string value like ab\'cd\"ef\"\n
    // and converts it to a REG_SZ : { "ab'cd\"ef\"\n" }
    static RegistryValue FromStringValue(const std::string& escstr) {
        return RegistryValue(unescape(escstr));
    }
    // this function takes an escaped string value like ab\'cd\"ef\"\n
    // and converts it to a REG_SZ : { "ab'cd\"ef\"\n" }
    static RegistryValue FromMultiStringValue(const std::string& multistr) {
        StringList list;

        bool bDoubleQuotes= multistr[0]=='\"';

        std::string::const_iterator i= multistr.begin();
        std::string::const_iterator iend= multistr.end();
        boost::smatch what;
        while (regex_search(i, iend, what, bDoubleQuotes?reMultiPartStringPattern_d:reMultiPartStringPattern_s)) {
            std::string escstr(what[1].first, what[1].second);
            list.push_back(unescape(escstr));

            i= what[0].second;
        }

        return RegistryValue(list);
    }
    // this function takes a string of format 00,11,22,33,44
    // and converts it to { 0x00, 0x11, 0x22, 0x33, 0x44 }
    static RegistryValue FromBinaryValue(const std::string& typestr, const std::string& hexdata) {
        int type= strtol(typestr.c_str(), 0, 0);
        ByteVector bv;
        size_t i=0;
        while (i<hexdata.size()) {
            bv.push_back((BYTE)strtoul(&hexdata[i], 0, 16));

            i= hexdata.find(",", i);
            if (i==hexdata.npos)
                break;

            i++;
        }
        return RegistryValue(type, bv);
    }
    static RegistryValue RegistryValue::FromValueSpec(const std::string& spec)
    {
        boost::smatch what;
        if (boost::regex_match(spec, what, reDwordPattern, boost::regbase::icase)) {
            std::string hexval(what[1].first, what[1].second);
            return FromDwordValue(hexval);
        }
        else if (boost::regex_match(spec, what, reStringPattern_s, boost::regbase::icase)) {
            std::string escstr(what[1].first, what[1].second);
            return FromStringValue(escstr);
        }
        else if (boost::regex_match(spec, what, reStringPattern_d, boost::regbase::icase)) {
            std::string escstr(what[1].first, what[1].second);
            return FromStringValue(escstr);
        }
        else if (boost::regex_match(spec, what, reMultiStringPattern, boost::regbase::icase)) {
            std::string multistr(what[1].first, what[1].second);
            return FromMultiStringValue(multistr);
        }
        else if (boost::regex_match(spec, what, reBinaryPattern, boost::regbase::icase)) {
            std::string typestr= (what[1].matched)
                ? std::string(what[1].first, what[1].second)
                : "3";
            std::string hexdata(what[2].first, what[2].second);
            return FromBinaryValue(typestr, hexdata);
        }
        debug("...%s.\n", spec.c_str());
		throw "unimplemented valuespec";
    }
    RegistryValue()
        : m_dwType(REG_NONE)
    {
        //printf("new regvalue none\n");
    }
    RegistryValue(DWORD dwType, const ByteVector& data)
        : m_dwType(dwType), m_data(data)
    {
        //printf("new regvalue type %d\n", m_dwType);
    }
    RegistryValue(const RegistryValue& rv)
        : m_dwType(rv.GetType()), m_data(rv.GetData())
    {
        //printf("copy of regvalue type %d\n", m_dwType);
    }

    DWORD GetType() const
    {
        return m_dwType;
    }

    // binary value
    RegistryValue(const ByteVector& data)
        : m_dwType(REG_BINARY), m_data(data)
    {
    }
    ByteVector GetData() const
    {
        return m_data;
    }

    // dword value
    RegistryValue(DWORD dwValue)
        : m_dwType(REG_DWORD), m_data(BV_FromDword(dwValue))
    {
    }
    DWORD GetDword() const
    {
        if (m_dwType!=REG_DWORD) throw "GetDword: not a dword";
        if (m_data.size()!=4) throw "GetDword: incorrect size";

        return BV_GetDword(m_data);
    }

    // string value
    RegistryValue(const std::string& strValue)
        : m_dwType(REG_SZ), m_data(BV_FromWString(ToWString(strValue)))
    {
        BV_AppendWord(m_data, 0); // add terminating (wchar_t)NUL
    }
    std::string GetString() const
    {
        if (m_dwType!=REG_SZ && m_dwType!=REG_EXPAND_SZ) throw "GetString: not a string";
        if (m_data.size()&1) throw "GetString: incorrect size";

        std::string str;
        for (ByteVector::const_iterator i=m_data.begin() ; i!=m_data.end() && (*i)!=0 ; i+=2)
            str += (char)(*i);

        return str;
    }

    // string list value
    RegistryValue(const StringList& listValue)
        : m_dwType(REG_MULTI_SZ)
    {
        m_data.clear();
        ByteVector::iterator out= m_data.begin();
        for (StringList::const_iterator i= listValue.begin() ; i!=listValue.end() ; ++i)
        {
            BV_AppendWString(m_data, ToWString(*i));
            BV_AppendWord(m_data, 0); // add terminating (wchar_t)NUL
        }
        BV_AppendWord(m_data, 0); // add terminating (wchar_t)NUL
    }
    StringList GetStringList() const
    {
        if (m_dwType!=REG_MULTI_SZ) throw "GetStringList: not a string";
        if (m_data.size()&1) throw "GetStringList: incorrect size";

        StringList list;
        std::string str;
        for (ByteVector::const_iterator i= m_data.begin() ; i!=m_data.end() ; i+=2)
        {
            if (*i==0)
            {
                list.push_back(str);
                str.clear();
            }
            else
            {
                str += (char)(*i);
            }
        }
        if (!str.empty()) {
            throw "Invalid multi_sz: no terminating NUL";
        }
        if (!list.back().empty()) {
            throw "Invalid multi_sz: no closing element";
        }
        list.pop_back();
        return list;
    }
    std::string quotestring(const std::string& str) const
    {
        std::string qstr= "\"";
        for (size_t i=0 ; i<str.size() ; i++)
        {
            if (str[i]<' ' || str[i]>'~')
                qstr+=stringformat("\\%02x", str[i]);
            else
                qstr+=str[i];
        }
        return qstr+"\"";
    }
    std::string MultiszAsString() const
    {
        StringList list= GetStringList();

        for (size_t i=0 ; i<list.size() ; ++i)
            list[i]= quotestring(list[i]);
        return std::string("multi_sz:")+JoinStringList(list, ",");
    }
    std::string AsString() const
    {
        switch(m_dwType) {
case REG_NONE:   return "none";
case REG_SZ: return quotestring(GetString());
case REG_EXPAND_SZ: return std::string("expand:")+quotestring(GetString());
case REG_BINARY: return std::string("hex:")+ascdump(m_data);
case REG_DWORD: return stringformat("dword:%08lx", *(DWORD*)vectorptr(m_data));
//case REG_DWORD_BE: return stringformat("dwordbe:%08lx", *(DWORD*)vectorptr(m_data));
case REG_LINK: return "link";
case REG_MULTI_SZ: return MultiszAsString();
case REG_RESOURCE_LIST: return "rsclist";
case REG_FULL_RESOURCE_DESCRIPTOR: return "rscdesc";
case REG_RESOURCE_REQUIREMENTS_LIST: return "rscreqslist";
case REG_QWORD: return "qword";
default: return "UNKNOWN";

        }
    }

private:
    DWORD m_dwType;
    ByteVector m_data;
    //Value* m_value;
};
const boost::regex RegistryValue::reDwordPattern= boost::regex("dword:\\s*(\\w+)");
const boost::regex RegistryValue::reStringPattern_s=       boost::regex("^\\s*'((?:[^'\\\\]|\\\\['\\\\0rtn])*)'\\s*$");
const boost::regex RegistryValue::reStringPattern_d=     boost::regex("^\\s*\"((?:[^\"\\\\]|\\\\[\"\\\\0rtn])*)\"\\s*$");
const boost::regex RegistryValue::reMultiPartStringPattern_s=   boost::regex("'((?:[^'\\\\]|\\\\['\\\\0rtn])*)'");
const boost::regex RegistryValue::reMultiPartStringPattern_d= boost::regex("\"((?:[^\"\\\\]|\\\\[\"\\\\0rtn])*)\"");
const boost::regex RegistryValue::reMultiStringPattern= boost::regex("^\\s*multi_sz:\\s*(.*?)\\s*$");
const boost::regex RegistryValue::reBinaryPattern= boost::regex("^\\s*hex(?:\\((\\w+)\\))?:\\s*(.*?)\\s*$");

typedef std::map<std::string,RegistryValue> RegistryValueMap;
typedef std::vector<RegistryKey> RegistryKeyList;


class RegistryKey {
public:
    static RegistryKey FromKeySpec(const std::string& keyspec)
    {
        HKEY hRoot;
        std::string path;

        size_t slashpos= keyspec.find_first_of("/\\");

        if (stricmp(keyspec.substr(0,slashpos).c_str(), "hkcr")==0)
            hRoot= HKEY_CLASSES_ROOT;
        else if (stricmp(keyspec.substr(0,slashpos).c_str(), "hkey_classes_root")==0)
            hRoot= HKEY_CLASSES_ROOT;
        else if (stricmp(keyspec.substr(0,slashpos).c_str(), "hkcu")==0)
            hRoot= HKEY_CURRENT_USER;
        else if (stricmp(keyspec.substr(0,slashpos).c_str(), "hkey_current_user")==0)
            hRoot= HKEY_CURRENT_USER;
        else if (stricmp(keyspec.substr(0,slashpos).c_str(), "hklm")==0)
            hRoot= HKEY_LOCAL_MACHINE;
        else if (stricmp(keyspec.substr(0,slashpos).c_str(), "hkey_local_machine")==0)
            hRoot= HKEY_LOCAL_MACHINE;
        else
            hRoot= HKEY_LOCAL_MACHINE;

        if (slashpos==keyspec.npos)
            path= "";
        else
            path= keyspec.substr(slashpos+1);

        //printf("FromKeySpec: %08lx - %s\n", hRoot, path.c_str());

        return RegistryKey(hRoot, path);
    }

    RegistryKey() : m_hRoot(0), m_path(), m_hKey(0)
    {
    }

    RegistryKey(HKEY hRoot) : m_hRoot(hRoot), m_path(), m_hKey(0)
    {
    }

    RegistryKey(const RegistryKey& hKey) : m_hRoot(hKey.GetRoot()), m_path(hKey.GetPath()), m_hKey(0)
    {
    }
    RegistryKey& operator=(const RegistryKey& rk)
    {
        m_hRoot=rk.GetRoot();
        m_path= rk.GetPath();
        m_hKey= 0;
        return *this;
    }
    RegistryKey(const RegistryKey& hRoot, const std::string& path) : m_hRoot(hRoot.GetRoot()), m_hKey(0)
    {
        m_path= hRoot.GetPath();
        if (!m_path.empty())
            m_path += "\\";
        m_path += path;
    }
    ~RegistryKey()
    {
        Close();
    }

    void Close()
    {
        __RegCloseKey(m_hKey);
        m_hKey= 0;
    }
    bool Open()
    {
        if (m_hKey)
            return true;

        LONG res= __RegOpenKeyEx(m_hRoot, ToWString(m_path).c_str(), 0, KEY_ALL_ACCESS, &m_hKey);
        if (res!=ERROR_SUCCESS)
        {
            error(res, "Failed to open key %hs:'%hs'", GetRootName().c_str(), m_path.c_str());
            m_hKey= 0;
            return false;
        }
        m_dwMaxSubKeyNameLength= STD_NAME_BUF_SIZE;
        m_dwMaxValueNameLength= STD_NAME_BUF_SIZE;
        m_dwMaxValueLength= STD_VALUE_BUF_SIZE;
/*
 *
 * this does not seem to work on ce/rapi
 *
        __RegQueryInfoKey(m_hKey, NULL, NULL, NULL, NULL, &m_dwMaxSubKeyNameLength,
                NULL, NULL, &m_dwMaxValueNameLength, &m_dwMaxValueLength, NULL, NULL);

        if ( m_dwMaxSubKeyNameLength==0 && m_dwMaxValueNameLength==0 && m_dwMaxValueLength==0)
        {
            debug("ERROR: noinfo\n");
            m_dwMaxSubKeyNameLength= STD_NAME_BUF_SIZE;
            m_dwMaxValueNameLength= STD_NAME_BUF_SIZE;
            m_dwMaxValueLength= STD_VALUE_BUF_SIZE;
        }
*/
        return true;
    }
    HKEY GetHKey()
    {
        Open();

        return m_hKey;
    }

    void info()
    {
        std::wstring regclass; regclass.resize(256);
        DWORD cbClass= regclass.size();

        DWORD nSubKeys=0;
        DWORD maxSubKeyNameLength=0;
        DWORD maxClassLength=0;
        DWORD nValues=0;
        DWORD maxValueNameLength=0;
        DWORD maxValueLength=0;

        Open();

        // BUG: only the key and value counts are returned
        LONG res= __RegQueryInfoKey(m_hKey, stringptr(regclass), &cbClass, NULL, &nSubKeys,
                &maxSubKeyNameLength, &maxClassLength, &nValues,
                &maxValueNameLength, &maxValueLength, NULL, NULL);
        if (res != ERROR_SUCCESS)
        {
            error(res, "__RegQueryInfoKey(%hs:'%hs')", GetRootName().c_str(), m_path.c_str());
            return ;
        }
        regclass.resize(cbClass);

        printf("regclass='%hs'  nkeys=%d nvals=%d  maxkey=%d maxcls=%d maxname=%d maxval=%d\n",
                regclass.c_str(), nSubKeys, maxSubKeyNameLength,
                maxClassLength, nValues, maxValueNameLength,&maxValueLength);
    }

    // returns the roothive of this key
    HKEY GetRoot() const
    {
        return m_hRoot;
    }

    // returns the full path to this key
    std::string GetPath() const
    {
        return m_path;
    }

    // returns the name of this key
    std::string GetName() const
    {
        size_t pos= m_path.find_last_of("\\/");
        if (pos!=m_path.npos)
            return m_path.substr(pos+1);
        else
            return m_path;
    }

    // returns the name of the hive of this key
    std::string GetRootName() const
    {
        if (m_hRoot==HKEY_LOCAL_MACHINE) return "HKLM";
        if (m_hRoot==HKEY_CURRENT_USER) return "HKCU";
        if (m_hRoot==HKEY_CLASSES_ROOT) return "HKCR";
        return "unknown";
    }

    // creates a new subkey
    bool CreateSubKey(const std::string& keyname)
    {
        if (!Open())
            return false;

        HKEY hk;
        DWORD disp;
        LONG res= __RegCreateKeyEx(m_hKey, ToWString(keyname).c_str(), 0, NULL, 0, 0, NULL, &hk, &disp);
        if (res != ERROR_SUCCESS)
        {
            error(res, "Failed to create subkey %hs:'%hs'\\'%hs'", GetRootName().c_str(), m_path.c_str(), keyname.c_str());
            return false;
        }
        __RegCloseKey(hk);
        return true;
    }


    // deletes a childless subkey from this key
    bool DeleteSubKey(const std::string& keyname)
    {
        if (!Open())
            return false;
        LONG res= __RegDeleteKey(m_hKey, ToWString(keyname).c_str());
        if (res != ERROR_SUCCESS)
        {
            error(res, "Failed to delete1 subkey %hs:'%hs'\\'%hs'", GetRootName().c_str(), m_path.c_str(), keyname.c_str());
            return false;
        }
        return true;
    }

    // delete all children
    bool DeleteChildren()
    {
        bool bRes= true;
        StringList children= GetSubKeyNames();
        for (StringList::iterator i= children.begin() ; i!=children.end() ; ++i)
        {
            bRes = bRes && DeleteSubKey(*i);
        }
        return bRes;
    }

    bool DeleteValue(const std::string& valuename)
    {
        if (!Open())
            return false;
        LONG res= __RegDeleteValue(m_hKey, ToWString(valuename).c_str());
        if (res != ERROR_SUCCESS)
        {
            error(res, "Failed to delete value %hs:'%hs'\\'%hs'", GetRootName().c_str(), m_path.c_str(), valuename.c_str());
            return false;
        }
        return true;
    }

    //-------------------------------------------------------------------
    // deletes this key and all it's children
    bool DeleteKey()
    {
        if (!DeleteChildren())
            return false;

        Close();

        if (!GetParent().DeleteSubKey(GetName()))
            return false;

        return true;
    }

    //-------------------------------------------------------------------
    // creates registry key
    bool CreateKey()
    {
        if (Open()) {
            Close();
            return true;
        }

        if (!GetParent().CreateKey())
            return false;

        if (!GetParent().CreateSubKey(GetName()))
            return false;

        debug("created key %hs\\%hs\n", GetRootName().c_str(), GetPath().c_str());
        return true;
    }


    void DumpKey()
    {
        printf("[%hs\\%hs]\n", GetRootName().c_str(), GetPath().c_str());
        DumpValues();
        printf("\n");
        DumpChildren();
    }

    //-------------------------------------------------------------------
    void DumpValues()
    {
        RegistryValueMap map= GetValues();

        for (RegistryValueMap::iterator i= map.begin() ; i!=map.end() ; ++i)
        {
            printf("%hs=%hs\n", (*i).first.c_str(), (*i).second.AsString().c_str());
        }
    }

    void DumpChildren()
    {
        RegistryKeyList list= GetSubKeys();

        for (RegistryKeyList::iterator i= list.begin() ; i!=list.end() ; ++i)
        {
            (*i).DumpKey();
        }
    }

    //-------------------------------------------------------------------
    // returns parent to this key
    RegistryKey GetParent()
    {
        size_t pos= m_path.find_last_of("\\/");
        if (pos!=m_path.npos)
            return RegistryKey(m_hRoot, m_path.substr(0, pos));
        else
            return RegistryKey(m_hRoot);
    }
    StringList GetSubKeyNames()
    {
        StringList list;
        if (!Open())
            throw "GetSubKeyNames: could not open";

        for (int i=0 ; true ; i++)
        {
            DWORD cbName = m_dwMaxSubKeyNameLength+1;
            std::wstring wname; wname.resize(cbName);
            LONG res =__RegEnumKeyEx(m_hKey, i, stringptr(wname), &cbName, NULL, NULL, NULL, NULL);
            if (res==ERROR_NO_MORE_ITEMS)
                break;
            if (res!=ERROR_SUCCESS)
            {
                error(res, "__RegEnumKeyEx(%hs:'%hs', %d)", GetRootName().c_str(), GetPath().c_str(), i);
                continue;
            }
            wname.resize(cbName);
            list.push_back(ToString(wname));
        }
        return list;
    }
    RegistryKeyList GetSubKeys()
    {
        RegistryKeyList list;
        if (!Open())
            throw "GetSubkeys: could not open";

        for (int i=0 ; true ; i++)
        {
            DWORD cbName = m_dwMaxSubKeyNameLength+1;
            std::wstring wname; wname.resize(cbName);
            LONG res =__RegEnumKeyEx(m_hKey, i, stringptr(wname), &cbName, NULL, NULL, NULL, NULL);
            if (res==ERROR_NO_MORE_ITEMS)
                break;
            if (res!=ERROR_SUCCESS)
            {
                error(res, "__RegEnumKeyEx(%hs:'%hs', %d)", GetRootName().c_str(), GetPath().c_str(), i);
                continue;
            }
            wname.resize(cbName);
            list.push_back(RegistryKey(*this, ToString(wname)));
        }
        return list;
    }
    StringList GetValueNames()
    {
        StringList list;
        if (!Open())
            throw "GetValueNames: could not open";

        for (int i=0 ; true ; i++)
        {
            DWORD cbName = m_dwMaxValueNameLength+1;
            std::wstring wname; wname.resize(cbName);
            LONG res= __RegEnumValue(m_hKey, i, stringptr(wname), &cbName, NULL, NULL, NULL, NULL);
            if (res==ERROR_NO_MORE_ITEMS)
                break;
            if (res!=ERROR_SUCCESS)
            {
                error(res, "__RegEnumValue(%hs:'%hs', %d)", GetRootName().c_str(), GetPath().c_str(), i);
                continue;
            }
            wname.resize(cbName);
            list.push_back(ToString(wname));
        }
        return list;
    }
    RegistryValueMap GetValues()
    {
        RegistryValueMap map;
        if (!Open())
            throw "GetValues: could not open";

        for (int i=0 ; true ; i++)
        {
            DWORD cbName = m_dwMaxValueNameLength+1;
            std::wstring wname; wname.resize(cbName);
            DWORD cbData = m_dwMaxValueLength;
            ByteVector data; data.resize(cbData);
            DWORD dwType;
            LONG res= __RegEnumValue(m_hKey, i, stringptr(wname), &cbName, NULL, &dwType, vectorptr(data), &cbData);
            if (res==ERROR_NO_MORE_ITEMS)
                break;
            if (res!=ERROR_SUCCESS)
            {
                error(res, "__RegEnumValue(%hs:'%hs', %d)", GetRootName().c_str(), GetPath().c_str(), i);
                continue;
            }
            wname.resize(cbName);
            data.resize(cbData);

            map[ToString(wname)]= RegistryValue(dwType, data);
        }
        return map;
    }
    RegistryValue GetValue(const std::string& valuename)
    {
        if (!Open())
            throw "GetValue: could not open";
        DWORD cbData = m_dwMaxValueLength;
        ByteVector data; data.resize(cbData);
        DWORD dwType;
        LONG res= __RegQueryValueEx(m_hKey, ToWString(valuename).c_str(), NULL, &dwType, vectorptr(data), &cbData);
        if (res!=ERROR_SUCCESS)
        {
            error(res, "__RegQueryValueEx(%hs:'%hs' '%hs')", GetRootName().c_str(), GetPath().c_str(), valuename.c_str());

            throw "GetValue: value does not exist";
        }
        data.resize(cbData);
        return RegistryValue(dwType, data);
    }
    bool SetValue(const std::string& valuename, const RegistryValue& value)
    {
        if (!Open())
            throw "SetValue: could not open";

        //printf("setting [%s] %s to %s\n", m_path.c_str(), valuename.c_str(), value.AsString().c_str());
        ByteVector data= value.GetData();
        LONG res= __RegSetValueEx(m_hKey, ToWString(valuename).c_str(), NULL, value.GetType(), vectorptr(data), data.size());
        if (res!=ERROR_SUCCESS)
        {
            error(res, "__RegSetValueEx(%hs:'%hs' '%hs')", GetRootName().c_str(), GetPath().c_str(), valuename.c_str());
            throw "SetValue: could not create";
        }
        debug("set value %hs\\%hs  %hs : %hs\n", GetRootName().c_str(), GetPath().c_str(), valuename.c_str(), value.AsString().c_str());
        return true;
    }

private:
        HKEY m_hRoot;
        std::string m_path;

        HKEY m_hKey;
        DWORD m_dwMaxSubKeyNameLength;
        DWORD m_dwMaxValueNameLength;
        DWORD m_dwMaxValueLength;
};

size_t findendquote(const std::string& str, size_t pos, char quotechar)
{
    bool bEscaped= false;
    while (pos<str.size())
    {
        if (bEscaped)
        {
            bEscaped= false;
        }
        else if (str[pos]==quotechar)
        {
            return pos+1;
        }
        else if (str[pos]=='\\')
        {
            bEscaped= true;
        }
        ++pos;
    }
    return std::string::npos;
}
std::string unescape(const std::string& str)
{
    std::string result;
    size_t pos= 0;
    bool bEscaped= false;
    while (pos<str.size())
    {
        if (bEscaped)
        {
            switch(str[pos])
            {
                case 'n': result+='\n'; break;
                case 'r': result+='\r'; break;
                case 't': result+='\t'; break;
                case 'x': if (pos+2<str.size())
                          {
                              result += (char)strtoul(str.substr(pos+1,2).c_str(),0,16); 
                              pos+=2;
                          }
                          break;
                case '0':
                case '1':
                case '2': if (pos+2<str.size())
                          {
                              result += (char)strtoul(str.substr(pos,3).c_str(),0,8); 
                              pos+=2;
                          }
                          break;
                default:
                          result+=str[pos];
            }
            bEscaped= false;
        }
        else if (str[pos]=='\\')
        {
            bEscaped= true;
        }
        else {
            result += str[pos];
        }
        ++pos;
    }

    return result;
}
bool ReadLine(FILE *f, std::string& line)
{
    line.clear();
    line.resize(1024);

    char *p= fgets(stringptr(line), 1024, f);
    if (p==NULL) {
        line.clear();
        return false;
    }
    line.resize(strlen(p));
    // todo: handle long lines ( > 1024 )
    //  - read until line has EOLN

    while (line.size() && (line[line.size()-1]=='\r' || line[line.size()-1]=='\n')) {
        line.resize(line.size()-1);
    }
    return true;
}

bool ProcessRegFile(const std::string& filename)
{
    FILE *f= fopen(filename.c_str(), "r");
    if (f==NULL) {
        perror(filename.c_str());
        return false;
    }

    RegistryKey curkey;
    std::string line;

    while (ReadLine(f, line)) {
        while (line.size() && isspace(line[line.size()-1])) {
            line.resize(line.size()-1);
        }

        if (line.size()==0)
            continue;
        else if (line[0]=='[' && line[line.size()-1]==']') {
            curkey= RegistryKey::FromKeySpec(line.substr(1, line.size()-2));
            curkey.CreateKey();
        }
        else if (line[0]==';') {
            // skip comments.
        }
        else {
            std::string valuename= GetNameFromSetSpec(line, 0);
            std::string valuespec= GetValueSpecFromSetSpec(line, 0);

            //printf("valline: %hs  = %hs\n", valuename.c_str(), valuespec.c_str());

            curkey.SetValue(valuename, RegistryValue::FromValueSpec(valuespec));
        }
    }

    fclose(f);
    return true;
}

bool IsValueNameSpec(const std::string& spec)
{
    if (spec[0]!=':')
        return false;
    if (spec.find('=')==std::string::npos)
        return true;

    if (spec[1]=='\'' || spec[1]=='\"')
    {
        size_t endquote= findendquote(spec, 2, spec[1]);

        if (endquote==std::string::npos)
        {
            throw "IsValueNameSpec: missing endquote";
        }
        return endquote==spec.size();
    }
    return false;
}

std::string GetNameFromValueNameSpec(const std::string& spec, size_t start)
{
    if (spec[start]=='\'' || spec[start]=='\"')
    {
        size_t endquote= findendquote(spec, start+1, spec[start]);
        //printf("GetNameFromValueNameSpec:%s\n", spec.substr(1, endquote-3).c_str());
        return unescape(spec.substr(start, endquote-start-1));
    }
    //printf("GetNameFromValueNameSpec:%s\n", spec.substr(1).c_str());
    return spec.substr(start);
}

bool IsSetSpec(const std::string& spec)
{
    if (spec[0]!=':')
        return false;
    if (spec.find('=')==std::string::npos)
        return false;

    if (spec[1]=='\'' || spec[1]=='\"')
    {
        size_t endquote= findendquote(spec, 2, spec[1]);

        if (endquote==std::string::npos)
        {
            throw "IsSetSpec: missing endquote\n";
        }
        //printf("isset: endq=%d size=%d c=%c\n", endquote, spec.size(), spec[endquote]);
        if (endquote==spec.size())
            return false;

        return (spec[endquote]=='=');
    }
    return true;
}
std::string GetNameFromSetSpec(const std::string& spec, size_t start)
{
    if (spec[start]=='\'' || spec[start]=='\"')
    {
        size_t endquote= findendquote(spec, start+1, spec[start]);
        //printf("GetNameFromSetSpec:%s\n", spec.substr(1, endquote-3).c_str());

        //todo: check for missing quote.
        return unescape(spec.substr(start+1, endquote-start-2));
    }
    size_t eqpos= spec.find('=');

    //printf("GetNameFromSetSpec:%s\n", spec.substr(1,eqpos-1).c_str());
    return spec.substr(start,eqpos-start);
}

std::string GetValueSpecFromSetSpec(const std::string& spec, size_t start)
{
    if (spec[start]=='\'' || spec[start]=='\"')
    {
        size_t endquote= findendquote(spec, start+1, spec[start]);
        //printf("GetValueSpecFromSetSpec:%s\n", spec.substr(endquote).c_str());

        //todo: check for missing quote.
        size_t eqpos= spec.find('=', endquote);
        return spec.substr(eqpos+1);
    }
    size_t eqpos= spec.find('=');

    //printf("GetValueSpecFromSetSpec:%s\n", spec.substr(eqpos+1).c_str());
    return spec.substr(eqpos+1);
}

bool IsRegFile(const std::string& filespec)
{
    return filespec[0]=='@';
}
std::string GetFilenameFromRegfileSpec(const std::string& filespec)
{
    return filespec.substr(1);
}

void checkparameters(RegAction action, const RegistryKeyList& keys, const StringList& names, const RegistryValueMap& values)
{
    if (action==DO_CREATE && names.size())
        throw "checkparameters: create with name";
    if (action==DO_CREATE && values.size())
        throw "checkparameters: create with 'name=value'";
    if (action==DO_CREATE && keys.size()==0)
        throw "checkparameters: create without keys";
    if (action==DO_DELETE && values.size())
        throw "checkparameters: delete with 'name=value'";
    if (action==DO_DELETE && keys.size()==0)
        throw "checkparameters: delete without keys";
    if (action==DO_LIST && values.size())
        throw "checkparameters: list with 'name=value'";
    if (action==DO_LIST && keys.size()==0)
        throw "checkparameters: list without keys";
    if (action==DO_SET && names.size())
        throw "checkparameters: set value with bare name";
    if (action==DO_SET && values.size()==0)
        throw "checkparameters: set without values";
}

void usage()
{
    printf("Usage: pregutl [keys] [values] [@regfile]\n");
    printf("  -d   delete value or key\n");
    printf("  -c   create key\n");
    printf("  -s   set value\n");
    printf("  -r   recurse\n");
    printf("\n");
    printf("key-values are specified as follows:\n");
    printf("    :valuename  or :'value name'\n");
    printf("    :valuename=value\n");
    printf("\n");
    printf("values are specified as follows:\n");
    printf("    dword:01234567  or \"str\\n\\\"str\" or\n");
    printf("    multi_sz:\"...\",\"...\" or  hex:ff,11,22\n");
    printf("    note that you may also use single (') quotes instead\n");
    printf("    note also, that backslashes in strings must be escaped: \"\\\\windows\"\n");
    printf("    otherwise it would be almost impossible to enter quotes on the dos cmdline\n");
    printf("\n");
    printf("keys are specified as follows:\n");
    printf("    HKLM\\Software\\HTC\n");
    printf("\n");
    printf("if multiple keys and keyvals are specified pregutl\n");
    printf("operates on all combinations\n");
}

int main(int argc, char *argv[])
{
    DebugStdOut();

    RegAction action= DO_LIST;
    bool fRecurse= false;

    RegistryKeyList keys;
    StringList names;
    RegistryValueMap values;
    StringList regfiles;

    try {
    for (int i=1 ; i<argc ; i++)
    {
        if (argv[i][0]=='-') switch(argv[i][1])
        {
            case 'd': action= DO_DELETE; break;
            case 'c': action= DO_CREATE; break;
            case 's': action= DO_SET; break;
            case 'r': fRecurse= true; break;
            default:
                usage();
                return 1;
        }
        else if (IsValueNameSpec(argv[i]))
        {
            names.push_back(GetNameFromValueNameSpec(argv[i], 1));
        }
        else if (IsSetSpec(argv[i]))
        {
            values[GetNameFromSetSpec(argv[i], 1)]= RegistryValue::FromValueSpec(GetValueSpecFromSetSpec(argv[i], 1));
        }
        else if (IsRegFile(argv[i]))
        {
            regfiles.push_back(GetFilenameFromRegfileSpec(argv[i]));
        }
        else
            keys.push_back(RegistryKey::FromKeySpec(std::string(argv[i])));
    }
    }
    catch(char *e) {
        printf("ERROR: %s\n", e);
        return 1;
    }
    if (!regfiles.empty())
        action= DO_LOAD;

    try {
        checkparameters(action, keys, names, values);
    }
    catch(char *e) {
        printf("ERROR: %s\n", e);
        return 1;
    }


#ifdef WINCEREGUTL
    if (FAILED(CeRapiInit()))
    {
        printf( "Failed to init rapi\n");
        return 1;
    }
#endif

    try {
    for (RegistryKeyList::iterator i= keys.begin() ; i!=keys.end() ; ++i)
    {
        switch(action)
        {
        case DO_CREATE:
            (*i).CreateKey();
            break;
        case DO_DELETE:
            if (names.size())
                for (StringList::iterator j= names.begin() ; j!=names.end() ; ++j)
                    (*i).DeleteValue(*j);
            else
                (*i).DeleteKey();
            break;
        case DO_LIST:
            (*i).DumpKey();
            break;
        case DO_SET:
            for (RegistryValueMap::iterator j= values.begin() ; j!=values.end() ; ++j)
                    (*i).SetValue((*j).first, (*j).second);
            break;
        case DO_LOAD:
            debug("ERROR: not expecting load here\n");
            break;
        }
    }
    for (StringList::iterator i= regfiles.begin() ; i!=regfiles.end() ; ++i)
    {
        ProcessRegFile(*i);
    }
    }
    catch(char *e) {
        printf("ERROR: %s\n", e);
        return 1;
    }

#ifdef WINCEREGUTL
    CeRapiUninit();
#endif

    return 0;
}


