/* (C) 2003 XDA Developers  itsme@xs4all.nl
 *
 * $Header: /var/lib/cvs/nah6utils/src/psetmem.cpp,v 1.6 2004/01/05 18:01:47 itsme Exp $
 */
#include <windows.h>
#include "ItsUtils.h"
#include "dllversion.h"

#include "debug.h"
#include "args.h"
#include <stdio.h>
#include <string.h>
#include <rapi.h>

#include <vector>
using namespace std;

HANDLE GetRapiProcessHandle();
void WriteProcessMemory(HANDLE hProc, DWORD dwOffset, DWORD dwLength, BYTE *pData);
HANDLE ITGetProcessHandle(char *szProcessName);

void usage()
{
    printf("Usage: psetmem [-p procname] [-h prochandle] start data ...\n");
    printf("    numbers can be specified as 0x1234abcd\n");
    //printf("    -1 -2 -4 : data are bytes/words/dwords\n");
}

int main( int argc, char *argv[])
{
    DebugStdOut();

    DWORD dwOffset=0;
    char *szProcessName= NULL;
    HANDLE hProc= INVALID_HANDLE_VALUE;
    bool bDirectMemoryAccess= false;    // false: use readprocessmemory, true: use readmemory

    vector<DWORD> data;
    int nDataSize= 1;

    int argsfound=0;
    for (int i=1 ; i<argc ; i++)
    {
        if (argv[i][0]=='-') switch(argv[i][1])
        {
            case 'h': HANDLEULOPTION(hProc, HANDLE); break;
            case 'p': HANDLESTROPTION(szProcessName); break;
            case 'm': bDirectMemoryAccess= true; break;

            case '1': case '2': case '4':
                nDataSize= argv[i][1]-'0';
                break;
            default:
                usage();
                return 1;
        }
        else switch (argsfound++)
        {
            case 0: dwOffset= strtoul(argv[i], 0, 0); break;
            default:
                    {
                    DWORD arg= strtoul(argv[i], 0, 0);
                    if ( (nDataSize==1 && arg>=0x100)
                            || (nDataSize==2 && arg>=0x10000) )
                    {
                        printf("data larger than specified size: %x >= %x\n", arg, 256<<nDataSize);
                        usage();
                        return 1;
                    }
                    data.push_back(arg);
                    }
        }
    }

    if (argsfound==0)
    {
        usage();
        return 1;
    }
    if (nDataSize==0) nDataSize= 1;

    DWORD bufsize= data.size()*nDataSize;
    BYTE *buf= new BYTE[bufsize];
    for (size_t i=0 ; i<data.size() ; i++)
    {
        switch(nDataSize)
        {
            case 1: buf[i]= (BYTE)data[i]; break;
            case 2: ((WORD*)buf)[i]= (WORD)data[i]; break;
            case 4: ((DWORD*)buf)[i]= data[i]; break;
        }
    }

    CheckITSDll();

    if (hProc!=INVALID_HANDLE_VALUE)
    {
        // - do nothing, process handle already there.
    }
    else if (szProcessName==NULL)
    {
        hProc= GetRapiProcessHandle();
        if (hProc==INVALID_HANDLE_VALUE) {
            debug("error getting process context\n");
            return 1;
        }
    }
    else {
        hProc= ITGetProcessHandle(szProcessName);
        if (hProc==INVALID_HANDLE_VALUE) {
            debug("error getting process context\n");
            return 1;
        }
    }
    if (bDirectMemoryAccess)
        hProc= NULL;

    WriteProcessMemory(hProc, dwOffset, bufsize, buf);

    CeRapiUninit();

    return 0;
}

HANDLE GetRapiProcessHandle()
{
    DWORD outsize=0;
    GetContextResult *r=NULL;

    HRESULT res= CeRapiInvoke(L"\\Windows\\ItsUtils.dll", L"ITGetContext",
            0, NULL, &outsize, (BYTE**)&r, NULL, 0);
    if (res)
    {
        error(res, "ITGetContext");
        return INVALID_HANDLE_VALUE;
    }

    HANDLE hProc= r->hProcess;
    LocalFree(r);
    return hProc;
}
HANDLE ITGetProcessHandle(char *szProcessName)
{
    DWORD insize= (strlen(szProcessName)+1)*sizeof(WCHAR);
    WCHAR *inbuf= (WCHAR*)LocalAlloc(LPTR, insize);

	_snwprintf((WCHAR*)inbuf, insize/sizeof(WCHAR), L"%hs", szProcessName);

    DWORD outsize=0;
    HANDLE *outbuf=NULL;

    HRESULT res= CeRapiInvoke(L"\\Windows\\ItsUtils.dll", L"ITGetProcessHandle",
            insize, (BYTE*)inbuf,
            &outsize, (BYTE**)&outbuf, NULL, 0);
    if (res) {
        error(res, "ITGetProcessHandle");
        return INVALID_HANDLE_VALUE;
    }

    HANDLE hproc= *outbuf;

    LocalFree(outbuf);

    return hproc;
}

bool ITWriteProcessMemory(HANDLE hProc, DWORD dwOffset, BYTE *buffer, DWORD dwBufLength, DWORD *pdwNumberOfBytesWritten)
{
    WriteProcessMemoryParams *inbuf;
    DWORD outsize=0;
    WriteProcessMemoryResult *outbuf=NULL;

    DWORD insize= sizeof(WriteProcessMemoryParams)+dwBufLength;
    inbuf= (WriteProcessMemoryParams *)LocalAlloc(LPTR, insize);

    inbuf->hProcess= hProc;
    inbuf->dwOffset= dwOffset;
    inbuf->nSize= dwBufLength;
    memcpy(inbuf->buffer, buffer, dwBufLength);

    outbuf= NULL; outsize= 0;
    HRESULT res= CeRapiInvoke(L"\\Windows\\ItsUtils.dll", L"ITWriteProcessMemory",
            insize, (BYTE*)inbuf,
            &outsize, (BYTE**)&outbuf, NULL, 0);
    if (res)
    {
        error(res, "ITWriteProcessMemory(%08lx, %08lx, %08lx)", hProc, dwOffset, dwBufLength);
        return false;
    }
    *pdwNumberOfBytesWritten= outbuf->dwNumberOfBytesWritten;
    LocalFree(outbuf);

    return true;
}

void WriteProcessMemory(HANDLE hProc, DWORD dwOffset, DWORD dwLength, BYTE *pData)
{
    while (dwLength)
    {
        DWORD dwNumberOfBytesWritten;
        if (!ITWriteProcessMemory(hProc, dwOffset, pData, min(dwLength, 32768), &dwNumberOfBytesWritten))
            return;

        if (dwNumberOfBytesWritten==0 || dwNumberOfBytesWritten>dwLength)
            return;
        dwLength -= dwNumberOfBytesWritten;
        dwOffset += dwNumberOfBytesWritten;
        pData += dwNumberOfBytesWritten;
    }
}

