/* (C) 2003 XDA Developers  itsme@xs4all.nl
 *
 * $Header: /var/lib/cvs/nah6utils/src/rapiinit.cpp,v 1.2 2004/01/05 18:01:47 itsme Exp $
 */
#include <windows.h>
#include <rapi.h>
#include <string>
#include <vector>

#include "rapiinit.h"


typedef std::vector<BYTE> ByteVector;
#define vectorptr(v)  (&(v)[0])

std::wstring ToWString(char* buf)
{
    WCHAR *wbuf= new WCHAR[strlen(buf)+1];

    _snwprintf(wbuf, strlen(buf), L"%hs", buf);

    std::wstring wstr(wbuf, wbuf+strlen(buf));
    delete wbuf;

    return wstr;
}

std::wstring ToWString(const std::string& str)
{
    return ToWString((char*)str.c_str());
}

bool WaitForDevice()
{
    char *szTitle= "ShipPhone";
    RAPIINIT rinit; 
    rinit.cbSize= sizeof(RAPIINIT);
    rinit.heRapiInit= NULL;
    rinit.hrRapiInit= -1;
    if (CeRapiInitEx(&rinit))
    {
        // ERROR initializing rapi
        return false;
    }

    do {
        // msgwait is used to be able to handle windows messages while waiting.
        DWORD res= MsgWaitForMultipleObjects(1, &rinit.heRapiInit, false, 1000, 0);
        if (res==WAIT_OBJECT_0)
            break;

        if (res!=WAIT_TIMEOUT)
        {
            // ERROR waiting for rapi init
            return false;
        }
        res= MessageBox(0, "Please connect your XDA to activesync", szTitle, MB_RETRYCANCEL);
        if (res==IDCANCEL)
        {
            // User decided not to connect
            return false;
        }
    } while(1);
    if (rinit.hrRapiInit)
    {
        // ERROR connecting to XDA
        return false;
    }
    return true;
}

bool CeCopyFileToDevice(const std::string& srcfile, const std::string& dstfile) 
{
    WIN32_FIND_DATA wfd;
    HANDLE hFind = FindFirstFile( srcfile.c_str(), &wfd);
    if (INVALID_HANDLE_VALUE == hFind)
    {
        // ERROR: Source/host file does not exist
        return false;
    }
    FindClose( hFind);

    if (wfd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
    {
        // ERROR: Source/host file specifies a directory
        return false;
    }

    std::string dstname= dstfile;

    DWORD dwAttr = CeGetFileAttributes( ToWString(dstname).c_str());
    if (0xFFFFFFFF  != dwAttr)
    {
        if (dwAttr & FILE_ATTRIBUTE_DIRECTORY)
        {
            dstname += "\\";
            dstname += wfd.cFileName;
        }
        else
        {
            // ERROR: File already exists.
            return false;
        }
    }

    HANDLE hSrc = CreateFile( srcfile.c_str(), GENERIC_READ, FILE_SHARE_READ,
                NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
    if (INVALID_HANDLE_VALUE == hSrc)
    {
        // ERROR: Unable to open source/host file
        return false;
    }

    HANDLE hDest = CeCreateFile( ToWString(dstname).c_str(), GENERIC_WRITE, FILE_SHARE_READ,
                NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
    if (INVALID_HANDLE_VALUE == hDest )
    {
        // ERROR: Unable to open WinCE file
        return false;
    }

    // Copying srcfile to WCE: dstname

    ByteVector buffer; buffer.resize(32768);

    DWORD dwNumRead;
    do
    {
        if (ReadFile( hSrc, vectorptr(buffer), (DWORD)buffer.size(), &dwNumRead, NULL))
        {
            DWORD dwNumWritten;
            if (!CeWriteFile( hDest, vectorptr(buffer), dwNumRead, &dwNumWritten, NULL))
            {
                // ERROR: Writing WinCE file
                CeCloseHandle( hDest);
                CloseHandle (hSrc);
                return false;
            }
        }
        else
        {
            // ERROR: Reading source file
            CeCloseHandle( hDest);
            CloseHandle (hSrc);
            return false;
        }
    } while (dwNumRead);
    CeCloseHandle( hDest);
    CloseHandle (hSrc);
    return true;
}

